theory Commands
  imports Main
begin

type_synonym ('L, 'S) rel
  = "'L \<Rightarrow> 'S \<Rightarrow> 'S \<Rightarrow> bool"

fun not' where
"not' P l s s' = (\<not> P l s s')"

fun both_not where
"both_not b l s s' = (\<not> b s \<and> \<not> b s')"

fun both where
"both b l s s' = (b s \<and> b s')"

fun and' where
"and' P Q l s s' = (P l s s' \<and> Q l s s')"

fun imp' where
"imp' P Q l s s' = (P l s s' \<longrightarrow> Q l s s')"

datatype ('I, 'O, 'L, 'E, 'S) prog
  = Input  'L "'I \<Rightarrow> 'S \<Rightarrow> 'S"
  | Output 'L "'S \<Rightarrow> 'O"
  | Checkpoint "'S \<Rightarrow> 'E"
  | Step   "'S \<Rightarrow> 'S"
  | Assume "('L, 'S) rel" (* relational assumption *)
  | Assert "('L, 'S) rel" (* relational assertion  *)
  | Seq    "('I, 'O, 'L, 'E, 'S) prog" "('I, 'O, 'L, 'E, 'S) prog"
  | If     "'S \<Rightarrow> bool" "('I, 'O, 'L, 'E, 'S) prog" "('I, 'O, 'L, 'E, 'S) prog"
  | While  "'S \<Rightarrow> bool" "('I, 'O, 'L, 'E, 'S) prog"

definition "Skip \<equiv> Step id"

datatype ('I, 'O, 'L, 'E, 'S) event
  = Tick
  | In  'L 'I
  | Out 'L 'O
  | Ckp 'E
  | Asm "('L, 'S) rel"

datatype ('I, 'O, 'L, 'E, 'S) conf
  = Stop 'S
  | Run  'S "('I, 'O, 'L, 'E, 'S) prog"

inductive step where
"step (s,In  l x)     (Run s (Input  l f))   (Stop (f x s))" |
"step (s,Out l (f s)) (Run s (Output l f))   (Stop s)"       |
"step (s,Ckp (f s))   (Run s (Checkpoint f)) (Stop s)"     |
"step (s,Tick)        (Run s (Step f))       (Stop (f s))"   |
"step (s,Tick)        (Run s (Assert \<phi>))     (Stop s)"       |
"step (s,Asm \<phi>)       (Run s (Assume \<phi>))     (Stop s)"       |

"step e (Run s c1) (Stop s')
  \<Longrightarrow> step e (Run s (Seq c1 c2))  (Run s' c2)" |
"step e (Run s c1) (Run s' c1')
  \<Longrightarrow> step e (Run s (Seq c1 c2))  (Run s' (Seq c1' c2))" |

"p s
  \<Longrightarrow> step (s,Tick) (Run s (If p c1 c2))  (Run s c1)" |
"\<not> p s
  \<Longrightarrow> step (s,Tick) (Run s (If p c1 c2))  (Run s c2)" |

"\<not> p s
  \<Longrightarrow> step (s,Tick) (Run s (While p c))  (Stop s)" |
"p s
  \<Longrightarrow> step (s,Tick) (Run s (While p c))  (Run s (Seq c (While p c)))"

inductive_cases [elim!]:    "step e (Run s (Input  l f)) k"
inductive_cases [elim!]:    "step e (Run s (Output l f)) k"
inductive_cases [elim!]:    "step e (Run s (Checkpoint f)) k"
inductive_cases [elim!]:    "step e (Run s (Step f)) k"
inductive_cases [elim!]:    "step e (Run s (Assert \<phi>)) k"
inductive_cases [elim!]:    "step e (Run s (Assume \<phi>)) k"
inductive_cases step_seqE[elim!]:    "step e (Run s (Seq c1 c2)) k"
inductive_cases [elim!]:    "step e (Run s (If p c1 c2)) k"
inductive_cases step_while: "step e (Run s (While p c)) k"

inductive steps where
steps_done[intro]: "steps [] k k" |
steps_step: "step e k1 k2 \<Longrightarrow> steps \<sigma> k2 k3 \<Longrightarrow> steps (e#\<sigma>) k1 k3"

lemma steps_rec:
  "steps (e#\<sigma>) k k'' = (\<exists> k'. step e k k' \<and> steps \<sigma> k' k'')"
  by (auto intro: steps_step elim: steps.cases)

lemma steps0[simp]:
"steps [] k k' = (k = k')"
  by (auto elim: steps.cases)

lemma steps1:
"steps [e] k k' = step e k k'"
  unfolding steps_rec[of e "[]" k k']
  by auto

lemma steps_rev:
  "steps (\<sigma>@[e]) k k'' = (\<exists> k'. steps \<sigma> k k' \<and> step e k' k'')"
  by (induction \<sigma> arbitrary: k)
     (auto simp add: steps1 steps_rec)

lemma step1[simp]:
"step e k k' \<Longrightarrow> steps [e] k k'"
  using steps_step[of e k k' "[]" k'] by auto

lemma step_Stop:
"step \<sigma> k k' \<Longrightarrow> k = Stop s \<Longrightarrow> False"
  by (induction \<sigma> k k' rule: step.induct) auto
corollary [simp]:
"step \<sigma> (Stop s) k' = False"
  using step_Stop by metis

lemma steps_Stop[simp]:
"steps \<sigma> (Stop s) k = (k = (Stop s) \<and> \<sigma> = [])"
  by (auto elim: steps.cases)

lemma step_exists:
  obtains e k where "step e (Run s c) k"
proof -
  have "\<exists> e k. step e (Run s c) k"
  proof (induction c)
    case (Seq c1 c2)
    then obtain e k where "step e (Run s c1) k"
      by auto
    then show ?case
      by (cases k) (meson step.intros)+
  next
    case (If p c1 c2)
    then show ?case
      by (cases "p s") (meson step.intros)+
  next
    case (While p c)
    then show ?case
      by (cases "p s") (meson step.intros)+
  qed (meson step.intros)+

  with that show ?thesis
    by auto
qed


end