theory Guarantee
  imports Secure
begin



definition assumption_failed where
"assumption_failed l \<sigma> \<sigma>'
  = (\<exists> n s s' \<phi>. n < length \<sigma> \<and> n < length \<sigma>' \<and> \<sigma> ! n = (s,Asm \<phi>) \<and> \<sigma>' ! n = (s',Asm \<phi>) \<and> \<not> \<phi> l s s')"

(* taking up to and excluding the nth item is sufficient
   because that Asm step will be dropped by hist anyhow *)
definition assumption_failed_at where
"assumption_failed_at n P H A \<phi> l \<tau> \<tau>' \<sigma> \<sigma>'
  = (\<exists> s s'. n < length \<sigma> \<and> n < length \<sigma>' \<and> \<sigma> ! n = (s,Asm \<phi>) \<and> \<sigma>' ! n = (s',Asm \<phi>) \<and>
             P l s s' \<and> auditing (\<sigma> ! n) l P H A \<and> (\<tau>@hist (take n \<sigma>), \<tau>'@hist (take n \<sigma>')) \<in> H l s s' \<and> \<not> \<phi> l s s')"

lemma assumption_failed_nil[simp]:
  "assumption_failed l [] [] = False"
  unfolding assumption_failed_def by auto

lemma assumption_failed_rec[simp]:
  "assumption_failed l (e # \<sigma>2) (e' # \<sigma>2') =
    ((\<exists> s s' \<phi>. e = (s,Asm \<phi>) \<and> e' = (s',Asm \<phi>) \<and> \<not> \<phi> l s s')
      \<or> assumption_failed l \<sigma>2 \<sigma>2')"
  unfolding assumption_failed_def
  by (auto, case_tac n, auto) 

lemma assumption_failed_here[intro]:
  assumes "assumption_failed_at n P H A \<phi> l \<tau> \<tau>' \<sigma> \<sigma>'"
  shows   "assumption_failed l \<sigma> \<sigma>'"
  using assms unfolding assumption_failed_def assumption_failed_at_def by blast

lemma assumption_failed_at_nil[simp]:
  "assumption_failed_at n P H A \<phi> l \<tau> \<tau>' [] [] = False"
  unfolding assumption_failed_at_def by auto

lemma assumption_failed_at_rec[simp]:
  "assumption_failed_at n P H A \<phi> l \<tau> \<tau>' (e # \<sigma>2) (e' # \<sigma>2') =
    ((\<exists> s s'. n = 0 \<and> e = (s,Asm \<phi>) \<and> e' = (s',Asm \<phi>) \<and> P l s s' \<and> auditing e l P H A \<and> (\<tau>@hist [e], \<tau>'@hist [e']) \<in> H l s s' \<and> \<not> \<phi> l s s')
      \<or> (\<exists> m. n = Suc m \<and> assumption_failed_at m P H A \<phi> l (\<tau>@hist[e]) (\<tau>'@hist [e']) \<sigma>2 \<sigma>2'))"
  unfolding assumption_failed_at_def
  by (cases n) (auto simp add: hist_Cons)

lemma step_asm_helpme:
  "step e k k' \<Longrightarrow> e = (s0,Asm \<phi>) \<Longrightarrow> s0 = state k \<and> s0 = state k'"
  by (induction e k k' rule: step.induct) auto

lemma step_asm[elim]:
  assumes "step (s0,Asm \<phi>) (Run s c1) k"
  obtains "s0 = s" and "state k = s"
  using step_asm_helpme[OF assms] by simp

(* This key lemma connects step-wise characterization of security
   to some high-level properties of pairs of traces *)
lemma secure_stepsE:
  assumes "secure (length \<sigma> + n) l P1 H1 c1 Q H A"
  assumes "P1 l s1 s1'" "(\<tau>, \<tau>') \<in> H1 l s1 s1'"
  assumes "steps \<sigma>  (Run s1  c1) k3"
  assumes "steps \<sigma>' (Run s1' c1) k3'"
  assumes "length \<sigma> = length \<sigma>'"
  obtains (failed) m P' H' \<phi> k  k'     where
            "lockstep l (take (Suc m) \<sigma>) (take (Suc m) \<sigma>')"
            "steps (take (Suc m) \<sigma>)  (Run s1  c1) k"
            "steps (take (Suc m) \<sigma>') (Run s1' c1) k'"
            "matched_conf k k'"
            "assumption_failed_at m P' H' A \<phi> l \<tau> \<tau>' \<sigma> \<sigma>'" |
          (stop)      s3 s3'    where
            "lockstep l \<sigma> \<sigma>'"
            "\<not> assumption_failed l \<sigma> \<sigma>'"
            "k3 = Stop s3"   "k3' = Stop s3'"
            "Q l s3 s3'" "(\<tau>@hist \<sigma>, \<tau>'@hist \<sigma>') \<in> H l s3 s3'" |
          (run)    P3 H3 s3 s3' c3 where
            "lockstep l \<sigma> \<sigma>'"
            "\<not> assumption_failed l \<sigma> \<sigma>'"
            "k3 = Run s3 c3" "k3' = Run s3' c3"
            "P3 l s3 s3'" "(\<tau>@hist \<sigma>, \<tau>'@hist \<sigma>') \<in> H3 l s3 s3'"
            "secure n l P3 H3 c3 Q H A"
  using assms
proof (induction \<sigma> arbitrary: \<sigma>' P1 H1 \<tau> \<tau>' s1 s1' c1 k3 k3')
  case Nil
  then show ?case
    by auto
next
  case (Cons e \<sigma>2)
  note asmI = Cons.prems(1)
  note stopI = Cons.prems(2)
  note runI = Cons.prems(3)

  obtain e' \<sigma>2' where \<sigma>'[simp]: "\<sigma>' = e'#\<sigma>2'"
    by (metis Cons.prems(9) Suc_length_conv)
  from Cons.prems(7,8) obtain k2 k2' where
    step:   "step e   (Run s1  c1) k2" and
    step':  "step e'  (Run s1' c1) k2'" and
    steps:  "steps \<sigma>2  k2  k3" and
    steps': "steps \<sigma>2' k2' k3'"
    by (auto simp: steps_rec)

  from Cons.prems(4) have "secure (Suc (length \<sigma>2 + n)) l P1 H1 c1 Q H A"
    by simp
  from this Cons.prems(5,6) step step'
  show ?case
  proof (cases rule: secure_stepE)
    case asm
    with step step' Cons.prems(5,6) obtain P' \<phi> where "assumption_failed_at 0 P' H1 A \<phi> l \<tau> \<tau>' (e # \<sigma>2) \<sigma>'"
      by (auto elim!: lockstep.cases step_asm)
    with asm step step' show ?thesis
      by (auto intro!: asmI)
  next
    case (stop s2 s2')
    with step step' steps steps' show ?thesis
      by (fastforce intro!: stopI)
  next
    case (run P2 H2 c2 s2 s2')
    note IH = Cons.IH[of \<sigma>2' s2 c2 s2' "\<tau>@hist [e]" "\<tau>'@hist [e']" k3 k3' P2 H2]
    show ?thesis
    proof (rule IH, goal_cases)
      case (1 m k k' P' H' \<phi>)
      then have "assumption_failed_at (Suc m) P' H' A \<phi> l \<tau> \<tau>' (e # \<sigma>2) \<sigma>'"
        by auto
      moreover from 1 run have "lockstep l (take (Suc (Suc m)) (e # \<sigma>2)) (take (Suc (Suc m)) \<sigma>')"
        by auto
      moreover from step 1(2) run have "steps (take (Suc (Suc m)) (e # \<sigma>2))  (Run s1 c1) k"
        by (auto intro: steps_step)
      moreover from step' 1(3) run have "steps (take (Suc (Suc m)) \<sigma>') (Run s1' c1) k'"
        by (auto intro: steps_step)
      ultimately show ?case
        using asmI step step' 1(4) by blast
    next
      case (2 s3 s3')
      with run step step' show ?case
        by (fastforce intro!: stopI simp add: hist_Cons)
    next
      case (3 s3 c3 s3' P3 H3)
      with run step step' show ?case
        by (fastforce intro!: runI[where ?P3.0 = P3 and ?H3.0 = H3] simp add: hist_Cons)
    next
      case 4
      with run show ?case by simp
    next
      case 5
      with run show ?case by simp
    next
      case 6
      with run show ?case by simp
    next
      case 7
      with steps run show ?case by simp
    next
      case 8
      with steps' run show ?case by simp
    next
      case 9
      with Cons.prems(9) show ?case by simp
    qed
  qed
qed


fun visible where
"visible l [] = []" |
"visible l ((s,Tick)#\<sigma>)  = visible l \<sigma>" |
"visible l ((s,Ckp e)#\<sigma>) = visible l \<sigma>" |
"visible l ((s,Asm \<phi>)#\<sigma>) = visible l \<sigma>" |
"visible l ((s,In  l' x)#\<sigma>)  = (if l' \<le> l then Inl x # visible l \<sigma> else visible l \<sigma>)" |
"visible l ((s,Out l' x)#\<sigma>)  = (if l' \<le> l then Inr x # visible l \<sigma> else visible l \<sigma>)"

lemma visible_append[simp]:
  "visible l (\<sigma>1@\<sigma>2) = visible l \<sigma>1 @ visible l \<sigma>2"
  by (induction l \<sigma>1 rule: visible.induct) auto

(* what is directly observable to an attacker *)
definition equiv where
"equiv l \<sigma> \<sigma>' = (visible l \<sigma> = visible l \<sigma>' \<and> length \<sigma> = length \<sigma>')"

lemma equiv_empty[simp,intro]:
  "equiv l [] []"
  unfolding equiv_def by simp

lemma equiv_append[intro]:
  assumes "equiv l \<sigma>1 \<sigma>1'" "equiv l \<sigma>2 \<sigma>2'"
  shows   "equiv l (\<sigma>1@\<sigma>2) (\<sigma>1'@\<sigma>2')"
  using assms unfolding equiv_def by simp

lemma equiv_Cons[intro]:
  assumes "equiv l [e] [e']" "equiv l \<sigma>2 \<sigma>2'"
  shows   "equiv l (e#\<sigma>2) (e'#\<sigma>2')"
  using assms equiv_append[of l "[e]" "[e']" \<sigma>2 \<sigma>2']
  by simp

lemma lockstep_equiv[intro,simp]:
  assumes "lockstep l \<sigma> \<sigma>'"
  shows   "equiv l \<sigma> \<sigma>'"
  using assms unfolding equiv_def 
  by (induction rule: lockstep.induct) auto

definition uncertainty where
"uncertainty \<sigma> l P c s 
  \<equiv> {s'. \<exists> \<sigma>' k'. P l s s' \<and> steps \<sigma>' (Run s' c) k' \<and> equiv l \<sigma> \<sigma>'}"

(* Note: added same lengths and precondition , because it is established from uncertainty,
         and we need it later for the policy stuff *)
definition release where
"release \<sigma> l P c s
  = {s'. \<exists> \<sigma>' k'. P l s s' \<and> steps \<sigma>' (Run s' c) k' \<and> equiv l \<sigma> \<sigma>' \<and>  assumption_failed l \<sigma> \<sigma>' }"

(* shows that the last step of an execution
   bounds the decrease in uncertainty by the release policy *)
theorem attacker_learns_only_failed_assumptions:
  assumes steps:  "steps \<sigma> (Run s1 c1) (Run s2 c2)"
  assumes step:   "step  e (Run s2 c2) k3"
  assumes H1:     "\<And> s1'. ([],[]) \<in> H1 l s1 s1'"
  assumes secure: "\<And> n.   secure n l P1 H1 c1 Q H A"
  shows   "uncertainty \<sigma> l P1 c1 s1 - uncertainty (\<sigma>@[e]) l P1 c1 s1
           \<subseteq> release \<sigma> l P1 c1 s1"
proof (auto, goal_cases)
  case (1 s1')

  from 1(1) obtain \<sigma>' k' where
    P1:      "P1 l s1 s1'" and
    steps':  "steps \<sigma>' (Run s1' c1) k'" and
    eq:      "equiv l \<sigma> \<sigma>'"
    unfolding uncertainty_def by auto

  from eq have 
    lengths[simp]: "length \<sigma> = length \<sigma>'"
    unfolding equiv_def by auto

  from assms have steps2: "steps (\<sigma>@[e]) (Run s1 c1) k3"
    unfolding steps_rev by auto

  from secure[of "length \<sigma> + (Suc 0)"] P1 H1 steps steps' lengths
  show ?case
  proof (cases rule: secure_stepsE)
    case (failed m P' H' \<phi> k k')
    with P1 eq steps' show ?thesis
      unfolding release_def by auto
  next
    case (stop s3 s3')
    then show ?thesis by simp
  next
    case (run P2 H2 s2 s2' c2)
    obtain e' k3' where step': "step e' (Run s2' c2) k3'"
      by (rule step_exists)

    with run steps' have
      steps2': "steps (\<sigma>'@[e']) (Run s1' c1) k3'"
      unfolding steps_rev by auto

    from 1(2) P1 steps2' have
      neqs: "\<not> equiv l (\<sigma> @ [e]) (\<sigma>'@[e'])"
      unfolding uncertainty_def by auto
    with eq have neq: "\<not> equiv l [e] [e']"
      unfolding equiv_def by auto

    from run step have step: "step  e (Run s2 c2) k3"
      by simp (* Note: swaps blue variables to orange ones *)
    from run(7,5,6) step step' neq
    show ?thesis
      by (cases rule: secure_stepE) auto
  qed
qed

end