theory Policy
  imports Guarantee
begin

definition policy_excludes where
"policy_excludes D l \<tau> \<tau>'
  = (\<exists> z z' n \<phi> \<psi>. D = (\<phi>,\<psi>) \<and> n \<le> length \<tau> \<and> n \<le> length \<tau>' \<and>
                        \<phi> z l (take n \<tau>) \<and> \<phi> z' l (take n \<tau>') \<and> \<not> \<psi> z z' l (take n \<tau>) (take n \<tau>'))"

definition release' where
"release' D l \<sigma> c s
  = {s'. \<exists> \<sigma>' k'. steps \<sigma>' (Run s' c) k' \<and> equiv l \<sigma> \<sigma>' \<and> policy_excludes D l (hist \<sigma>) (hist \<sigma>') }"

lemma lockstep_hist_length:
  assumes "lockstep l \<sigma> \<sigma>'"
  shows   "length (hist \<sigma>) = length (hist \<sigma>')"
  using assms
  by (induction) auto

lemma hist_take_Suc:
  assumes "hist (take (Suc n) \<sigma>) = take m (hist \<sigma>)" "\<sigma> ! n = (s, Asm \<chi>)"
  shows   "hist (take n \<sigma>) = take m (hist \<sigma>)"
  using assms
    apply (induction n arbitrary: \<sigma> m)
     apply (case_tac \<sigma>, auto)
     apply (case_tac \<sigma>, auto)
    apply (case_tac b, auto)
     apply (case_tac m, auto)
  done

inductive audit where
"\<lbrakk>\<And> \<tau> \<tau>'      P H \<chi> s s'. (P,H,\<chi>) \<in> A \<Longrightarrow> P l s s' \<Longrightarrow> (\<tau>,\<tau>') \<in> H l s s' \<Longrightarrow> \<exists> z z'. \<phi> z l \<tau> \<and> \<phi> z' l \<tau>';
  \<And> z' z \<tau> \<tau>' P H \<chi> s s'. (P,H,\<chi>) \<in> A \<Longrightarrow> P l s s' \<Longrightarrow> (\<tau>,\<tau>') \<in> H l s s' \<Longrightarrow> \<phi> z l \<tau> \<Longrightarrow> \<phi> z' l \<tau>' \<Longrightarrow> \<psi> z z' l \<tau> \<tau>' \<Longrightarrow> \<chi> l s s'\<rbrakk>
  \<Longrightarrow> audit (\<phi>,\<psi>) l A"

inductive_cases auditE[elim!]: "audit (\<phi>,\<psi>) l A"

theorem audited_policy_covers_release:
  assumes steps:  "steps \<sigma> (Run s1 c1) k1"
  assumes H1:     "\<And> s1'. ([],[]) \<in> H1 l s1 s1'"
  assumes secure: "\<And> n. secure n l P1 H1 c1 Q H A"
  assumes [simp]: "D = (\<phi>,\<psi>)"
  assumes audit:  "audit D l A"

  shows   "release \<sigma> l P1 c1 s1
           \<subseteq> release' D l \<sigma> c1 s1"
proof (auto, goal_cases)
  case (1 s1')

  from 1 obtain \<sigma>' k1' where
      P1:     "P1 l s1 s1'" and
      steps':  "steps \<sigma>' (Run s1' c1) k1'" and
      eq:      "equiv l \<sigma> \<sigma>'" and
      failed_do_not_use_this_witness:  "assumption_failed l \<sigma> \<sigma>'"
    unfolding release_def by auto

  from eq have 
    lengths[simp]: "length \<sigma> = length \<sigma>'"
    unfolding equiv_def by auto

  obtain P' H' \<chi> k k' n where
    lockstep: "lockstep l (take (Suc n) \<sigma>) (take (Suc n) \<sigma>')"  and
    prefix:   "steps (take (Suc n) \<sigma>) (Run s1 c1) k" "steps (take (Suc n) \<sigma>') (Run s1' c1) k'" "matched_conf k k'" and
    failed:   "assumption_failed_at n P' H' A \<chi> l [] [] \<sigma> \<sigma>'"
    using secure[of "length \<sigma> + (Suc 0)"] P1 H1 steps steps' lengths
    by (blast elim: secure_stepsE intro: failed_do_not_use_this_witness)

  from failed obtain s s' P'' where
    n: "n < length \<sigma>" "n < length \<sigma>'" and
    s: "\<sigma> ! n = (s, Asm \<chi>)" "\<sigma>' ! n = (s', Asm \<chi>)" and
    \<chi>: "\<not> \<chi> l s s'" and
    H': "(hist (take n \<sigma>), hist (take n \<sigma>')) \<in> H' l s s'" and
    A: "(P'', H', \<chi>) \<in> A" and P'': "P'' l s s'"
    unfolding assumption_failed_at_def by auto

  from n(1) hist_prefix[of n "take (Suc n) \<sigma>"] obtain m where
    m: "m \<le> length (hist \<sigma>)" and z[simp]: "hist (take (Suc n) \<sigma>) = take m (hist \<sigma>)"
    by (metis hist_prefix nat_le_linear take_all)

  from n(2) hist_prefix obtain m' where
    m': "m' \<le> length (hist \<sigma>')" and z'[simp]: "hist (take (Suc n) \<sigma>') = take m' (hist \<sigma>')"
    by (metis hist_prefix nat_le_linear take_all)

  from lockstep lockstep_hist_length have "length (take m (hist \<sigma>)) = length (take m' (hist \<sigma>'))"
    by fastforce
  with m m' have [simp]: "m' = m" (* because the respective (take m _) and (take m' _) have exactly that length *)
    by simp

  from n have Suc_n: "length (take (Suc n) \<sigma>) = length (take (Suc n) \<sigma>')"
    by auto

  from z s(1) z' s(2) have
    [simp]: "hist (take n \<sigma>)  = take m (hist \<sigma>)"
            "hist (take n \<sigma>') = take m (hist \<sigma>')"
    by (auto intro: hist_take_Suc)

  (* this assumption has been audited, this needs to come from prove' vs some condition akin to secure *)
  from audit A P'' H' \<chi> obtain z z' where
    "\<phi> z l (take m (hist \<sigma>))" "\<phi> z' l (take m (hist \<sigma>'))" "\<not> \<psi> z z' l (take m (hist \<sigma>)) (take m (hist \<sigma>'))"
    by auto blast

  with steps' eq m m' show ?case
    unfolding release'_def policy_excludes_def
    by auto
qed


end