# Isabelle formalisation for Section 6

Of the policy audit (Section 6)
from the paper
"Assume but Verify: Deductive Verification of Leaked Information in Concurrent Applications"

## Building

Tested with Isabelle2022 https://isabelle.in.tum.de/index.html

Install Isabelle, making sure that the `isabelle` binary is in your `PATH` and then run:

```
isabelle build -c -d . -v Audit
```

It should produce output like:
```
[...]
Session Pure/Pure
Session Misc/Tools
Session HOL/HOL (main)
Session Unsorted/Audit
Cleaned Audit
Running Audit ...
Audit: theory Audit.Commands
Audit: theory Audit.Secure
Audit: theory Audit.Guarantee
Audit: theory Audit.Policy
Timing Audit (8 threads, 7.858s elapsed time, 19.930s cpu time, 1.694s GC time, factor 2.54)
Finished Audit (0:00:09 elapsed time, 0:00:20 cpu time, factor 2.27)
```

## Theory Organisation

Browse with

```
isabelle jedit filename.thy
```
* `Rules.thy`: audit rules and soundness proof
* `Policy.thy`: top-level policy soundness statement
* `Guarantee.thy`: policy agnostic guarantee 
* `Secure.thy`: semantic security property 
* `Commands.thy`: simple programming language and semantics


## Location of major definitions and theorems

The audit rules (Figure 7) are defined by the `prove` judgement in
Rules.thy

The proof establishes that these rules and checking the associated audit
obligations they generate (Definition 6.2) ensure the
policy-specific security guarantee (Theorem 6.5) as follows:

- theorem `prove_secure` in Rules.thy shows that the audit rules ensure a
   semantic condition that correctly captures the required policy audit
   obligations
- theorem `audited_policy_covers_release` in Policy.thy shows that if those
  audit obligations are fulfilled (Definition 6.2),
  then the policy specific security guarantee holds (Theorem 6.5).

The definition `audit` in Policy.thy formally defines in Isabelle the
meaning of Definition 6.2 in the paper.

