theory Rules
  imports Secure
begin

inductive prove where
prove_input:
  "\<lbrakk>\<And> s s' x x'. P l s s' \<Longrightarrow> Q l (f x s) (f x' s');
    \<And> s s' x x' \<tau> \<tau>'. (\<tau>, \<tau>') \<in> H l s s' \<Longrightarrow> (\<tau>, \<tau>') \<in> H' l (f x s) (f x' s');
    \<not> l' \<le> l\<rbrakk>
    \<Longrightarrow> prove l  P H (Input l' f) Q H' {}" |
prove_output:
  "\<lbrakk>\<And> s s'. l' \<le> l \<Longrightarrow> f s = f s'\<rbrakk>
    \<Longrightarrow> prove l P H (Output l' f) P H {}" |
prove_assume:
  "\<lbrakk>\<And> s s'. P l s s' \<Longrightarrow> \<phi> l s s' \<Longrightarrow> Q l s s'\<rbrakk>
    \<Longrightarrow> prove l P H (Assume \<phi>) Q H {(P,H,\<phi>)}" |
prove_checkpoint:
  "\<lbrakk>\<And> s s' \<tau> \<tau>'. (\<tau>,\<tau>') \<in> H l s s' \<Longrightarrow> (\<tau>@[f s],\<tau>'@[f s']) \<in> H' l s s'\<rbrakk>
    \<Longrightarrow> prove l P H (Checkpoint f) P H' {}" |
prove_step:
  "\<lbrakk>\<And> s s'. P l s s' \<Longrightarrow> Q l (f s) (f s');
    \<And> s s' \<tau> \<tau>'. (\<tau>, \<tau>') \<in> H l s s' \<Longrightarrow> (\<tau>, \<tau>') \<in> H' l (f s) (f s')\<rbrakk>
    \<Longrightarrow> prove l P H (Step f) Q H' {}" |
prove_assert:
  "\<lbrakk>\<And> s s'. P l s s' \<Longrightarrow> \<phi> l s s'\<rbrakk>
    \<Longrightarrow> prove l P H (Assert \<phi>) P H {}" |
prove_prepost_conseq:
  "\<lbrakk>prove l P' H1 c Q' H2 A;
    \<And> s s'. P l s s' \<Longrightarrow> P' l s s';
    \<And> s s'. Q' l s s' \<Longrightarrow> Q l s s'\<rbrakk>
    \<Longrightarrow> prove l P H1 c Q H2 A" |
prove_audit_conseq:
  "\<lbrakk>prove l P1 H1 c1 Q H A';
    weaker_audit l A' A\<rbrakk>
    \<Longrightarrow> prove l P1 H1 c1 Q H A" |
prove_frame:
  "\<lbrakk>prove l P1 H1 c1 Q H A;
    frame l F c1\<rbrakk>
    \<Longrightarrow> prove l (and' P1 F) H1 c1 (and' Q F) H { (and' P F, H, \<phi>) | P H \<phi>. (P,H,\<phi>) \<in> A }" |
prove_seq:
  "\<lbrakk>prove l P H c1 Q H' A; prove l Q H' c2 R H'' A\<rbrakk>
    \<Longrightarrow> prove l P H (Seq c1 c2) R H'' A" |
prove_if:
  "\<lbrakk>\<And> s s'. P1 l s s' \<Longrightarrow> b s = b s';
    prove l (and' (both     b) P1) H1 c1 Q H A;
    prove l (and' (both_not b) P1) H1 c2 Q H A\<rbrakk>
    \<Longrightarrow> prove l P1 H1 (If b c1 c2) Q H A" |
prove_while:
  "\<lbrakk>\<And> s s'. I l s s' \<Longrightarrow> b s = b s';
    \<And> s s'. P l s s' \<Longrightarrow> I l s s';
    \<And> s s'. I l s s' \<Longrightarrow> \<not> b s \<Longrightarrow> \<not> b s' \<Longrightarrow> Q l s s';
    prove l (and' (both b) I) H c I H A\<rbrakk>
    \<Longrightarrow> prove l P H (While b c) Q H A"


lemma prove_secure:
  assumes "prove l P1 H1 c Q H A"
  shows   "secure n l P1 H1 c Q H A"
  using assms
proof (induction arbitrary: n)
  case (prove_prepost_conseq l P' H1 c Q' H2 A P Q)
  show ?case
    apply (rule secure_prepost_conseq)
      apply (rule prove_prepost_conseq)
    using prove_prepost_conseq by auto
next
  case (prove_audit_conseq l P1 H1 c1 Q H A' A)
  with secure_audit_conseq show ?case
    by blast
next
  case (prove_frame l P1 H1 c1 Q H A F)
  with secure_frame show ?case
    by blast
next
  case (prove_seq l P H c1 Q H' A c2 R H'')
  with secure_Seq show ?case
    by blast
next
  case (prove_if P1 l b H1 H' c1 Q H A c2)
  with secure_If show ?case
    by blast
next
  case (prove_while I l b P Q H c A)
  with secure_While have "secure n l I H (While b c) Q H A" 
    by blast
  then show ?case
    apply (rule secure_prepost_conseq)
    using prove_while by blast
qed (auto intro: secure_oneI simp add: equiv_def)


end