theory Secure
  imports Commands
begin

inductive lockstep where
lockstep_Nil[intro,simp]:
  "lockstep l [] []" |
lockstep_Tick[intro!]:
  "lockstep l \<sigma> \<sigma>' \<Longrightarrow> lockstep l ((s,Tick)#\<sigma>) ((s',Tick)#\<sigma>')" |
lockstep_Ckp[intro!]:
  "lockstep l \<sigma> \<sigma>' \<Longrightarrow> lockstep l ((s,Ckp e)#\<sigma>) ((s',Ckp e')#\<sigma>')" |
lockstep_Asm[intro!]:
  "lockstep l \<sigma> \<sigma>' \<Longrightarrow> lockstep l ((s,Asm \<phi>)#\<sigma>) ((s',Asm \<phi>)#\<sigma>')" |
lockstep_In[intro!]:
  "lockstep l \<sigma> \<sigma>' \<Longrightarrow> (l' \<le> l \<Longrightarrow> x = x') \<Longrightarrow> lockstep l ((s,In l' x)#\<sigma>)  ((s',In l' x')#\<sigma>')" |
lockstep_Out[intro!]:
  "lockstep l \<sigma> \<sigma>' \<Longrightarrow> (l' \<le> l \<Longrightarrow> x = x') \<Longrightarrow> lockstep l ((s,Out l' x)#\<sigma>) ((s',Out l' x')#\<sigma>')"

lemma lockstep_Nil[simp]:
  "lockstep l [] \<sigma>' \<longleftrightarrow> \<sigma>' = []"
  by (auto elim: lockstep.cases)
  
lemma lockstep_Tick[simp]:
  "lockstep l [(s,Tick)] [(s',Tick)] = True"
  by auto

lemma lockstep_rec[simp]:
  assumes "lockstep l [e] [e']"
  shows   "lockstep l (e#\<sigma>) (e'#\<sigma>') = lockstep l \<sigma> \<sigma>'"
  using assms by (fastforce elim: lockstep.cases) (* much faster than auto *)

lemma lengths_lockstep:
  assumes "length \<sigma> = length \<sigma>'"
  obtains (stop) "\<sigma> = []" "\<sigma>' = []" |
          (step) e e' \<sigma>1 \<sigma>1' where "\<sigma> = e#\<sigma>1" "\<sigma>' = e'#\<sigma>1'"
  using assms
  apply (cases \<sigma>)
  apply auto
  apply (cases \<sigma>')
   apply auto
  done

(* select all Ckp events from a schedule *)
fun hist where
"hist [] = []" |
"hist ((_,Ckp e)#\<sigma>) = e#hist \<sigma>" |
"hist (_#\<sigma>) = hist \<sigma>"

(* NOTE: don't use always as simplifier rule, leads to loops *)
lemma hist_Cons:
  "hist [e]@hist \<sigma> = hist (e#\<sigma>)"
  apply (cases e)
  apply (case_tac b)
      apply auto
  done

lemma length_hist:
  "length (hist \<sigma>) \<le> length \<sigma>"
  by (induction \<sigma> rule: hist.induct) auto

lemma hist_prefix:
  assumes "n \<le> length \<sigma>"
  shows   "\<exists> m. m \<le> length (hist \<sigma>) \<and> hist (take n \<sigma>) = take m (hist \<sigma>)"
  using assms
proof (induction n arbitrary: \<sigma>)
  case (Suc n)
  show ?case
  proof (cases \<sigma>)
    case (Cons a \<sigma>')
    with Suc obtain m where IH: "m \<le> length (hist \<sigma>')" "hist (take n \<sigma>') = take m (hist \<sigma>')"
      by auto
    obtain s e where [simp]: "a = (s,e)"
      by fastforce
    from IH Cons show ?thesis
    proof (cases e)
      case (Ckp f)
      with IH Cons have "(Suc m) \<le> length (hist \<sigma>)" "hist (take (Suc n) \<sigma>) = take (Suc m) (hist \<sigma>)"
        by auto
      then show ?thesis
        by blast
    qed auto
  qed auto
qed auto

fun assuming where
"assuming (_,Asm \<phi>) l s s' = (\<phi> l s s')" |
"assuming  e l s s' = True"

lemma not_assumingE[elim]:
  assumes "\<not> assuming e l s s'"
  obtains s0 \<phi> where "e = (s0,Asm \<phi>)" "\<not> \<phi> l s s'"
  using assms
  apply (cases e)
  apply (case_tac b)
      apply auto
  done

(* provisioning for the consequence rule:
   we can instead collect weaker path conditions,
   but they must be covered by the current one

   note, if necessary, we can do the same for H and \<phi> *)
fun auditing where
"auditing (_,Asm \<phi>) l P H A
  = (\<exists> P'. ((P',H,\<phi>) \<in> A) \<and> (\<forall> s s'. P l s s' \<longrightarrow> P' l s s'))" |
"auditing  e l P H A = True"

fun state where
"state (Run s _) = s" |
"state (Stop s)  = s"

fun matched_conf where
"matched_conf (Stop s)   k' = (\<exists> s'.    k' = Stop s')" | 
"matched_conf (Run  s c) k' = (\<exists> s' c'. k' = Run s' c')"

inductive secure where
secure_stopI[intro]:
  "secure 0 l P1 H1 c1 Q H A" |
secure_stepI:
  "(\<And> s1 s1' \<tau> \<tau>' e e' k2 k2'.
    \<lbrakk>P1 l s1 s1'; (\<tau>, \<tau>') \<in> H1 l s1 s1';
     step e  (Run s1  c1) k2;
     step e' (Run s1' c1) k2'\<rbrakk>
      \<Longrightarrow> lockstep l [e] [e'] \<and> auditing e l P1 H1 A \<and>
          (  (matched_conf k2 k2' \<and> \<not> assuming e l (state k2) (state k2'))
           \<or> (\<exists> s2 s2'.          k2 = Stop s2   \<and> k2' = Stop s2'   \<and> Q  l s2 s2' \<and> (\<tau>@hist [e], \<tau>'@hist [e']) \<in> H l s2 s2')
           \<or> (\<exists> P2 H2 c2 s2 s2'. k2 = Run s2 c2 \<and> k2' = Run s2' c2 \<and> P2 l s2 s2' \<and> (\<tau>@hist [e], \<tau>'@hist [e']) \<in> H2 l s2 s2' \<and> secure n l P2 H2 c2 Q H A)))
    \<Longrightarrow> secure (Suc n) l P1 H1 c1 Q H A"

lemma secure_oneI:
  assumes
    "(\<And> s1 s1' \<tau> \<tau>' e e' k2 k2'.
      \<lbrakk>P1 l s1 s1'; (\<tau>, \<tau>') \<in> H1 l s1 s1';
       step e  (Run s1  c1) k2;
       step e' (Run s1' c1) k2'\<rbrakk>
        \<Longrightarrow> lockstep l [e] [e'] \<and> auditing e l P1 H1 A \<and>
            (  (matched_conf k2 k2' \<and> \<not> assuming e l (state k2) (state k2'))
             \<or> (\<exists> s2 s2'.          k2 = Stop s2   \<and> k2' = Stop s2'   \<and> Q  l s2 s2' \<and> (\<tau>@hist [e], \<tau>'@hist [e']) \<in> H l s2 s2')))"
    shows "secure n l P1 H1 c1 Q H A"
proof (cases n)
  case 0
  then show ?thesis
    by auto
next
  case (Suc m)
  show ?thesis 
    unfolding Suc
  proof (rule secure_stepI, goal_cases)
    case (1 s1 s1' \<tau> \<tau>' e e' k2 k2')
    from assms[OF this] show ?case
      by auto
  qed
qed

lemma secure_stepE:
  assumes sec: "secure (Suc n) l P1 H1 c1 Q H A"
  assumes pre:
          "P1 l s1 s1'"
          "(\<tau>, \<tau>') \<in> H1 l s1 s1'"
          "step e  (Run s1  c1) k2"
          "step e' (Run s1' c1) k2'"
        obtains 
          (asm)                        "lockstep l [e] [e']" "auditing e l P1 H1 A" "matched_conf k2 k2'" "\<not> assuming e l (state k2) (state k2')" |
          (stop) s2 s2'          where "lockstep l [e] [e']" "auditing e l P1 H1 A" "k2 = Stop s2"   "k2' = Stop s2'"   "Q  l s2 s2'" "assuming e l s2 s2'" "(\<tau>@hist [e], \<tau>'@hist [e']) \<in> H l s2 s2'" |
          (run)  P2 H2 c2 s2 s2' where "lockstep l [e] [e']" "auditing e l P1 H1 A" "k2 = Run s2 c2" "k2' = Run s2' c2" "P2 l s2 s2'" "assuming e l s2 s2'" "(\<tau>@hist [e], \<tau>'@hist [e']) \<in> H2 l s2 s2'" "secure n l P2 H2 c2 Q H A"
  using sec
proof (cases)
  case secure_stepI
  from this[OF pre] show ?thesis
    using asm stop run
    by fastforce
qed

definition weaker_audit where
"weaker_audit l A A' = (\<forall> P H \<phi>. (P,H,\<phi>) \<in> A \<longrightarrow> (\<exists> P'. (P',H,\<phi>) \<in> A' \<and> (\<forall> s s'. P l s s' \<longrightarrow> P' l s s')))"

lemma auditing_conseq[intro]:
  assumes "auditing e l P H A" "weaker_audit l A A'"
  shows   "auditing e l P H A'"
  using assms
  apply (cases e)
  apply (case_tac b)
  unfolding weaker_audit_def apply auto
  by blast

(* TODO: replace by proper semantic inclusion lemma *)
lemma secure_audit_conseq:
  assumes "secure n l P1 H1 c1 Q H A" "weaker_audit l A A'"
  shows   "secure n l P1 H1 c1 Q H A'"
  using assms
proof induction
  case (secure_stopI l P1 H1 c1 Q H A)
  then show ?case
    by auto
  next
    case (secure_stepI P1 l H1 c1 A Q H n)
    show ?case
    proof (rule secure.secure_stepI, goal_cases)
      case (1 s1 s1' \<tau> \<tau>' e e' k2 k2')
      with secure_stepI.IH[OF 1] secure_stepI.prems show ?case
        by blast
    qed
  qed

lemma auditing_pre_conseq[intro]:
  assumes "auditing e l P' H A"
  assumes "\<And> s s'. P l s s' \<Longrightarrow> P' l s s'"
  shows   "auditing e l P H A"
  using assms
  apply (cases e)
  apply (case_tac b)
  by auto

lemma auditing_frame_conseq:
  assumes "auditing e l P H A"
  shows   "auditing e l (and' P F) H { (and' P F, H, \<phi>) | P H \<phi>. (P,H,\<phi>) \<in> A }"
  using assms
  apply (cases e)
  apply (case_tac b)
  by auto

lemma secure_prepost_conseq:
  assumes "secure n l P1' H1 c1 Q' H A"
  assumes "\<And> s s'. P1 l s s' \<Longrightarrow> P1' l s s'"
  assumes "\<And> s s'. Q' l s s' \<Longrightarrow> Q  l s s'"
  shows   "secure n l P1 H1 c1 Q H A"
  using assms
proof (induction n arbitrary: P1 P1' H1 c1)
  case 0
  then show ?case
    by auto
next
  case (Suc n)
  note sec = Suc.prems(1)
  note pre = Suc.prems(2)
  note post = Suc.prems(3)

  show ?case
  proof (rule secure_stepI, goal_cases)
    case (1 s1 s1' \<tau> \<tau>' e e' k2 k2')
    with pre have P1: "P1' l s1 s1'"
      by blast
    from sec P1 1(2,3,4)
    show ?case
    proof (cases rule: secure_stepE)
      case asm
      with pre show ?thesis
        by auto
    next
      case (stop s2 s2')
      with pre post show ?thesis
        by auto
    next
      case (run P2 H2 c2 s2 s2')
      from Suc.IH post run have "secure n l P2 H2 c2 Q H A"
        by simp
      with pre run show ?thesis
        by auto
    qed
  qed
qed

inductive frame where
  "\<lbrakk>\<And> e e' s1 s1' s2 s2'.    F l s1 s1' \<Longrightarrow> step e  (Run s1  c1) (Stop s2) \<Longrightarrow> step e' (Run s1' c1) (Stop s2') \<Longrightarrow> F l s2 s2';
    \<And> e e' s1 s1' s2 s2' c2. F l s1 s1' \<Longrightarrow> step e  (Run s1  c1) (Run s2  c2) \<Longrightarrow> step e' (Run s1' c1) (Run s2' c2) \<Longrightarrow> F l s2 s2' \<and> frame l F c2\<rbrakk>
    \<Longrightarrow> frame l F c1"

lemma frame_stop:
  assumes "frame l F c1" "F l s1 s1'"
  assumes "step e  (Run s1  c1) (Stop s2)"
  assumes "step e' (Run s1' c1) (Stop s2')"
  shows   "F l s2 s2'"
  using assms
  by cases blast

lemma frame_run:
  assumes "frame l F c1" "F l s1 s1'"
  assumes "step e1  (Run s1  c1) (Run s2  c2)"
  assumes "step e1' (Run s1' c1) (Run s2' c2)"
  shows   "F l s2 s2'" "frame l F c2"
  using assms
  by (cases, blast)+

lemma secure_frame:
  assumes "secure n l P1 H1 c1 Q H A" "frame l F c1"
  shows   "secure n l (and' P1 F) H1 c1 (and' Q F) H { (and' P F, H, \<phi>) | P H \<phi>. (P,H,\<phi>) \<in> A }"
  using assms
proof (induction n arbitrary: P1 H1 c1)
  case 0
  then show ?case
    by auto
next
  case (Suc n)
  note IH = Suc.IH
  note sec = Suc.prems(1)
  note frame = Suc.prems(2)

  show ?case
  proof (rule secure_stepI, goal_cases)
    case (1 s1 s1' \<tau> \<tau>' e e' k2 k2')
    note step = 1(3) and step' = 1(4)
    from 1 have P1: "P1 l s1 s1'" and F1: "F l s1 s1'"
      by auto
    from sec P1 1(2) step step'
    show ?case
    proof (cases rule: secure_stepE)
      case asm
      then show ?thesis
        by (fastforce intro: auditing_frame_conseq)
    next
      case (stop s2 s2')
      from stop F1 frame step step' show ?thesis
        apply (auto simp del: and'.simps intro: auditing_frame_conseq)
        by (auto intro: frame_stop)
    next
      case (run P2 H2 c2 s2 s2')
      from F1 frame step step' have F2: "F l s2 s2'" "frame l F c2"
        unfolding run by (auto intro: frame_run)
      with run have "secure n l (and' P2 F) H2 c2 (and' Q F) H { (and' P F, H, \<phi>) | P H \<phi>. (P,H,\<phi>) \<in> A }"
        by (blast intro: IH)
      moreover have "auditing e l (and' P1 F) H1 { (and' P F, H, \<phi>) | P H \<phi>. (P,H,\<phi>) \<in> A }"
        using run by (intro auditing_frame_conseq)
      moreover have "(and' P2 F) l s2 s2'"
        using run F2 by simp
      ultimately show ?thesis
        using run by (auto simp del: and'.simps)
    qed
  qed
qed

lemma secure_Seq:
  assumes c1: "secure n l P1 H1 c1 P2 H2 A"
  (* the additional assumption m \<le> is useful in the while rule *)
  assumes c2: "\<And> m. m \<le> n \<Longrightarrow> secure m l P2 H2 c3 P3 H3 A"
  shows   "secure n l P1 H1 (Seq c1 c3) P3 H3 A"
  using assms
proof (induction n arbitrary: P1 H1 c1)
  case 0
  then show ?case 
    by auto
next
  case (Suc n)
  note sec = Suc.prems(1)  

  show ?case
  proof (rule secure_stepI, goal_cases)
    case (1 s1 s1' \<tau> \<tau>' e e' K2 K2')
    obtain k2 s2 where
      step: "step e (Run s1 c1) k2" and
      k2:   "K2 = Run s2 c3 \<and> k2 = Stop s2 \<or> (\<exists> c2. K2 = Run s2 (Seq c2 c3) \<and> k2 = Run s2 c2)"
      by (cases rule: step_seqE[OF 1(3)]) auto

    obtain k2' s2' where
      step': "step e' (Run s1' c1) k2'" and
      k2':   "K2' = Run s2' c3 \<and> k2' = Stop s2' \<or> (\<exists> c2'. K2' = Run s2' (Seq c2' c3) \<and> k2' = Run s2' c2')"
      by (cases rule: step_seqE[OF 1(4)]) auto

    from sec 1(1,2) step step'
    show ?case
    proof (cases rule: secure_stepE)
      case asm
      with k2 k2' show ?thesis
        by auto
    next
      case (stop)
      from Suc have "secure n l P2 H2 c3 P3 H3 A"
        by auto
      with stop k2 k2' show ?thesis
        by auto
    next
      case (run P2 H2 c2 s2 s2')
      from Suc run have "secure n l P2 H2 (Seq c2 c3) P3 H3 A"
        by auto
      with run k2 k2' show ?thesis
        by auto
    qed
  qed
qed

lemma secure_If_Suc:
  assumes "\<And> s s'. P1 l s s' \<Longrightarrow> b s = b s'"
  assumes "secure n l (and' (both     b) P1) H1 c1 Q H A"
  assumes "secure n l (and' (both_not b) P1) H1 c2 Q H A"
  shows   "secure (Suc n) l P1 H1 (prog.If b c1 c2) Q H A"
  using assms by (auto intro: secure_stepI)

lemma secure_If:
  assumes "\<And> s s'. P1 l s s' \<Longrightarrow> b s = b s'"
  assumes "\<And> n. secure n l (and' (both     b) P1) H1 c1 Q H A"
  assumes "\<And> n. secure n l (and' (both_not b) P1) H1 c2 Q H A"
  shows   "secure n l P1 H1 (prog.If b c1 c2) Q H A"
  using assms
  by (cases n) (auto intro: secure_If_Suc)

lemma secure_While:
    assumes "\<And> s s'. I l s s' \<Longrightarrow> b s = b s'"
    assumes "\<And> s s'. I l s s' \<Longrightarrow> \<not> b s \<Longrightarrow> \<not> b s' \<Longrightarrow> Q l s s'"
    assumes "\<And> n.    secure n l (and' (both b) I) H c I H A"
    shows "secure n l I H (While b c) Q H A"
  using assms
proof (induction n rule: nat_less_induct)
  case (1 n)
  then have IH: "\<And> m. m < n \<Longrightarrow> secure m l I H (While b c) Q H A"
    by auto
  note b = 1(2)
  note Q = 1(3)
  note c = 1(4)

  show ?case
  proof (cases n)
    case 0
    then show ?thesis 
      by auto
  next
    case (Suc m)
    show ?thesis
      unfolding Suc
    proof (rule secure_stepI, simp, goal_cases)
      case (1 s1 s1' \<tau> \<tau>' e e' k2 k2')
      note I = 1(1) note H = 1(2)
      from 1(3,4) Q I H b show ?case

      proof (elim step_while, goal_cases)
        case 4
        (* this obtains hides the concrete instance of P2 and H2 which would confuse automation otherwise *)
        obtain P2 H2 where c: "P2 l s1 s1'" "(\<tau>, \<tau>') \<in> H2 l s1 s1'" "secure m l P2 H2 (Seq c (While b c)) Q H A"
        proof
          from I 4 show "(and' (both b) I) l s1 s1'"
            by simp
        next
          from H show "(\<tau>, \<tau>') \<in> H l s1 s1'" .
        next
          show "secure m l (and' (both b) I) H (Seq c (While b c)) Q H A"
            by (fastforce simp add: Suc intro: secure_Seq c IH)
        qed
        from 4 I H c show ?case
          by auto
      qed auto
    qed
  qed
qed

end