theory Knowledge
  imports SecCSL
begin
context concurrency
begin

(* Extract the stack from a configuration *)
fun stack :: "'a conf \<Rightarrow> 'a state" where
  "stack (Run _ _ s _) = s" |
  "stack (Stop _ s _) = s" |
  "stack Abort = undefined"

(*
  Attacker knowledge (here modelled as uncertainty about the initial stack and heap) given an 
  execution beginning from an initial command~@{term c1}, stack~@{term s1} and locks~@{term \<Gamma>1},
  under precondition~@{term p1} and schedule~@{term \<sigma>}.

  Specifically the attacker knows that execution must have begun in a state~@{term "(s1',H1')"}
  such that: (1) the precondition holds and 
             (2) for every intermediate configuration~@{term k1} arrived at after some 
                 prefix~@{term "\<sigma>1"} of~@{term \<sigma>} (where @{term "\<sigma> = \<sigma>1@\<sigma>2"}),
                   @{term "(s1',H1')"} has a @{term sched_eq} execution resulting in some
                   configuration~@{term k1'} that is 
                     (2.a) low-equivalent to @{term k1} and for which
                     (2.b) any initial assumption made in @{term \<sigma>2} must hold.
*)
definition
  uncertainty :: "bool \<Rightarrow> 'level \<Rightarrow> assert \<Rightarrow> ('level,'a) sched \<Rightarrow> cmd \<Rightarrow> ctx \<Rightarrow> 'a state \<Rightarrow> 'a heap \<Rightarrow>  ('a state \<times> 'a heap) set"
  where
  "uncertainty sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1  \<equiv> 
   {(s1',H1'). local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1', H1') \<and> (\<forall> \<sigma>1 \<sigma>2 k1. \<sigma> = \<sigma>1@\<sigma>2 \<and> 
                cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k1 \<longrightarrow>
                (\<exists> \<sigma>1' k1'.  cmd_steps \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k1' \<and> sched_eq \<sigma>1 \<sigma>1' \<and>
                         (ct_eq lev \<sigma>1 \<sigma>1') \<and> 
                         ((\<forall>k2 \<sigma>3 \<sigma>4. cmd_step \<sigma>3 k1 k2 \<and> \<sigma>2=\<sigma>3@\<sigma>4 \<longrightarrow> \<not> asm lev \<sigma>3 (stack k1) (stack k1')))))}"

lemma uncertainty_ninD:
  "(s1',H1') \<notin> uncertainty sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1 \<Longrightarrow> local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1', H1') \<Longrightarrow> 
  \<exists> \<sigma>1 \<sigma>2 k1. \<sigma> = \<sigma>1 @ \<sigma>2 \<and> cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k1 \<and>
        (\<forall>\<sigma>1' k1'. sched_eq \<sigma>1 \<sigma>1' \<longrightarrow> (ct_eq lev \<sigma>1 \<sigma>1') \<longrightarrow> 
              cmd_steps \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k1' \<longrightarrow>
              (\<exists>k2 \<sigma>3. cmd_step \<sigma>3 k1 k2 \<and> (\<exists>\<sigma>4. \<sigma>2 = \<sigma>3 @ \<sigma>4) \<and> asm lev \<sigma>3 (stack k1) (stack k1')))"
  by(clarsimp simp: uncertainty_def split del: if_splits, blast)

lemma uncertaintyD:
  "(s1',H1') \<in> uncertainty sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1 \<Longrightarrow>
   local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1', H1') \<and> 
    (\<forall> \<sigma>1 \<sigma>2 k1. \<sigma> = \<sigma>1@\<sigma>2 \<and> 
                cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k1 \<longrightarrow>
                (\<exists> \<sigma>1' k1'.  cmd_steps \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k1' \<and> sched_eq \<sigma>1 \<sigma>1' \<and>
                         (ct_eq lev \<sigma>1 \<sigma>1') \<and>
                         ((\<forall>k2 \<sigma>3 \<sigma>4. cmd_step \<sigma>3 k1 k2 \<and> \<sigma>2=\<sigma>3@\<sigma>4 \<longrightarrow> \<not> asm lev \<sigma>3 (stack k1) (stack k1')))))"
  by(auto simp: uncertainty_def)

lemma uncertaintyD_unsafe:
  "(s1',H1') \<in> uncertainty sh lev p1 (\<sigma>1@\<sigma>2) c1 \<Gamma>1 s1 H1 \<Longrightarrow>
   local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1', H1') \<and> 
    (\<forall> k1. cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k1 \<longrightarrow>
                (\<exists> \<sigma>1' k1'.  cmd_steps \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k1' \<and> sched_eq \<sigma>1 \<sigma>1' \<and>
                         (ct_eq lev \<sigma>1 \<sigma>1') \<and> 
                         ((\<forall>k2 \<sigma>3 \<sigma>4. cmd_step \<sigma>3 k1 k2 \<and> \<sigma>2=\<sigma>3@\<sigma>4 \<longrightarrow> \<not> asm lev \<sigma>3 (stack k1) (stack k1')))))"
  by(auto simp: uncertainty_def)

lemma cmd_step_is_Run:
  "cmd_step \<sigma> k1 k2 \<Longrightarrow> \<exists>a b c d. k1 = Run a b c d"
  by(induct rule: cmd_step.induct, auto)

lemma cmd_step_blah:
  "cmd_step (\<sigma>3 @ \<sigma>4) k1 k2 \<Longrightarrow> cmd_step \<sigma>3 k1 k2a \<Longrightarrow>
  \<sigma>4 = [] \<and> k2a = k2"
  using cmd_step_is_Run cmd_sched_determ cmd_step_determ 
  by (metis append_eq_appendI cmd_step_Run self_append_conv) (* a little slow *)

lemma cmd_steps_sched_determ:
  assumes steps1: "cmd_steps \<sigma>1 k1 k2"
  assumes steps1a: "cmd_steps \<sigma>1a k1 ka" 
  assumes eq: "\<sigma>1 @ \<sigma>2 = \<sigma>1a @ \<sigma>2a"
  assumes len: "length \<sigma>1 \<ge> length \<sigma>1a"
  obtains \<sigma>1b where "\<sigma>1 = \<sigma>1a@\<sigma>1b"
  using steps1 steps1a eq len proof(induction arbitrary: \<sigma>1a ka rule: cmd_steps.induct)
case (cmd_steps_refl k)
  then show ?case 
    by simp
next
  case (cmd_steps_step \<sigma> k k' \<sigma>' k'')
  note that  = this
  from `cmd_steps \<sigma>1a k ka` show ?case proof(cases rule: cmd_steps.cases)
    case cmd_steps_refl
    then show ?thesis 
      using cmd_steps_step.prems(1) by blast
  next
    case (cmd_steps_step \<sigma>a ka' \<sigma>a')
    with that  show ?thesis 
      by (metis append_Nil2 append_assoc append_eq_append_conv_if)
  qed
qed

lemma cmd_steps_appD:
  "cmd_steps (\<sigma>1 @ \<sigma>2) k k2 \<Longrightarrow> cmd_steps \<sigma>1 k k1 \<Longrightarrow> cmd_steps \<sigma>2 k1 k2"
  apply(induct "\<sigma>1 @ \<sigma>2" k k2 arbitrary: \<sigma>1 \<sigma>2 k1 rule: cmd_steps.induct, clarsimp+)
  apply(erule cmd_steps.cases, clarsimp+)
   apply (fastforce simp: cmd_steps_step)
  apply clarsimp
  by (metis append_eq_append_conv cmd_sched_determ cmd_step_determ cmd_step_is_Run) 

lemma cmd_steps_trans:
  "cmd_steps \<sigma>1 k k1 \<Longrightarrow> cmd_steps \<sigma>2 k1 k2 \<Longrightarrow> cmd_steps (\<sigma>1 @ \<sigma>2) k k2"
  apply(induct arbitrary: \<sigma>2 k2 rule: cmd_steps.induct, clarsimp+)
  using cmd_steps_step by blast


lemma is_prefix_blah:
  assumes steps1: "cmd_steps \<sigma>1 k1 k2"
  assumes step2: "cmd_step \<sigma>2 k2 k3"
  assumes steps1a: "cmd_steps \<sigma>1a k1 ka" 
  assumes eq: "(\<sigma>1 @ \<sigma>2) = (\<sigma>1a @ \<sigma>2a)" 
  assumes nNil: "\<sigma>2a \<noteq> []"
  obtains \<sigma>1b where "\<sigma>1 = \<sigma>1a @ \<sigma>1b"
proof(cases "length \<sigma>1 \<ge> length \<sigma>1a")
  case True
  then show ?thesis using cmd_steps_sched_determ 
    using eq steps1 steps1a that by blast
next
  case False
  hence less: "length  \<sigma>1 < length \<sigma>1a" by auto
  from this obtain \<sigma>a where "\<sigma>1a = \<sigma>1 @ \<sigma>a" using cmd_steps_sched_determ 
    by (metis eq less_imp_le_nat steps1 steps1a)
  from this have "cmd_steps \<sigma>a k2 ka" using cmd_steps_appD 
    using steps1 steps1a by blast
  have steps12: "cmd_steps (\<sigma>1@\<sigma>2) k1 k3" using cmd_steps_trans 
    using cmd_step1_intro step2 steps1 by auto
  have steps2a: "cmd_steps \<sigma>2a ka k3"
    using cmd_steps_appD eq steps12 steps1a by auto
  from this obtain \<sigma>2a1 \<sigma>2a2 kb where
    "cmd_step \<sigma>2a1 ka kb" "cmd_steps \<sigma>2a2 kb k3" "\<sigma>2a = \<sigma>2a1@\<sigma>2a2" 
    by (metis cmd_steps.cases nNil)
  from this show ?thesis 
  proof -
    obtain tts :: "'a conf \<Rightarrow> ('a \<Rightarrow> 'a option) \<Rightarrow> (char list \<Rightarrow> 'a) \<Rightarrow> char list set \<Rightarrow> cmd \<Rightarrow> ('level,'a) tick list \<Rightarrow> ('level,'a) tick list" and ttsa :: "'a conf \<Rightarrow> ('a \<Rightarrow> 'a option) \<Rightarrow> (char list \<Rightarrow> 'a) \<Rightarrow> char list set \<Rightarrow> cmd \<Rightarrow> ('level,'a) tick list \<Rightarrow> ('level,'a) tick list" and cc :: "'a conf \<Rightarrow> ('a \<Rightarrow> 'a option) \<Rightarrow> (char list \<Rightarrow> 'a) \<Rightarrow> char list set \<Rightarrow> cmd \<Rightarrow> ('level,'a) tick list \<Rightarrow> 'a conf" where
      "\<forall>x0 x1 x2 x3 x4 x5. (\<exists>v6 v7 v8. x5 = v6 @ v7 \<and> cmd_step v6 (Run x4 x3 x2 x1) v8 \<and> cmd_steps v7 v8 x0) = (x5 = tts x0 x1 x2 x3 x4 x5 @ ttsa x0 x1 x2 x3 x4 x5 \<and> cmd_step (tts x0 x1 x2 x3 x4 x5) (Run x4 x3 x2 x1) (cc x0 x1 x2 x3 x4 x5) \<and> cmd_steps (ttsa x0 x1 x2 x3 x4 x5) (cc x0 x1 x2 x3 x4 x5) x0)"
      by moura
    then have f1: "\<forall>ts c C f fa ca. \<not> cmd_steps ts (Run c C f fa) ca \<or> ts = [] \<and> ca = Run c C f fa \<or> ts = tts ca fa f C c ts @ ttsa ca fa f C c ts \<and> cmd_step (tts ca fa f C c ts) (Run c C f fa) (cc ca fa f C c ts) \<and> cmd_steps (ttsa ca fa f C c ts) (cc ca fa f C c ts) ca"
      using cmd_steps_Run by presburger
    obtain cca :: "'a conf \<Rightarrow> cmd" and CC :: "'a conf \<Rightarrow> char list set" and aa :: "'a conf \<Rightarrow> char list \<Rightarrow> 'a" and zz :: "'a conf \<Rightarrow> 'a \<Rightarrow> 'a option" where
      "\<forall>x1. (\<exists>v3 v4 v5 v6. x1 = Run v3 v4 v5 v6) = (x1 = Run (cca x1) (CC x1) (aa x1) (zz x1))"
      by moura
    then have f2: "\<forall>ts c ca. \<not> cmd_step ts c ca \<or> c = Run (cca c) (CC c) (aa c) (zz c)"
      using cmd_step_is_Run by presburger
    then have f3: "cmd_steps \<sigma>a (Run (cca k2) (CC k2) (aa k2) (zz k2)) (Run (cca ka) (CC ka) (aa ka) (zz ka))"
      by (metis \<open>cmd_step \<sigma>2a1 ka kb\<close> \<open>cmd_steps \<sigma>a k2 ka\<close> step2)
    have f4: "ttsa (Run (cca ka) (CC ka) (aa ka) (zz ka)) (zz k2) (aa k2) (CC k2) (cca k2) \<sigma>a @ \<sigma>2a1 @ \<sigma>2a2 \<noteq> []"
      using \<open>\<sigma>2a = \<sigma>2a1 @ \<sigma>2a2\<close> nNil by fastforce
    have "\<sigma>a @ \<sigma>2a1 @ \<sigma>2a2 = \<sigma>2"
      using \<open>\<sigma>1a = \<sigma>1 @ \<sigma>a\<close> \<open>\<sigma>2a = \<sigma>2a1 @ \<sigma>2a2\<close> eq by auto
    then have "\<sigma>a \<noteq> tts (Run (cca ka) (CC ka) (aa ka) (zz ka)) (zz k2) (aa k2) (CC k2) (cca k2) \<sigma>a @ ttsa (Run (cca ka) (CC ka) (aa ka) (zz ka)) (zz k2) (aa k2) (CC k2) (cca k2) \<sigma>a \<or> \<not> cmd_step (tts (Run (cca ka) (CC ka) (aa ka) (zz ka)) (zz k2) (aa k2) (CC k2) (cca k2) \<sigma>a) (Run (cca k2) (CC k2) (aa k2) (zz k2)) (cc (Run (cca ka) (CC ka) (aa ka) (zz ka)) (zz k2) (aa k2) (CC k2) (cca k2) \<sigma>a) \<or> \<not> cmd_steps (ttsa (Run (cca ka) (CC ka) (aa ka) (zz ka)) (zz k2) (aa k2) (CC k2) (cca k2) \<sigma>a) (cc (Run (cca ka) (CC ka) (aa ka) (zz ka)) (zz k2) (aa k2) (CC k2) (cca k2) \<sigma>a) (Run (cca ka) (CC ka) (aa ka) (zz ka))"
      using f4 f2 by (metis append_assoc cmd_step_blah step2)
    then have "\<sigma>a = [] \<and> Run (cca ka) (CC ka) (aa ka) (zz ka) = Run (cca k2) (CC k2) (aa k2) (zz k2)"
      using f3 f1 by blast
    then show ?thesis
      by (simp add: \<open>\<sigma>1a = \<sigma>1 @ \<sigma>a\<close> that)
  qed
qed

lemma uncertainty_in_ninD_unsafe:
  assumes inu: "(s1',H1') \<in> uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1"
  assumes ninu: "(s1',H1') \<notin> uncertainty sh lev p1 (\<sigma>1@\<sigma>2) c1 \<Gamma>1 s1 H1"
  assumes inv: "local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1', H1')" 
  assumes steps: "cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k1"
  assumes step: "cmd_step \<sigma>2 k1 k2"
  assumes seq1: "sched_eq \<sigma>1 \<sigma>1' \<and> (ct_eq lev \<sigma>1 \<sigma>1')"
  assumes steps': "cmd_steps \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k1'"
  assumes step': "cmd_step \<sigma>2' k1' k2'"
  assumes seq2: "sched_eq \<sigma>2 \<sigma>2'"
  shows "asm lev \<sigma>2 (stack k1) (stack k1') \<or> \<not> (ct_eq lev \<sigma>2 \<sigma>2')"
proof -
  from step' steps' have steps2': "cmd_steps (\<sigma>1'@\<sigma>2') (Run c1 (set \<Gamma>1) s1' H1') k2'" using cmd_steps_trans 
    using cmd_step1_intro by blast
  from steps step have steps2: "cmd_steps (\<sigma>1@\<sigma>2) (Run c1 (set \<Gamma>1) s1 H1) k2" 
    using cmd_steps_trans 
    using cmd_step1_intro by blast
  from seq1 seq2 have seq12: "sched_eq (\<sigma>1 @ \<sigma>2) (\<sigma>1' @ \<sigma>2')"
    using sched_eq_appI by blast
  obtain \<sigma>1a \<sigma>2a k1a where
    sched_eq: "\<sigma>1 @ \<sigma>2 = \<sigma>1a @ \<sigma>2a" and
    steps1: "cmd_steps \<sigma>1a (Run c1 (set \<Gamma>1) s1 H1) k1a"  and
    stuff: "(\<forall>\<sigma>1a' k1'. 
          cmd_steps \<sigma>1a' (Run c1 (set \<Gamma>1) s1' H1') k1' \<longrightarrow>
          sched_eq \<sigma>1a \<sigma>1a' \<and> (ct_eq lev \<sigma>1a \<sigma>1a') \<longrightarrow>
          (\<exists>k2 \<sigma>3. cmd_step \<sigma>3 k1a k2 \<and> (\<exists>\<sigma>4. \<sigma>2a = \<sigma>3 @ \<sigma>4) \<and> asm lev \<sigma>3 (stack k1a) (stack k1')))"
    using ninu[THEN uncertainty_ninD, OF inv] by (clarsimp, blast)
  show ?thesis
  proof(cases "\<sigma>2a = []")
    case True
    then have \<sigma>1a_def [simp]: "\<sigma>1a = \<sigma>1 @ \<sigma>2" using sched_eq by simp
    with True steps2' steps2 stuff show ?thesis
      using seq1 seq12 by auto
  next
    case False

    obtain \<sigma>1b where \<sigma>1_def: "\<sigma>1 = \<sigma>1a @ \<sigma>1b" using is_prefix_blah 
      using False assms(4) local.step sched_eq steps1 by blast
    (* so \<sigma>1a is a prefix of \<sigma>1. Now if it is a strict prefix, I think we have a contradiction.
       Otherwise, asm must hold. *)
    then show ?thesis proof(cases "\<sigma>1b = []")
      case True
      with \<sigma>1_def have "\<sigma>1a = \<sigma>1" by simp
      then show ?thesis 
        by (metis append_Nil2 cmd_step_blah cmd_steps_determ local.step same_append_eq sched_eq seq1 steps steps' steps1 stuff)
    next
      case False 
      obtain \<sigma>1a' k1a' where steps1a': "cmd_steps \<sigma>1a' (Run c1 (set \<Gamma>1) s1' H1') k1a'" and
        seqa: "sched_eq \<sigma>1a \<sigma>1a' \<and> (ct_eq lev \<sigma>1a \<sigma>1a')" and
        blah: "(\<forall>k2 \<sigma>3 \<sigma>4. cmd_step \<sigma>3 k1a k2 \<and> \<sigma>1b = \<sigma>3 @ \<sigma>4 \<longrightarrow> \<not> asm lev \<sigma>3 (stack k1a) (stack k1a'))" 
        using \<sigma>1_def steps1 uncertaintyD[OF inu] by metis
      from steps steps1 \<sigma>1_def 
      have steps1b: "cmd_steps \<sigma>1b k1a k1" using cmd_steps_appD by blast

      with False obtain \<sigma>1b1 \<sigma>1b2 k1b where meh: "\<sigma>1b = \<sigma>1b1 @ \<sigma>1b2" "cmd_step \<sigma>1b1 k1a k1b" "cmd_steps \<sigma>1b2 k1b k1" 
        by (metis cmd_steps.cases)
      with blah have nasm: "\<not> asm lev \<sigma>1b1 (stack k1a) (stack k1a')" by blast
 
      from stuff[rule_format, OF steps1a']
      obtain k2a \<sigma>2a1 \<sigma>2a2 where meh2: "cmd_step \<sigma>2a1 k1a k2a" "\<sigma>2a = \<sigma>2a1 @ \<sigma>2a2" "asm lev \<sigma>2a1 (stack k1a) (stack k1a')" 
        using seqa by blast 
      from sched_eq \<sigma>1_def have \<sigma>2a_def: "\<sigma>2a = \<sigma>1b @ \<sigma>2" by simp
      have "\<sigma>2a1 = \<sigma>1b1" using cmd_sched_determ `cmd_step \<sigma>2a1 k1a k2a` `cmd_step \<sigma>1b1 k1a k1b` \<sigma>2a_def 
        by (metis append.assoc cmd_step_is_Run meh(1) meh2(2))
      with nasm meh2 have "False"  by simp
      thus ?thesis by blast
    qed
  qed
qed

   
lemma uncertaintyD_unsafe2:
  "(s1',H1') \<in> uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1 \<Longrightarrow> cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k1 \<Longrightarrow>
   local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1', H1') \<and>  
                (\<exists> \<sigma>1' k1'.  cmd_steps \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k1' \<and> 
                         sched_eq \<sigma>1 \<sigma>1' \<and> (ct_eq lev \<sigma>1 \<sigma>1') 
                         )"
  apply(clarsimp simp: uncertainty_def split del: if_splits)
  apply(drule_tac x=\<sigma>1 in spec)
  apply(drule_tac x="[]" in spec)
  by auto

lemma sched_asm_extend:
  "sched_asm \<sigma> = Some p \<Longrightarrow> sched_asm (\<sigma>@\<sigma>1) = Some p"
  by(induct arbitrary: \<sigma>1 rule: sched_asm.induct, auto)

lemma sched_asm_shorten:
  "cmd_step \<sigma>2 k1 k2 \<Longrightarrow> sched_asm (\<sigma>2@\<sigma>3) = Some p \<Longrightarrow> sched_asm \<sigma>2 = Some p"
  by(induct rule: cmd_step.induct, auto)

lemma asm_shorten:
  "\<lbrakk>asm lev (\<sigma> @ \<sigma>2) s s'; cmd_step \<sigma> k1 k2\<rbrakk> \<Longrightarrow> asm lev \<sigma> s s'"
  using sched_asm_shorten 
  by (meson asm.cases asm.intros)

(* Sanity check that attacker uncertainty can never increase. *)
lemma uncertainty_only_decreases:
  notes if_splits[split del]
  shows "uncertainty sh lev p1 (\<sigma>1@\<sigma>2) c1 \<Gamma>1 s1 H1 \<subseteq> 
        uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1"
  (* FIXME: horrible proof *)
  apply(subst (2) uncertainty_def)
  apply(clarsimp split del: if_splits | safe)+
   apply(auto simp: uncertainty_def)[1]
  apply(subst (asm) uncertainty_def)
  apply clarify
  apply(drule_tac x="\<sigma>1a" in spec)
  apply(drule_tac x="\<sigma>2a@\<sigma>2" in spec)
  apply(drule_tac x="k1" in spec)
  apply(clarsimp)
  apply(rule_tac x=\<sigma>1' in exI, simp)
  apply(rule_tac x=k1' in exI, simp)
  apply clarsimp
  done

(*
  Describe what it means for an attacker to learn nothing, for execution of a program
  @{term c1} under precondition~@{term p1}. Learning nothing means that the attacker can
  deduce only that execution must have begun in a state satisfying the precondition,
  whenever a non-assumption execution occurs.
*)
definition
  learns_nothing ::  "bool \<Rightarrow> 'level \<Rightarrow> assert \<Rightarrow> cmd \<Rightarrow> bool"
  where
  "learns_nothing sh lev p1 c1 \<equiv>
    \<forall> \<Gamma>1 \<sigma> s1 H1 k2.  no_asm \<sigma> \<and> set \<Gamma>1 \<subseteq> locks \<and>
       cmd_steps \<sigma> (Run c1 (set \<Gamma>1) s1 H1) k2 \<longrightarrow>
        uncertainty sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1 \<supseteq> 
           {(s1',H1'). local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1',H1')}"

lemma learns_nothing:
  "cmd_mod c1 \<inter> invvars = {} \<Longrightarrow> prove sh lev p1 c1 q1 \<Longrightarrow> learns_nothing sh lev p1 c1"
  using noninterference[where f=Emp] secure learns_nothing_def uncertainty_def
  apply(clarsimp simp: learns_nothing_def uncertainty_def split del: if_splits)
  apply(drule_tac f=Emp in noninterference)
        apply(blast intro: secure)
       apply assumption
      apply simp
     apply assumption
    apply assumption
   apply assumption
  by (metis asm_no_asm no_asm_appD)


(*
  Suppose command @{term c1} executes under schedule @{term \<sigma>1}. Then it makes
  one more step, consistent with schedule~@{term \<sigma>2}. This definition defines
  what the attacker is allowed to learn. Specifically the only thing they are
  allowed to learn is that any assumption made in the ``one more step'' has
  to hold. We define this by saying that the attacker is allowed to exclude
  from its uncertainty those initial states in which the assumption doesn't hold
  after the @{term \<sigma>1} execution.
*)
definition allowed_release :: "'level \<Rightarrow> ('level,'a) sched \<Rightarrow> cmd \<Rightarrow> ctx \<Rightarrow> 'a state \<Rightarrow> ('level,'a) sched \<Rightarrow> ('a state \<times> 'a heap) set" where
  "allowed_release lev \<sigma>1 c1 \<Gamma>1 s2 \<sigma>2 \<equiv> 
      {(s1',H1'). \<exists>\<sigma>1' k2'. cmd_steps \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k2' \<and>
                        sched_eq \<sigma>1 \<sigma>1'  \<and> (ct_eq lev \<sigma>1 \<sigma>1') \<and>
                        asm lev \<sigma>2 s2 (stack k2')}" 

(*
  A first attempt at the general security property.

  The following definition seems to implicitly assume 
  that uncertainty can only ever decrease. However, see
  below that that doesn't seem to be the case with the
  current definitions.

  Therefore the following definition seems very suspect...
*)
definition controlled_leakage ::  "bool \<Rightarrow> 'level \<Rightarrow> assert \<Rightarrow> cmd \<Rightarrow> bool"
  where
  "controlled_leakage sh lev p1 c1 \<equiv>
    \<forall> \<Gamma>1 \<sigma>1 s1 H1 k2 \<sigma>2 k2'.  set \<Gamma>1 \<subseteq> locks \<and>
       cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2 \<and>
       cmd_step \<sigma>2 k2 k2' \<longrightarrow>
        (uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1 -
         uncertainty sh lev p1 (\<sigma>1@\<sigma>2) c1 \<Gamma>1 s1 H1) 
        \<subseteq> allowed_release lev \<sigma>1 c1 \<Gamma>1 (stack k2) \<sigma>2"

lemma controlled_leakageD[dest]:
  "\<lbrakk>controlled_leakage sh lev p1 c1;
    set \<Gamma>1 \<subseteq> locks;
    cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2;
    cmd_step \<sigma>2 k2 k2'\<rbrakk> \<Longrightarrow>
        (uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1 -
         uncertainty sh lev p1 (\<sigma>1@\<sigma>2) c1 \<Gamma>1 s1 H1) 
        \<subseteq> allowed_release lev \<sigma>1 c1 \<Gamma>1 (stack k2) \<sigma>2"
  by(auto simp: controlled_leakage_def)

definition no_leakage ::  "bool \<Rightarrow> 'level \<Rightarrow> assert \<Rightarrow> cmd \<Rightarrow> bool"
  where
  "no_leakage sh lev p1 c1 \<equiv>
    \<forall> \<Gamma>1 \<sigma>1 s1 H1 k3 k2 \<sigma>2.  set \<Gamma>1 \<subseteq> locks \<and>
       cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2 \<and>
       cmd_steps \<sigma>2 k2 k3 \<and>
       no_asm (\<sigma>1@\<sigma>2) \<longrightarrow>
        (uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1 \<subseteq>
         uncertainty sh lev p1 (\<sigma>1@\<sigma>2) c1 \<Gamma>1 s1 H1)"

lemma no_leakageD[dest]:
  "\<lbrakk>no_leakage sh lev p1 c1; set \<Gamma>1 \<subseteq> locks;
    cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2;
    cmd_steps \<sigma>2 k2 k3;
    no_asm (\<sigma>1@\<sigma>2)\<rbrakk> \<Longrightarrow>
        (uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1 \<subseteq>
         uncertainty sh lev p1 (\<sigma>1@\<sigma>2) c1 \<Gamma>1 s1 H1)"
  by(auto simp: no_leakage_def)

inductive cmd_steps_rev :: "('level,'a) sched \<Rightarrow> 'a conf \<Rightarrow> 'a conf \<Rightarrow> bool" where
cmd_steps_rev_refl[intro]: "cmd_steps_rev [] k k" |
cmd_steps_rev_step:        "cmd_steps_rev \<sigma> k k' \<Longrightarrow> cmd_step \<sigma>' k' k'' \<Longrightarrow> cmd_steps_rev (\<sigma>@\<sigma>') k k''"


lemma cmd_steps_rev_Nil':
  "cmd_steps_rev \<sigma> k k' \<Longrightarrow> \<sigma> = [] \<Longrightarrow> k' = k"
  by(induct rule: cmd_steps_rev.induct, auto)

lemma cmd_steps_rev_Nil[simp]:
  "cmd_steps_rev [] k k' =  (k' = k)"
  apply(rule iffI)
   apply(blast intro: cmd_steps_rev_Nil')
  apply(blast intro: cmd_steps_rev_refl)
  done

lemma cmd_steps_rev_step1_intro[intro]:
  "cmd_step \<sigma>1 k1 k \<Longrightarrow> cmd_steps_rev \<sigma>1 k1 k"
  by (metis append.left_neutral cmd_steps_rev.simps cmd_steps_rev_refl)

lemma cmd_steps_rev_step:
  "cmd_steps_rev \<sigma>2 k2 k3 \<Longrightarrow> cmd_step \<sigma>1 k1 k2 \<Longrightarrow> cmd_steps_rev (\<sigma>1@\<sigma>2) k1 k3"
  apply(induct rule: cmd_steps_rev.induct, auto)
  by (metis append.assoc cmd_steps_rev_step)

lemma cmd_steps_rev:
  "cmd_steps \<sigma> k1 k2 = cmd_steps_rev \<sigma> k1 k2"
proof
  assume "cmd_steps \<sigma> k1 k2"
  thus "cmd_steps_rev \<sigma> k1 k2"
  proof(induct rule: cmd_steps.induct)
    case (cmd_steps_refl k)
    then show ?case
      by blast
  next
    case (cmd_steps_step \<sigma> k k' \<sigma>' k'')
    then show ?case using cmd_steps_rev_step by blast
  qed
next
  assume "cmd_steps_rev \<sigma> k1 k2"
  thus "cmd_steps \<sigma> k1 k2"
  proof(induct rule: cmd_steps_rev.induct)
    case (cmd_steps_rev_refl k)
    then show ?case by blast
  next
    case (cmd_steps_rev_step \<sigma> k k' \<sigma>' k'')
    then show ?case 
      using cmd_step1_intro cmd_steps_trans by blast
  qed
qed

lemma controlled_leakage_no_leakage':
  "\<lbrakk>cmd_steps \<sigma>2 k2 k3; cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2; controlled_leakage sh lev p1 c1; 
    (s1', H1') \<in> uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1;
        set \<Gamma>1 \<subseteq> locks; no_asm (\<sigma>1 @ \<sigma>2)\<rbrakk>
       \<Longrightarrow> (s1', H1') \<in> uncertainty sh lev p1 (\<sigma>1 @ \<sigma>2) c1 \<Gamma>1 s1 H1"
  unfolding cmd_steps_rev
proof(induction "\<sigma>2" "k2" k3 arbitrary: \<sigma>1 rule: cmd_steps_rev.induct)
  case cmd_steps_rev_refl
  then show ?case 
    by auto
next
  case (cmd_steps_rev_step \<sigma> k1 k2 \<sigma>s k3)
  have use_IH: "(s1', H1') \<in> uncertainty sh lev p1 (\<sigma>1 @ \<sigma>) c1 \<Gamma>1 s1 H1"
    by (metis append_assoc cmd_steps_rev_step.IH cmd_steps_rev_step.prems(1) cmd_steps_rev_step.prems(2) cmd_steps_rev_step.prems(3) cmd_steps_rev_step.prems(4) cmd_steps_rev_step.prems(5) no_asm_appD)  
  moreover have "no_asm \<sigma>s" 
    using cmd_steps_rev_step.prems(5) no_asm_appD by blast
  hence "\<And>lev \<sigma>1 c1 \<Gamma>1 s. allowed_release lev \<sigma>1 c1 \<Gamma>1 s \<sigma>s = {}"
    by(auto simp: allowed_release_def elim: asm.cases)
  moreover from 
    `cmd_steps_rev \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k1`
    `cmd_steps_rev \<sigma> k1 k2` 
  have "cmd_steps (\<sigma>1@\<sigma>) (Run c1 (set \<Gamma>1) s1 H1) k2" using cmd_steps_trans
    using cmd_steps_rev by auto
  ultimately show ?case
    using controlled_leakageD[OF `controlled_leakage sh lev p1 c1`, OF `set \<Gamma>1 \<subseteq> locks`]
  proof -
    have "\<forall>f fa p ts. p \<notin> uncertainty sh lev p1 ts c1 \<Gamma>1 fa f - uncertainty sh lev p1 (ts @ \<sigma>s) c1 \<Gamma>1 fa f \<or> \<not> cmd_steps ts (Run c1 (set \<Gamma>1) fa f) k2"
      using \<open>\<And>s lev c1 \<sigma>1 \<Gamma>1. allowed_release lev \<sigma>1 c1 \<Gamma>1 s \<sigma>s = {}\<close> \<open>\<And>s1 k2' k2 \<sigma>2 \<sigma>1 H1. \<lbrakk>cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2; cmd_step \<sigma>2 k2 k2'\<rbrakk> \<Longrightarrow> uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1 - uncertainty sh lev p1 (\<sigma>1 @ \<sigma>2) c1 \<Gamma>1 s1 H1 \<subseteq> allowed_release lev \<sigma>1 c1 \<Gamma>1 (stack k2) \<sigma>2\<close> cmd_steps_rev_step.hyps(2) by blast
    then show ?thesis
      by (metis (no_types) Diff_iff \<open>cmd_steps (\<sigma>1 @ \<sigma>) (Run c1 (set \<Gamma>1) s1 H1) k2\<close> append.assoc use_IH)
  qed
qed

lemma uncertainty_NilI:
  "local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1', H1') \<Longrightarrow>
   (s1', H1') \<in> uncertainty sh lev p1 [] c1 \<Gamma>1 s1 H1"
  apply(clarsimp simp: uncertainty_def split del: if_splits)
  apply(simp add: inv_def split del: if_splits)
  by blast

lemma no_leakage_learns_nothing':
  "\<lbrakk>no_leakage sh lev p1 c1; local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1', H1'); no_asm \<sigma>;
        set \<Gamma>1 \<subseteq> locks; cmd_steps \<sigma> (Run c1 (set \<Gamma>1) s1 H1) k\<rbrakk>
       \<Longrightarrow> (s1', H1') \<in> uncertainty sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1"
  apply(drule (1) no_leakageD)  
     apply(rule cmd_steps_refl)
    apply assumption
   apply simp
  using uncertainty_NilI
  by (simp add: subset_eq)
  
lemma no_leakage_learns_nothing:
  "no_leakage sh lev p1 c1 \<Longrightarrow> learns_nothing sh lev p1 c1"
  by(clarsimp simp: learns_nothing_def no_leakage_learns_nothing')

lemma controlled_leakage_no_leakage:
  "controlled_leakage sh lev p1 c1 \<Longrightarrow> no_leakage sh lev p1 c1"
  by(clarsimp simp: no_leakage_def controlled_leakage_no_leakage')
  
lemma controlled_leakage_learns_nothing:
  "controlled_leakage sh lev p1 c1 \<Longrightarrow> learns_nothing sh lev p1 c1"
  by(blast intro: controlled_leakage_no_leakage no_leakage_learns_nothing)


lemma sched_eq_appD:
  "sched_eq (\<sigma>1 @ \<sigma>2) \<sigma>' \<Longrightarrow> \<exists> \<sigma>1' \<sigma>2'. \<sigma>' = \<sigma>1' @ \<sigma>2' \<and> sched_eq \<sigma>1 \<sigma>1' \<and> sched_eq \<sigma>2 \<sigma>2'"
  apply(induct "\<sigma>1 @ \<sigma>2" arbitrary: \<sigma>1 \<sigma>2 \<sigma>', auto)
  apply(case_tac \<sigma>1, auto)
  by (metis append_Cons)

lemma cmd_step_determ:
"cmd_step \<sigma> k k' \<Longrightarrow> cmd_step \<sigma>' k k'' \<Longrightarrow> sched_eq \<sigma> \<sigma>' \<Longrightarrow> (k' = k'' \<and> \<sigma> = \<sigma>')"
  apply(induction \<sigma> k k' arbitrary: \<sigma>' k'' rule: cmd_step.induct)
                      apply auto[8]
               apply blast
              apply blast
             apply blast
            apply (auto elim: cmd_step_while)
  done

lemma sched_eq_appD_step:
  assumes 
     seq: "sched_eq (\<sigma>1 @ \<sigma>2) \<sigma>'" 
          "\<sigma>' = \<sigma>1' @ \<sigma>2'" and
   step': "cmd_step \<sigma>1' k k1'" and
    step: "cmd_step \<sigma>1 k k1"
  shows "sched_eq \<sigma>1 \<sigma>1' \<and> sched_eq \<sigma>2 \<sigma>2'"
  using step step' seq apply(induction arbitrary: \<sigma>1' \<sigma>2 \<sigma>' k1' rule: cmd_step.induct)
  by (auto elim: cmd_step_while)

lemma cmd_steps_determ:
  "cmd_steps \<sigma> k k' \<Longrightarrow> cmd_steps \<sigma>' k k'' \<Longrightarrow> sched_eq \<sigma> \<sigma>' \<Longrightarrow> k'' = k' \<and> \<sigma> = \<sigma>'"
proof(induction arbitrary: \<sigma>' k'' rule: cmd_steps.induct)
  case (cmd_steps_refl k)
  then show ?case 
    by auto
next
  case (cmd_steps_step \<sigma>1 k k1 \<sigma>2 k2 \<sigma>' k')
  note that = this
  with cmd_steps_step(4) show ?case proof(cases rule: cmd_steps.cases)
    case cmd_steps_refl
    with cmd_steps_step show ?thesis 
      by (metis append_is_Nil_conv cmd_step_sched_nonempty list.distinct(1) sched_eq.elims(2))
  next
    case (cmd_steps_step \<sigma>1' k1' \<sigma>2')
    from `sched_eq (\<sigma>1 @ \<sigma>2) \<sigma>'` `\<sigma>' = \<sigma>1' @ \<sigma>2'` `cmd_step \<sigma>1' k k1'` `cmd_step \<sigma>1 k k1`
    have "sched_eq \<sigma>1 \<sigma>1'" "sched_eq \<sigma>2 \<sigma>2'" using sched_eq_appD_step by blast+    
    then show ?thesis using cmd_step_determ that
      using local.cmd_steps_step(1) local.cmd_steps_step(2) local.cmd_steps_step(3) by blast
  qed
qed

lemma tick_eq_trans:
  "tick_eq a b \<Longrightarrow> tick_eq b c \<Longrightarrow> tick_eq a c"
  by(cases a, auto)

lemma tick_eq_sym:
  "tick_eq a b \<Longrightarrow> tick_eq b a"
  by(cases a, auto)

lemma sched_eq_trans:
  "sched_eq a b \<Longrightarrow> sched_eq b c \<Longrightarrow> sched_eq a c"
  apply(induct a arbitrary: b c)
   apply clarsimp+
  using tick_eq_trans by blast

lemma sched_eq_sym:
  "sched_eq a b \<Longrightarrow> sched_eq b a"
  apply(induct a arbitrary: b c)
   apply clarsimp+
  using tick_eq_sym by blast

(*
  Attempt to generalise @{thm secure_steps_ex} etc.

  Note that we need to state @{term secure} in the @{term Run} conclusion explicitly
  (and similarly for the side conditions on @{term \<Gamma>} and @{term c}),
  while @{thm secure_steps_ex} etc. does not, because the latter unwinds @{term cmd_steps}
  from the front (taking just one step) whereas we need to unwind from the back and so need
  an induction hypothesis that preserves @{term secure} across the inductive unwinding.
*)
lemma secure_uncertainty:
  assumes
    "cmd_steps_rev \<sigma> (Run c1 (set \<Gamma>1) s1  H1)  k"
  assumes
    "secure sh lev n p1 c1 q"
  assumes
    "n > length \<sigma>"
  assumes
    "cmd_steps_rev \<sigma>' (Run c1 (set \<Gamma>1) s1'  H1')  k'"
    "sched_eq \<sigma> \<sigma>' \<and> (ct_eq lev \<sigma> \<sigma>')"
    "(s1',H1') \<in> uncertainty sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1"
    "cmd_mod c1 \<inter> invvars = {}"
    "set \<Gamma>1 \<subseteq> locks"
  obtains
    (stop) \<Gamma> s H s' H' where
      "k = Stop (set \<Gamma>) s H"
      "k' = Stop (set \<Gamma>) s' H'"
      "inv sh lev \<Gamma> q Emp (s,H) (s',H')"
      "step sh lev c1 \<Gamma>1 p1 \<Gamma> q s1 s"
      "set \<Gamma> \<subseteq> locks" |
    (rec) p \<Gamma> c s H s' H' m where
      "k = Run c (set \<Gamma>) s H"
      "k' = Run c (set \<Gamma>) s' H'"
      "inv sh lev \<Gamma> p Emp (s,H) (s',H')"
      "step sh lev c1 \<Gamma>1 p1 \<Gamma> p s1 s"
      "secure sh lev m p c q" "m \<ge> n - length \<sigma>"
      "set \<Gamma> \<subseteq> locks"
      "cmd_mod c \<subseteq> cmd_mod c1"
      "cmd_vars c \<subseteq> cmd_vars c1"
  using assms
proof (induction  \<sigma> "(Run c1 (set \<Gamma>1) s1  H1)" k arbitrary: \<sigma>' c1 \<Gamma>1 s1 H1 p1 q s1' H1' k' n rule: cmd_steps_rev.induct)
  case cmd_steps_rev_refl
  from cmd_steps_rev_refl(2-) show ?case
    by(auto simp: uncertainty_def)
next
  case (cmd_steps_rev_step \<sigma>1 k1 \<sigma>2 k2)
  from `length (\<sigma>1 @ \<sigma>2) < n` have "length \<sigma>1 < n" by auto
  from `(s1', H1') \<in> uncertainty sh lev p1 (\<sigma>1 @ \<sigma>2) c1 \<Gamma>1 s1 H1` have
  "(s1', H1') \<in> uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1" using uncertainty_only_decreases by blast
  from `cmd_steps_rev \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k1` 
       `(s1', H1') \<in> uncertainty sh lev p1 (\<sigma>1 @ \<sigma>2) c1 \<Gamma>1 s1 H1`
       `cmd_step \<sigma>2 k1 k2`
  obtain \<sigma>1' k1' where
  inv: "local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1', H1')" and
  seq: "sched_eq \<sigma>1 \<sigma>1' \<and> (ct_eq lev \<sigma>1 \<sigma>1')" and
  steps': "cmd_steps \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k1'" and
  nasm: "\<not> asm lev \<sigma>2 (stack k1) (stack k1')"
    unfolding uncertainty_def using cmd_steps_rev
    by blast+
  from steps' have abc: "cmd_steps_rev \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k1'" by(simp add: cmd_steps_rev)

  show ?case    
    apply(rule cmd_steps_rev_step(2)[OF refl _ _ `secure sh lev n p1 c1 q` `length \<sigma>1 < n`
                                        `cmd_steps_rev \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k1'` seq
                                        `(s1', H1') \<in> uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1`
                                        `cmd_mod c1 \<inter> invvars = {}` `set \<Gamma>1 \<subseteq> locks`])
    using `cmd_step \<sigma>2 k1 k2`
          apply blast
        
  proof -
    fix  c \<Gamma> s H s' H' p m 
    assume k1_def: "k1 = Run c (set \<Gamma>) s H" and k1'_def: "k1' = Run c (set \<Gamma>) s' H'" and
           inv: "local.inv sh lev \<Gamma> p Emp (s, H) (s', H')" and 
           s: "step sh lev c1 \<Gamma>1 p1 \<Gamma> p s1 s" and m: "m \<ge> n - length \<sigma>1" and 
           sec: "secure sh lev m p c q" and \<Gamma>: "set \<Gamma> \<subseteq> locks" and mod: "cmd_mod c \<subseteq> cmd_mod c1"
       and vars: "cmd_vars c \<subseteq> cmd_vars c1"
    
    obtain m' where m_def: "m = Suc m'" using m `n > length \<sigma>1`
      by (metis diff_is_0_eq diff_zero not_less old.nat.exhaust)

    obtain \<sigma>2' k2' where "cmd_step \<sigma>2' k1' k2'" "sched_eq \<sigma>2 \<sigma>2'" using `cmd_step \<sigma>2 k1 k2` 
      `k1 = Run c (set \<Gamma>) s H` `k1' = Run c (set \<Gamma>) s' H'`
      using cmd_step_Run by blast

    have "sched_eq (\<sigma>1 @ \<sigma>2) \<sigma>'" 
      by (simp add: cmd_steps_rev_step.prems(6))
    have "sched_eq \<sigma>1 \<sigma>1'"
      by (simp add: local.seq)
    from `sched_eq (\<sigma>1 @ \<sigma>2) \<sigma>'` `sched_eq \<sigma>1 \<sigma>1'` `sched_eq \<sigma>2 \<sigma>2'` 
       cmd_steps_rev_step(8) steps' `cmd_step \<sigma>2' k1' k2'`
    have "sched_eq \<sigma>' (\<sigma>1' @ \<sigma>2')" using sched_eq_trans sched_eq_sym
      by (meson sched_eq_appI)
    hence det: "k2' = k'" using cmd_step1_intro cmd_steps_determ cmd_steps_rev cmd_steps_trans local.cmd_steps_rev_step(8) steps' 
      by (meson \<open>cmd_step \<sigma>2' k1' k2'\<close>)

    from `cmd_step \<sigma>2 k1 k2` have "\<sigma>2 \<noteq> []" 
      using cmd_step_sched_nonempty by blast

    with `n > length (\<sigma>1 @ \<sigma>2)` have "n > Suc (length \<sigma>1)" 
      using less_imp_Suc_add by force

    have len: "n - length (\<sigma>1 @ \<sigma>2) \<le> m'"
    proof -
      have "Suc (length \<sigma>1) \<le> length \<sigma>1 + length \<sigma>2"
        by (simp add: Suc_leI \<open>\<sigma>2 \<noteq> []\<close>)
      then show ?thesis
       using m m_def by auto
    qed

    have p: "pre sh lev \<sigma>2 \<sigma>2' (set \<Gamma>) p Emp c k2 k2' (s,H) (s',H')"
      apply(rule pre_intro, simp)
      using `cmd_mod c \<subseteq> cmd_mod c1` `cmd_mod c1 \<inter> invvars = {}` apply blast
         apply(rule `set \<Gamma> \<subseteq> locks`)
      using `cmd_step \<sigma>2 k1 k2` `k1 = Run c (set \<Gamma>) s H` 
        apply simp
      using `cmd_step \<sigma>2' k1' k2'` `k1' = Run c (set \<Gamma>) s' H'` 
        apply simp
      apply(rule `sched_eq \<sigma>2 \<sigma>2'`)
      by (rule `local.inv sh lev \<Gamma> p Emp (s, H) (s', H')`)

    from secure_step[OF sec[simplified m_def] this] nasm k1_def k1'_def
    have rec: "rec sh lev m' (set \<Gamma>) p Emp c k2 k2' q (s, H) (s', H')" by auto
    thus ?thesis apply -
      apply(erule rec_elim) 
       apply(rule cmd_steps_rev_step(4))
           apply assumption
      using det 
          apply simp
         apply assumption
      using s step_trans
      using \<open>cmd_vars c \<subseteq> cmd_vars c1\<close> 
        apply blast
      using step_elim apply blast
      apply(rule cmd_steps_rev_step(5))
              apply assumption
      using det 
             apply  simp
            apply assumption
      using s step_trans
      using \<open>cmd_vars c \<subseteq> cmd_vars c1\<close> 
           apply blast
          apply assumption
         apply (rule len)
      using step_elim 
        apply blast
      using cmd_step_preserves_mod cmd_steps_rev_step.hyps(3) k1_def mod 
       apply fastforce
      by (metis cmd_step_preserves_vars cmd_steps_rev_step.hyps(3) conf_vars.simps(3) k1_def rev_subsetD subsetI vars)
  qed
qed

lemma asm_sched_eq:
  "cmd_step \<sigma>2 k  k1 \<Longrightarrow>
   asm lev \<sigma>2 s s' \<Longrightarrow> sched_eq \<sigma>2 \<sigma>2' \<Longrightarrow> 
    \<sigma>2 = \<sigma>2'"
  by(induct \<sigma>2 k k1 arbitrary: \<sigma>2' s s' rule: cmd_step.induct, auto simp: asm.simps)

lemma controlled_release:
  "cmd_mod c1 \<inter> invvars = {} \<Longrightarrow> prove sh lev p1 c1 q1 \<Longrightarrow> controlled_leakage sh lev p1 c1"
  apply(clarsimp simp: controlled_leakage_def)
  apply(rename_tac s1' H1' k3)
  apply(frule uncertainty_ninD)
  using uncertaintyD 
   apply blast
  apply clarsimp
  apply(frule uncertaintyD_unsafe2)
   apply assumption
  apply clarsimp
  apply(subst (asm) cmd_steps_rev)
  apply(frule_tac \<sigma>'=\<sigma>1' in secure_uncertainty)
          apply(erule secure)
         apply blast
        apply(subst cmd_steps_rev[symmetric])
         apply assumption
        apply blast
      apply assumption
     apply assumption
     apply assumption
    prefer 3
    apply assumption

   apply clarsimp

  apply(subgoal_tac "\<exists>\<sigma>2' k2'. cmd_step \<sigma>2' k1' k2' \<and> sched_eq \<sigma>2 \<sigma>2'")
   apply(elim exE)
   apply(subgoal_tac "(ct_eq lev \<sigma>2 \<sigma>2')")
    apply(drule uncertainty_in_ninD_unsafe)
             apply(fastforce dest: sym)
            apply assumption
           apply(fastforce simp: cmd_steps_rev)
          apply assumption+
        apply blast
       apply assumption+
     apply blast
    apply blast
   apply(case_tac "asm lev \<sigma>2 (stack k2) (stack k1') ")
    apply(clarsimp simp: allowed_release_def)
     apply (metis stack.simps(1))
    apply clarsimp
    apply(subgoal_tac "\<exists>m'. m = Suc m'")
     apply(erule exE)
     apply clarsimp 
     apply(drule_tac \<sigma>=\<sigma>2 and \<sigma>'=\<sigma>2' and f=Emp and k=k3 and k'=k2'  in secure_step)
      apply(rule pre_intro)
            apply clarsimp
           apply blast
          apply (assumption, assumption)
        apply assumption
       apply assumption
     apply assumption
    apply (metis asm_sched_eq ct_eq_refl)
   apply (metis Suc_n_not_le_n diff_0_eq_0 diff_diff_cancel diff_is_0_eq not0_implies_Suc)
  using cmd_step_Run  
  by meson

definition
  exclusion_knowledge :: "bool \<Rightarrow> 'level \<Rightarrow> assert \<Rightarrow> ('level,'a) sched \<Rightarrow> cmd \<Rightarrow> ctx \<Rightarrow> 'a state \<Rightarrow> 'a heap \<Rightarrow>  ('a state \<times> 'a heap) set"
  where
  "exclusion_knowledge sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1 \<equiv> UNIV - uncertainty sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1"

lemma subset_cong_blah:
  "A = B \<Longrightarrow> (A \<subseteq> C) = (B \<subseteq> C)"
  by(rule arg_cong)


(*
  Prove the definition above equivalent to the more ``standard'' one phrased in terms of
  exclusion knowledge, even though I prefer the above formalisation.
*)
lemma controlled_leakage_def2:
  "controlled_leakage sh lev p1 c1 =
 (\<forall> \<Gamma>1 \<sigma>1 s1 H1 k2 \<sigma>2 k2'.  set \<Gamma>1 \<subseteq> locks \<and>
       cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2 \<and>
       cmd_step \<sigma>2 k2 k2' \<longrightarrow>
        (exclusion_knowledge sh lev p1 (\<sigma>1@\<sigma>2) c1 \<Gamma>1 s1 H1 -
          exclusion_knowledge sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1) 
        \<subseteq> allowed_release lev \<sigma>1 c1 \<Gamma>1 (stack k2) \<sigma>2)"
  apply(simp add: exclusion_knowledge_def controlled_leakage_def)
  apply(rule iff_allI)+
  apply(rule imp_cong, rule refl)
  apply(rule subset_cong_blah)
  by blast

(*
  Attacker knowledge (here modelled as uncertainty about the initial stack and heap) given an 
  execution beginning from an initial command~@{term c1}, stack~@{term s1} and locks~@{term \<Gamma>1},
  under precondition~@{term p1} and schedule~@{term \<sigma>}.

  Specifically the attacker knows that execution must have begun in a state~@{term "(s1',H1')"}
  such that: (1) the precondition holds and 
             (2) for every configuration~@{term k1} arrived at after @{term \<sigma>},
                   @{term "(s1',H1')"} has a @{term sched_eq} execution resulting in some
                   configuration~@{term k1'} that is low-equivalent to @{term k1} 
*)  
definition
  uncertainty2 :: "bool \<Rightarrow> 'level \<Rightarrow> assert \<Rightarrow> ('level,'a) sched \<Rightarrow> cmd \<Rightarrow> ctx \<Rightarrow> 'a state \<Rightarrow> 'a heap \<Rightarrow>  ('a state \<times> 'a heap) set"
  where
  "uncertainty2 sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1  \<equiv> 
   {(s1',H1'). local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1', H1') \<and> 
                (\<exists> \<sigma>' k1'.  cmd_steps \<sigma>' (Run c1 (set \<Gamma>1) s1' H1') k1' \<and> sched_eq \<sigma> \<sigma>' \<and>
                         (ct_eq lev \<sigma> \<sigma>'))}"

lemma uncertainty_uncertainty2_subset:
  assumes "cmd_steps \<sigma> (Run c1 (set \<Gamma>1) s1 H1) k1"
  shows "uncertainty sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1 \<subseteq> uncertainty2 sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1"
  using assms by(clarsimp simp: uncertainty_def uncertainty2_def, blast)

definition allowed_release2 :: "'level \<Rightarrow> ('level,'a) sched \<Rightarrow> cmd \<Rightarrow> ctx \<Rightarrow> 'a state \<Rightarrow> 'a heap \<Rightarrow> ('level,'a) sched \<Rightarrow> ('a state \<times> 'a heap) set" where
  "allowed_release2 lev \<sigma>1 c1 \<Gamma>1 s1 H1 \<sigma>2 \<equiv> 
      {(s1',H1'). \<exists>\<sigma>1' k2'. cmd_steps \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k2' \<and>
                        sched_eq \<sigma>1 \<sigma>1'  \<and> (ct_eq lev \<sigma>1 \<sigma>1') \<and>
                        (\<exists>\<sigma>1a \<sigma>1b \<sigma>1a' \<sigma>1b' k1a k1a' r. \<sigma>1@\<sigma>2 = \<sigma>1a @ \<sigma>1b \<and> \<sigma>1' = \<sigma>1a' @ \<sigma>1b' \<and>
                           length \<sigma>1a = length \<sigma>1a' \<and>
                           cmd_steps \<sigma>1a (Run c1 (set \<Gamma>1) s1 H1) k1a \<and>
                           cmd_steps \<sigma>1a' (Run c1 (set \<Gamma>1) s1' H1') k1a' \<and> 
                           sched_asm \<sigma>1b = Some r \<and>
                           \<not> (rel (rexpr_sem lev r) (stack k1a, Map.empty) (stack k1a', Map.empty)))
                         }" 



definition controlled_leakage2 ::  "bool \<Rightarrow> 'level \<Rightarrow> assert \<Rightarrow> cmd \<Rightarrow> bool"
  where
  "controlled_leakage2 sh lev p1 c1 \<equiv>
    \<forall> \<Gamma>1 \<sigma>1 s1 H1 k2 \<sigma>2 k2'.  set \<Gamma>1 \<subseteq> locks \<and>
       cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2 \<and>
       cmd_step \<sigma>2 k2 k2' \<longrightarrow>
        (uncertainty2 sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1 -
         uncertainty2 sh lev p1 (\<sigma>1@\<sigma>2) c1 \<Gamma>1 s1 H1) 
        \<subseteq> allowed_release2 lev \<sigma>1 c1 \<Gamma>1 s1 H1 \<sigma>2"


lemma asm_extend:
  "asm lev \<sigma>1 s1 s1' \<Longrightarrow> asm lev (\<sigma>1@\<sigma>2) s1 s1'"
  apply(induct \<sigma>1)
   apply(simp add: asm.simps)
  using asm.simps sched_asm_extend 
  by meson



lemma sched_eq_appD2:
  "length \<sigma>1 = length \<sigma>1' \<Longrightarrow> sched_eq (\<sigma>1@\<sigma>2) (\<sigma>1'@\<sigma>2') \<Longrightarrow> 
   sched_eq \<sigma>1 \<sigma>1' \<and> sched_eq \<sigma>2 \<sigma>2'"
  apply(induct \<sigma>1 \<sigma>1' arbitrary: \<sigma>2 \<sigma>2' rule: list_induct2)
  by auto

lemma ct_eq_length_eq[simp,intro]:
  "ct_eq lev \<sigma>1 \<sigma>2 \<Longrightarrow> length \<sigma>1 = length \<sigma>2"
  by (meson ct_eq_def map_eq_imp_length_eq)

definition
  ctobs :: "'level \<Rightarrow> ('level,'a) tick \<Rightarrow> ('level,'a) tick" where
  "ctobs lev = (\<lambda>ev. case ev of Out l x \<Rightarrow> if l\<le>lev then Out l x else Out l undefined | _ \<Rightarrow> ev)"

lemma ct_ct_eq_ConsD:
  assumes ct: "ct"
  shows
  "ct_eq lev (x # xs) (y # ys) \<Longrightarrow> ct_eq lev [x] [y] \<and> ct_eq lev xs ys"
  apply(simp add: ct_eq_def)
  apply(simp add: ct)
  apply(cases x; cases y, auto simp: outs_def)
  done

lemma ct_ct_eq_Single_ctobs:
  assumes ct: "ct"
  shows "ct_eq lev [x] [y] \<Longrightarrow> ctobs lev x = ctobs lev y"
  apply(simp add: ct_eq_def, simp add: ct ctobs_def)
  by(cases x; cases y, auto simp: outs_def)


lemma ct_ctobs_ct_eq:
  assumes ct: "ct"
  assumes "map (ctobs lev) xs = map (ctobs lev) ys"
  shows
  "ct_eq lev (x # xs) (y # ys)
       \<Longrightarrow> map (ctobs lev) (x # xs) = map (ctobs lev) (y # ys)"
  by (metis (no_types, lifting) assms(2) concurrency.ct_ct_eq_ConsD concurrency_axioms ct ct_ct_eq_Single_ctobs list.simps(9))

lemma ct_eq_ctobs_ct_eq_ConsI:
  "ct_eq lev xs ys \<Longrightarrow> ctobs lev x = ctobs lev y \<Longrightarrow>
   ct_eq lev (x#xs) (y#ys)"
  by(cases x; cases y, auto simp: ct_eq_def ctobs_def outs_def split: if_splits)

lemma ctobs_ct_eq:
  assumes "map (ctobs lev) xs = map (ctobs lev) ys"
  shows "ct_eq lev xs ys"
proof -
  from `map (ctobs lev) xs = map (ctobs lev) ys` have
  "length xs = length ys" 
    using map_eq_imp_length_eq by blast
  from this `map (ctobs lev) xs = map (ctobs lev) ys` show ?thesis
    apply(induct rule: list_induct2)
     apply simp
    by (simp add: ct_eq_ctobs_ct_eq_ConsI)
qed

  
lemma ct_ct_eq_def2:
  assumes ct: "ct"
  shows "(ct_eq lev \<sigma>1 \<sigma>2) = 
  ((map (ctobs lev) \<sigma>1) = (map (ctobs lev) \<sigma>2))"
proof
  assume "ct_eq lev \<sigma>1 \<sigma>2"
  hence "length \<sigma>1 = length \<sigma>2" by simp
  from this `ct_eq lev \<sigma>1 \<sigma>2` show "(map (ctobs lev) \<sigma>1 = map (ctobs lev) \<sigma>2)"
    apply(induct rule: list_induct2)
     apply simp
    using ct ct_ct_eq_ConsD ct_ctobs_ct_eq by blast
next
  assume "map (ctobs lev) \<sigma>1 = map (ctobs lev) \<sigma>2 "
  thus "ct_eq lev \<sigma>1 \<sigma>2"
    using ctobs_ct_eq by blast
qed

lemma ct_ct_eq_appD2:
  assumes ct: "ct"
  shows "length \<sigma>1 = length \<sigma>1' \<Longrightarrow> ct_eq lev (\<sigma>1@\<sigma>2) (\<sigma>1'@\<sigma>2') \<Longrightarrow> 
   ct_eq lev \<sigma>1 \<sigma>1' \<and> ct_eq lev \<sigma>2 \<sigma>2'"
  using ct_ct_eq_def2[OF ct]
  by (metis (no_types, lifting) append_eq_append_conv length_map map_append)

lemma ct_ct_eq_appD:
  "ct \<Longrightarrow> ct_eq lev (\<sigma>1 @ \<sigma>2) \<sigma>' \<Longrightarrow> \<exists>\<sigma>1' \<sigma>2'. \<sigma>' = \<sigma>1'@\<sigma>2' \<and> ct_eq lev \<sigma>1 \<sigma>1' \<and> ct_eq lev \<sigma>2 \<sigma>2'"
  apply(rule_tac x="take (length \<sigma>1) \<sigma>'" in exI)
  apply(rule_tac x="drop (length \<sigma>1) \<sigma>'" in exI)
  by (metis (no_types, lifting) append_eq_conv_conj append_take_drop_id ct_ct_eq_def2 map_append take_map)

fun
  atomic_action where
  "atomic_action Fst = False" |
  "atomic_action Snd = False" |
  "atomic_action _ = True"

fun
  valid_action where
  "valid_action (Fst#as) = valid_action as" |
  "valid_action (Snd#as) = valid_action as" |
  "valid_action (x#xs) = (atomic_action x \<and> xs = [])" |
  "valid_action [] = False"

lemma cmd_step_valid_action[intro]:
  "cmd_step \<sigma> k k1 \<Longrightarrow> valid_action \<sigma>"
  by(induct rule: cmd_step.induct, auto)

inductive valid_actions where
  "valid_actions []" |
  "valid_action x \<Longrightarrow> valid_actions xs \<Longrightarrow> valid_actions (x@xs)"

lemma cmd_steps_valid_actions:
  "cmd_steps \<sigma> k k1 \<Longrightarrow> valid_actions \<sigma>"
  by(induct rule: cmd_steps.induct, auto intro: valid_actions.intros)

lemma
  "valid_action \<sigma>1 \<Longrightarrow> \<sigma>1 = [] \<Longrightarrow> False"
  by(cases \<sigma>1, auto)

lemma valid_action_does_the_thing:
  "cmd_step \<sigma> k1 k' \<Longrightarrow> valid_action \<sigma>1 \<Longrightarrow> \<sigma>1 @ \<sigma>2 = \<sigma> @ \<sigma>' \<Longrightarrow> 
   \<sigma>1 = \<sigma>"
proof(induct arbitrary: \<sigma>1 \<sigma>2 \<sigma>' rule: cmd_step.induct)
case (1 L s h)
  then show ?case by(cases \<sigma>1, auto)
next
  case (2 v e s x L h)
  then show ?case by(cases \<sigma>1, auto)
next
  case (3 a p s h x L)
  then show ?case by(cases \<sigma>1, auto)
next
case (4 a p s h v x L)
  then show ?case by(cases \<sigma>1, auto)
next
  case (5 a p s h e L)
  then show ?case by(cases \<sigma>1, auto)
next
  case (6 a p s h v v' e L)
  then show ?case by(cases \<sigma>1, auto)
next
  case (7 l L s h)
  then show ?case by(cases \<sigma>1, auto)
next
  case (8 l L s h)
  then show ?case by(cases \<sigma>1, auto)
next
  case (9 \<sigma> c1 L s h c2)
  then show ?case by(cases \<sigma>1, auto)
next
  case (10 \<sigma> c1 L s h L' s' h' c2)
  then show ?case by(cases \<sigma>1, auto)
next
  case (11 \<sigma> c1 L s h c1' L' s' h' c2)
  then show ?case by(cases \<sigma>1, auto)
next
  case (12 \<sigma> c1 L s h c2)
  then show ?case by(cases \<sigma>1, auto)
next
  case (13 \<sigma> c1 L s h L' s' h' c2)
  then show ?case by(cases \<sigma>1, auto)
next
  case (14 \<sigma> c1 L s h c1' L' s' h' c2)
  then show ?case by(cases \<sigma>1, auto)
next
  case (15 \<sigma> c2 L s h c1)
  then show ?case by(cases \<sigma>1, auto)
next
  case (16 \<sigma> c2 L s h L' s' h' c1)
  then show ?case by(cases \<sigma>1, auto)
next
  case (17 \<sigma> c2 L s h c2' L' s' h' c1)
then show ?case by(cases \<sigma>1, auto)
next
case (18 b s c1 c2 L h)
then show ?case by(cases \<sigma>1, auto)
next
  case (19 b s c1 c2 L h)
  then show ?case by(cases \<sigma>1, auto)
next
  case (20 b s c L h)
  then show ?case by(cases \<sigma>1, auto)
next
  case (21 b s c L h)
  then show ?case by(cases \<sigma>1, auto)
next
  case (22 p L s h)
  then show ?case by(cases \<sigma>1, auto)
next
  case (23 elev s l eval v L h)
  then show ?case by(cases \<sigma>1, auto)
qed
  
lemma cmd_steps_valid_action_cmd_step:
  "cmd_steps (\<sigma>1@\<sigma>2) k1 k3 \<Longrightarrow>
   valid_action \<sigma>1 \<Longrightarrow>
   \<exists>k2. cmd_step \<sigma>1 k1 k2"
  apply(erule cmd_steps.cases)
   apply(cases \<sigma>1, auto)[1]
  apply clarsimp
  using valid_action_does_the_thing 
  by blast
  
lemma cmd_steps_valid_actions_cmd_steps:
  "cmd_steps (\<sigma>1@\<sigma>2) k1 k3 \<Longrightarrow>
   valid_actions \<sigma>1 \<Longrightarrow>
   \<exists>k2. cmd_steps \<sigma>1 k1 k2"
proof(induct "\<sigma>1@\<sigma>2" k1 k3 arbitrary: \<sigma>1 \<sigma>2 rule: cmd_steps.induct)
  case (cmd_steps_refl k)
  then show ?case 
    by blast
next
  case (cmd_steps_step \<sigma>01 k0 k1 \<sigma>12 k2)
  from this(5) this(1-4)
  show ?case
  proof(induct rule: valid_actions.induct)
    case 1
    then show ?case by blast
  next
    case (2 x xs)
    then show ?case
      by (metis append.assoc append_eq_append_conv cmd_steps.cmd_steps_step valid_action_does_the_thing)
  qed
qed

lemma valid_action_ctobs:
  "valid_action \<sigma>1 \<Longrightarrow> map (ctobs lev) \<sigma>1 = map (ctobs lev) \<sigma>1' \<Longrightarrow> valid_action \<sigma>1'"
  by(induct \<sigma>1 arbitrary: \<sigma>1' rule: valid_action.induct, auto simp: ctobs_def split: tick.splits if_splits)

lemma ct_valid_actions_ct_eq:
  assumes ct: "ct"
  shows "valid_actions \<sigma>1 \<Longrightarrow> ct_eq lev \<sigma>1 \<sigma>1' \<Longrightarrow> valid_actions \<sigma>1'"
  apply(induct \<sigma>1 arbitrary: \<sigma>1' rule: valid_actions.induct)
  using ct ct_ct_eq_def2 valid_actions.simps apply auto[1]
  apply(drule ct_ct_eq_appD[OF ct])
  apply clarsimp
  using valid_action_ctobs 
  by (meson ct ct_ct_eq_def2 valid_actions.intros(2))  
  

lemma prefix_closed_thingy:
  assumes  ct: "ct"
  shows 
  "cmd_steps \<sigma>1 (Run c L s1 H1) k2 \<Longrightarrow>
   cmd_steps (\<sigma>1'@\<sigma>2') (Run c L s1' H1') k3' \<Longrightarrow>
   ct_eq lev \<sigma>1 \<sigma>1' \<Longrightarrow>
   \<exists>k2'. cmd_steps \<sigma>1' (Run c L s1' H1') k2'"
  using cmd_steps_valid_actions_cmd_steps cmd_steps_valid_actions ct_valid_actions_ct_eq ct by metis

lemma controlled_leakage2:
  assumes ct: "ct"
  shows "controlled_leakage sh lev p1 c1 \<Longrightarrow> controlled_leakage2 sh lev p1 c1"
  apply(clarsimp simp: controlled_leakage2_def)
  apply(rename_tac s1' H1' k3)
proof -
  fix \<Gamma>1 \<sigma>1 s1 H1 k2 \<sigma>2 s1' H1' k3
  assume "controlled_leakage sh lev p1 c1"
  assume "(s1', H1') \<in> uncertainty2 sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1"
  assume "(s1', H1') \<notin> uncertainty2 sh lev p1 (\<sigma>1 @ \<sigma>2) c1 \<Gamma>1 s1 H1"
  assume "set \<Gamma>1 \<subseteq> locks"
  assume "cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2"
  assume "cmd_step \<sigma>2 k2 k3"
  from `(s1', H1') \<in> uncertainty2 sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1`  obtain \<sigma>1' k1' where
    inv: "local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1', H1')" and
    steps1: "cmd_steps \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k1'" and
    sched_eq1: "sched_eq \<sigma>1 \<sigma>1'" and ct_eq1: "ct_eq lev \<sigma>1 \<sigma>1'"
    by(clarsimp simp: uncertainty2_def)
  from `(s1', H1') \<notin> uncertainty2 sh lev p1 (\<sigma>1 @ \<sigma>2) c1 \<Gamma>1 s1 H1` have 
    nomatch2: "\<And>\<sigma>'. sched_eq (\<sigma>1 @ \<sigma>2) \<sigma>' \<Longrightarrow>
                (\<forall>k'. \<not> cmd_steps \<sigma>' (Run c1 (set \<Gamma>1) s1' H1') k') \<or> \<not> ct_eq lev (\<sigma>1 @ \<sigma>2) \<sigma>'"
    using inv
    by(clarsimp simp: uncertainty2_def)
  show "(s1', H1') \<in> allowed_release2 lev \<sigma>1 c1 \<Gamma>1 s1 H1 \<sigma>2"
  proof(cases "(\<exists>\<sigma>1a \<sigma>1b \<sigma>1a' \<sigma>1b' k1a k1a' r. \<sigma>1@\<sigma>2 = \<sigma>1a @ \<sigma>1b \<and> \<sigma>1' = \<sigma>1a' @ \<sigma>1b' \<and>
                           length \<sigma>1a = length \<sigma>1a' \<and>
                           cmd_steps \<sigma>1a (Run c1 (set \<Gamma>1) s1 H1) k1a \<and>
                           cmd_steps \<sigma>1a' (Run c1 (set \<Gamma>1) s1' H1') k1a' \<and> 
                           sched_asm \<sigma>1b = Some r \<and>
                           \<not> (rel (rexpr_sem lev r) (stack k1a, Map.empty) (stack k1a', Map.empty)))") 
    case True 
    show ?thesis 
      apply(simp add: allowed_release2_def)
      apply(rule_tac x=\<sigma>1' in exI)
      apply(rule conjI, rule exI, rule steps1)
      apply(rule conjI, rule sched_eq1)
      apply(rule conjI, rule ct_eq1)
      using True
      by simp
  next
    case False
    have "(s1', H1') \<in> uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1"
    proof(rule ccontr)
      assume nin: "(s1', H1') \<notin> uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1"
      from nin inv obtain \<sigma>1a \<sigma>1b k1a where
        \<sigma>1_def: "\<sigma>1 = \<sigma>1a @ \<sigma>1b" and
        steps1a: "cmd_steps \<sigma>1a (Run c1 (set \<Gamma>1) s1 H1) k1a"  and
        blah: "\<And> \<sigma>1a' k1a'. ct_eq lev \<sigma>1a \<sigma>1a' \<Longrightarrow>
                     sched_eq \<sigma>1a \<sigma>1a' \<Longrightarrow>
                     cmd_steps \<sigma>1a' (Run c1 (set \<Gamma>1) s1' H1') k1a' \<Longrightarrow>
             (\<exists>k2 \<sigma>3.
                                cmd_step \<sigma>3 k1a k2 \<and>
                                (\<exists>\<sigma>4. \<sigma>1b = \<sigma>3 @ \<sigma>4) \<and> asm lev \<sigma>3 (stack k1a) (stack k1a'))"        
        by(clarsimp simp: uncertainty_def)
      let ?\<sigma>1a' = "take (length \<sigma>1a) \<sigma>1'"
      let ?\<sigma>1b' = "drop (length \<sigma>1a) \<sigma>1'"
      have "length \<sigma>1' = length \<sigma>1" using ct_eq1 
        by (metis (no_types, lifting) ct_eq_def length_map)
      with \<sigma>1_def have len1a: "length \<sigma>1a = length ?\<sigma>1a'"
        by simp
      have sched_eq1a: "sched_eq \<sigma>1a ?\<sigma>1a'" 
        using sched_eq_appD2 sched_eq1 len1a 
        by (metis \<sigma>1_def append_take_drop_id)
      have ct_eq1a: "ct_eq lev \<sigma>1a ?\<sigma>1a'" 
        using ct ct_ct_eq_appD2 
        by (metis \<sigma>1_def append_take_drop_id ct_eq1 len1a)
      from steps1 steps1a sched_eq1a ct_eq1a obtain k1a' where
        steps1a': "cmd_steps ?\<sigma>1a' (Run c1 (set \<Gamma>1) s1' H1') k1a'"         
        using prefix_closed_thingy 
        by (metis append_take_drop_id ct)
      obtain \<sigma>1ba \<sigma>1bb k1ba where "cmd_step \<sigma>1ba k1a k1ba \<and>
                               \<sigma>1b = \<sigma>1ba @ \<sigma>1bb \<and> asm lev \<sigma>1ba (stack k1a) (stack k1a')"
        using blah steps1a' sched_eq1a ct_eq1a by blast
      hence "asm lev \<sigma>1b (stack k1a) (stack k1a')" using asm_extend by blast
      from this obtain r where " sched_asm \<sigma>1b = Some r \<and>
       \<not> rel (rexpr_sem lev r) (stack k1a, Map.empty) (stack k1a', Map.empty)" 
        using asm.cases by force
      moreover have "sched_asm (\<sigma>1b@\<sigma>2) = Some r" using sched_asm_extend
        by (simp add: calculation)
      moreover have "\<sigma>1' = ?\<sigma>1a' @ ?\<sigma>1b'" by simp
      moreover have "\<sigma>1@\<sigma>2 = \<sigma>1a @ (\<sigma>1b@\<sigma>2)" using \<sigma>1_def by simp 
      ultimately   show "False" using steps1a steps1a' False len1a by metis
    qed
    from nomatch2 have "(s1', H1') \<notin> uncertainty sh lev p1 (\<sigma>1 @ \<sigma>2) c1 \<Gamma>1 s1 H1"
      apply(clarsimp simp: uncertainty_def)
      by (meson \<open>cmd_step \<sigma>2 k2 k3\<close> \<open>cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2\<close> cmd_step1_intro concurrency.cmd_steps_trans concurrency_axioms self_append_conv)
    from `controlled_leakage sh lev p1 c1` have
      "(s1', H1') \<in> allowed_release lev \<sigma>1 c1 \<Gamma>1 (stack k2) \<sigma>2"
      using \<open>(s1', H1') \<in> uncertainty sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1\<close> \<open>(s1', H1') \<notin> uncertainty sh lev p1 (\<sigma>1 @ \<sigma>2) c1 \<Gamma>1 s1 H1\<close> \<open>cmd_step \<sigma>2 k2 k3\<close> \<open>cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2\<close> \<open>set \<Gamma>1 \<subseteq> locks\<close> by blast
    from this obtain \<sigma>1'' k1'' where stuff:
      "cmd_steps \<sigma>1'' (Run c1 (set \<Gamma>1) s1' H1') k1'' \<and>
       sched_eq \<sigma>1 \<sigma>1'' \<and> ct_eq lev \<sigma>1 \<sigma>1'' \<and> asm lev \<sigma>2 (stack k2) (stack k1'')"
      by(auto simp: allowed_release_def)
    thus ?thesis
      apply(clarsimp simp: allowed_release2_def)
      apply(rule_tac x=\<sigma>1'' in exI)
      apply(rule conjI)
      using stuff apply auto[1]
      apply(rule conjI, simp add: stuff)
      apply(rule conjI, simp add: stuff)
      apply(rule_tac x=\<sigma>1 in exI, rule_tac x=\<sigma>2 in exI)
      apply(rule conjI, simp)
      apply(rule_tac x=\<sigma>1'' in exI)
      apply(rule conjI, rule_tac x="[]" in exI, simp)
      apply(rule conjI)
      using stuff ct_eq_def 
       apply (meson map_eq_imp_length_eq)
      apply(rule_tac x=k2 in exI)
      apply(rule conjI)
       apply(rule `cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2`)
      using asm.cases by fastforce
  qed
qed

(*
 Wrap up into a single definition what it means when, for a major execution
@{term "\<sigma>1@\<sigma>2"} from initial state @{term "(s1,H1)"} and minor execution
@{term "\<sigma>1'"} from initial state @{term "(s1',H1')"} executing command~@{term c1}
with initial locks @{term \<Gamma>1} to have a violated assumption somewhere in the
execution.

Essentially given  @{term "\<sigma>1          @ \<sigma>2"}
            and    @{term "\<sigma>1'         "}
we are looking for @{term "\<sigma>1a         @ \<sigma>1b"}
            and    @{term "\<sigma>1a' @ \<sigma>1b' "}
where after the equal length @{term \<sigma>1a} and @{term \<sigma>1a'} the next step
(@{term \<sigma>1b}) is an assumption that doesn't hold.

Note that the assumption can be the step @{term \<sigma>2} or an earlier one.
*)
definition
  has_violated_assumption where
  "has_violated_assumption lev c1 \<Gamma>1 s1 H1 s1' H1' \<sigma> \<sigma>1' \<equiv>
                        (\<exists>\<sigma>1a \<sigma>1b \<sigma>1a' \<sigma>1b' k1a k1a' r. \<sigma> = \<sigma>1a @ \<sigma>1b \<and> \<sigma>1' = \<sigma>1a' @ \<sigma>1b' \<and>
                           length \<sigma>1a = length \<sigma>1a' \<and>
                           cmd_steps \<sigma>1a (Run c1 (set \<Gamma>1) s1 H1) k1a \<and>
                           cmd_steps \<sigma>1a' (Run c1 (set \<Gamma>1) s1' H1') k1a' \<and> 
                           sched_asm \<sigma>1b = Some r \<and>
                           \<not> (rel (rexpr_sem lev r) (stack k1a, Map.empty) (stack k1a', Map.empty)))"

lemma allowed_release2_def2:
  "allowed_release2 lev \<sigma>1 c1 \<Gamma>1 s1 H1 \<sigma>2 =
      {(s1',H1'). \<exists>\<sigma>1' k2'. cmd_steps \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k2' \<and>
                        sched_eq \<sigma>1 \<sigma>1'  \<and> (ct_eq lev \<sigma>1 \<sigma>1') \<and>
                        has_violated_assumption lev c1 \<Gamma>1 s1 H1 s1' H1' (\<sigma>1@\<sigma>2) \<sigma>1'}"
  by(simp add: allowed_release2_def has_violated_assumption_def)

definition
  exclusion_knowledge2 :: "bool \<Rightarrow> 'level \<Rightarrow> assert \<Rightarrow> ('level,'a) sched \<Rightarrow> cmd \<Rightarrow> ctx \<Rightarrow> 'a state \<Rightarrow> 'a heap \<Rightarrow>  ('a state \<times> 'a heap) set"
  where
  "exclusion_knowledge2 sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1 \<equiv> UNIV - uncertainty2 sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1"


(*
  Prove the definition above equivalent to the more ``standard'' one phrased in terms of
  exclusion knowledge
*)
lemma controlled_leakage2_def2:
  "controlled_leakage2 sh lev p1 c1 =
 (\<forall> \<Gamma>1 \<sigma>1 s1 H1 k2 \<sigma>2 k2'.  set \<Gamma>1 \<subseteq> locks \<and>
       cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2 \<and>
       cmd_step \<sigma>2 k2 k2' \<longrightarrow>
        (exclusion_knowledge2 sh lev p1 (\<sigma>1@\<sigma>2) c1 \<Gamma>1 s1 H1 -
          exclusion_knowledge2 sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1) 
        \<subseteq> allowed_release2 lev \<sigma>1 c1 \<Gamma>1 s1 H1 \<sigma>2)"
  apply(simp add: exclusion_knowledge2_def controlled_leakage2_def)
  apply(rule iff_allI)+
  apply(rule imp_cong, rule refl)
  apply(rule subset_cong_blah)
  by blast

(*
  Sanity check the new uncertainty definition too. 
*)
lemma uncertainty2_only_decreases:
  notes if_splits[split del]
  assumes ct: "ct"
  assumes "cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k1"
  shows "uncertainty2 sh lev p1 (\<sigma>1@\<sigma>2) c1 \<Gamma>1 s1 H1 \<subseteq> 
        uncertainty2 sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1"
  (* FIXME: horrible proof *)
  apply(subst (2) uncertainty2_def)
  apply(clarsimp split del: if_splits | safe)+
   apply(auto simp: uncertainty2_def)[1]
  apply(subst (asm) uncertainty2_def)
  apply clarsimp
  apply(drule ct_ct_eq_appD[OF ct])
  apply clarsimp
  by (meson assms(2) concurrency.sched_eq_appD2 concurrency_axioms ct ct_eq_length_eq prefix_closed_thingy)

lemma uncertainty2_uncertainty_subset:
  assumes ct: "ct"
  assumes "cmd_steps \<sigma> (Run c1 (set \<Gamma>1) s1 H1) k1"
  assumes "no_asm \<sigma>"
  shows "uncertainty2 sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1 \<subseteq> uncertainty sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1"
  apply(clarsimp simp: uncertainty2_def)
  apply(rename_tac s1' H1' \<sigma>' k')
  apply(clarsimp simp: uncertainty_def)
  apply(drule ct_ct_eq_appD[OF ct])
  apply clarsimp
  by (metis asm_no_asm assms(3) concurrency.sched_eq_appD2 concurrency_axioms ct ct_eq_length_eq no_asm_appD prefix_closed_thingy)

lemma ct_uncertainty_uncertainty2_equiv:
  assumes ct: "ct"
  assumes "cmd_steps \<sigma> (Run c1 (set \<Gamma>1) s1 H1) k1"
  assumes "no_asm \<sigma>"
  shows "uncertainty2 sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1 = uncertainty sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1"
  using uncertainty_uncertainty2_subset uncertainty2_uncertainty_subset assms by blast

lemma learns_nothing_def2:
  assumes ct: "ct"
  shows
  "learns_nothing sh lev p1 c1 =
    (\<forall> \<Gamma>1 \<sigma> s1 H1 k2.  no_asm \<sigma> \<and> set \<Gamma>1 \<subseteq> locks \<and>
       cmd_steps \<sigma> (Run c1 (set \<Gamma>1) s1 H1) k2 \<longrightarrow>
        uncertainty2 sh lev p1 \<sigma> c1 \<Gamma>1 s1 H1 \<supseteq> 
           {(s1',H1'). local.inv sh lev \<Gamma>1 p1 Emp (s1, H1) (s1',H1')})"
  using learns_nothing_def ct_uncertainty_uncertainty2_equiv[OF ct] 
  by presburger

lemma no_leakage_def2:
  assumes ct: "ct"
  shows "no_leakage sh lev p1 c1 =
(\<forall>\<Gamma>1 \<sigma>1 s1 H1 k3 k2 \<sigma>2.
   set \<Gamma>1 \<subseteq> locks \<and>
   cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2 \<and> cmd_steps \<sigma>2 k2 k3 \<and> no_asm (\<sigma>1 @ \<sigma>2) \<longrightarrow>
   uncertainty2 sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1 \<subseteq> uncertainty2 sh lev p1 (\<sigma>1 @ \<sigma>2) c1 \<Gamma>1 s1 H1)"
  apply(safe | auto simp: no_leakage_def)+
  using  ct_uncertainty_uncertainty2_equiv[OF ct] cmd_steps_trans no_asm_appD
   apply (smt subsetCE)
  using  ct_uncertainty_uncertainty2_equiv[OF ct] cmd_steps_trans no_asm_appD
  by (smt subsetCE)

lemma controlled_leakage2_no_leakage':
  "\<lbrakk>cmd_steps \<sigma>2 k2 k3; cmd_steps \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k2; controlled_leakage2 sh lev p1 c1; 
    (s1', H1') \<in> uncertainty2 sh lev p1 \<sigma>1 c1 \<Gamma>1 s1 H1;
        set \<Gamma>1 \<subseteq> locks; no_asm (\<sigma>1 @ \<sigma>2)\<rbrakk>
       \<Longrightarrow> (s1', H1') \<in> uncertainty2 sh lev p1 (\<sigma>1 @ \<sigma>2) c1 \<Gamma>1 s1 H1"
  unfolding cmd_steps_rev
proof(induction "\<sigma>2" "k2" k3 arbitrary: \<sigma>1 rule: cmd_steps_rev.induct)
  case cmd_steps_rev_refl
  then show ?case 
    by auto
next
  case (cmd_steps_rev_step \<sigma> k1 k2 \<sigma>s k3)
  have use_IH: "(s1', H1') \<in> uncertainty2 sh lev p1 (\<sigma>1 @ \<sigma>) c1 \<Gamma>1 s1 H1"
    by (metis append_assoc cmd_steps_rev_step.IH cmd_steps_rev_step.prems(1) cmd_steps_rev_step.prems(2) cmd_steps_rev_step.prems(3) cmd_steps_rev_step.prems(4) cmd_steps_rev_step.prems(5) no_asm_appD)  
  have "\<And>lev. allowed_release2 lev (\<sigma>1 @ \<sigma>) c1 \<Gamma>1 s1 H1 \<sigma>s = {}"
    apply(clarsimp simp: allowed_release2_def elim: asm.cases)
    using cmd_steps_rev_step.prems(5) no_asm_appD sched_asm_no_asm by blast
  moreover from 
    `cmd_steps_rev \<sigma>1 (Run c1 (set \<Gamma>1) s1 H1) k1`
    `cmd_steps_rev \<sigma> k1 k2` 
  have "cmd_steps (\<sigma>1@\<sigma>) (Run c1 (set \<Gamma>1) s1 H1) k2" using cmd_steps_trans
    using cmd_steps_rev by auto
  ultimately show ?case
    using `controlled_leakage2 sh lev p1 c1`
    apply -
    apply(clarsimp simp: controlled_leakage2_def)
    by (metis DiffI append.assoc cmd_steps_rev_step.hyps(2) cmd_steps_rev_step.prems(4) empty_iff subsetI subset_antisym use_IH)
qed

lemma controlled_leakage2_no_leakage:
  assumes ct: "ct"
  shows "controlled_leakage2 sh lev p1 c1 \<Longrightarrow> no_leakage sh lev p1 c1"
  by(clarsimp simp: no_leakage_def2[OF ct] controlled_leakage2_no_leakage')
  
lemma controlled_leakage2_learns_nothing:
  assumes ct: "ct"
  shows "controlled_leakage2 sh lev p1 c1 \<Longrightarrow> learns_nothing sh lev p1 c1"
  by(blast intro: controlled_leakage2_no_leakage[OF ct] no_leakage_learns_nothing)

(* Note that the @{term ct} argument here applies the constant-time security definition. *)
lemma controlled_release2:
  "ct \<Longrightarrow> cmd_mod c1 \<inter> invvars = {} \<Longrightarrow> prove sh lev p1 c1 q1 \<Longrightarrow> controlled_leakage2 sh lev p1 c1"
  using controlled_release controlled_leakage2 by metis


end

end
