theory Locks
  imports Logic
begin

context concurrency
begin

lemma remove_lock:
"l \<in> set \<Gamma> \<Longrightarrow> \<exists> \<Gamma>'. set \<Gamma> - {l} = set \<Gamma>'"
proof -
  have "set \<Gamma> - {l} = set (filter (\<lambda> x. x \<noteq> l) \<Gamma>)"
    by auto
  then show "\<exists> \<Gamma>'. set \<Gamma> - {l} = set \<Gamma>'" ..
qed

lemma
lockinvs_Nil[simp]:   "lockinvs [] = Emp" and
lockinvs_Cons[simp]:  "l \<notin> set \<Gamma> \<Longrightarrow> lockinvs (l#\<Gamma>) = Star (lockinv l) (lockinvs \<Gamma>)" and
lockinvs_Cons'[simp]: "l \<in> set \<Gamma> \<Longrightarrow> lockinvs (l#\<Gamma>) = lockinvs \<Gamma>"
  unfolding lockinvs_def by auto

lemma lockinvs_sem_reorder:
"\<lbrakk>set \<Gamma> = set \<Gamma>'\<rbrakk>
  \<Longrightarrow> assert_sem lev (lockinvs \<Gamma>) = assert_sem lev (lockinvs \<Gamma>')"
  unfolding lockinvs_def by (auto intro: stars_reorder_perm perm_map_remdups)


(* use the four corollaries below *)
corollary lockinvs_sem_lock:
"\<lbrakk>l \<in> set \<Gamma>; set \<Gamma>' = set \<Gamma> - {l}\<rbrakk>
  \<Longrightarrow> assert_sem lev (lockinvs \<Gamma>) = star (assert_sem lev (lockinv l)) (assert_sem lev (lockinvs \<Gamma>'))"
  using lockinvs_sem_reorder[of \<Gamma> "l#\<Gamma>'" lev] by auto

corollary lockinvs_sem_unlock:
"\<lbrakk>l \<notin> set \<Gamma>; set \<Gamma>' = insert l (set \<Gamma>)\<rbrakk>
  \<Longrightarrow> assert_sem lev (lockinvs \<Gamma>') = star (assert_sem lev (lockinv l)) (assert_sem lev (lockinvs \<Gamma>))"
  using lockinvs_sem_reorder[of  \<Gamma>' "l#\<Gamma>" lev] by auto

lemma \<Gamma>_eq[intro!]:     "P \<Gamma> \<Longrightarrow> \<exists>\<Gamma>'. set \<Gamma>' = set \<Gamma> \<and> P \<Gamma>'" by auto
lemma \<Gamma>_insert[intro!]: "P (l # \<Gamma>) \<Longrightarrow> \<exists>\<Gamma>'. set \<Gamma>' = insert l (set \<Gamma>) \<and> P \<Gamma>'" by (drule \<Gamma>_eq, simp)

lemma lockinv_free[intro]:
"l \<in> locks \<Longrightarrow> assert_free (lockinv l) \<subseteq> invvars"
  unfolding invvars_def invs_def by auto


(* wrt loc *)
lemma locks_sem_coincidence[simp]:
"\<lbrakk>x \<notin> invvars; set \<Gamma> \<subseteq> locks\<rbrakk>
  \<Longrightarrow> assert_sem lev (lockinvs \<Gamma>) (s(x:=v),h) (s'(x:=v'),h') = assert_sem lev (lockinvs \<Gamma>) (s,h) (s',h')"
proof (induction \<Gamma> arbitrary: h h', simp) (* note: heap will be smaller! *)
  case (Cons l \<Gamma>)
    from Cons(2) Cons(3) have
      l: "\<And> h h'. assert_sem lev (lockinv l) (s(x:=v),h) (s'(x:=v'),h') = assert_sem lev (lockinv l) (s,h) (s',h')"
      unfolding invvars_def invs_def by auto
  show ?case proof (cases "l \<in> set \<Gamma>")
    case True
    with Cons show ?thesis by simp
  next
    case False
    with l Cons show ?thesis
      by (auto simp del: fun_upd_apply intro!: star_coincidence)
  qed
qed


end

end