theory Logic
  imports Separation
begin

type_synonym ctx = "lsym list"

locale concurrency = semantics _ _ _ _ _ to_level
  for to_level :: "'a \<Rightarrow> ('level::complete_lattice)" +
  fixes lockinv :: "lsym \<Rightarrow> assert"
  fixes shared :: assert
  fixes ct     :: bool    (* whether we are enforcing constant-time *)

  fixes locks :: locks
  assumes finite_locks[intro]: "finite locks"

context concurrency
begin

definition invs :: "assert set" where
"invs \<equiv> (insert shared {lockinv l | l. l \<in> locks})"

definition invvars :: "var set" where
"invvars = \<Union>(assert_free ` invs)"

definition subset :: "('a state \<Rightarrow> 'a set) \<Rightarrow> ('a state \<Rightarrow> 'a  set) \<Rightarrow> bool" where
"subset p q \<equiv> \<forall> s. p s \<subseteq> q s"

lemma
subset_refl[simp]:   "subset p p" and
subset_trans[intro]: "subset p q \<Longrightarrow> subset q r \<Longrightarrow> subset p r"
  unfolding subset_def by blast+

definition follows :: "'level \<Rightarrow> assert \<Rightarrow> assert \<Rightarrow> bool" where
"follows lev p q \<equiv>   implies (assert_sem lev p) (assert_sem lev q) \<and>
                      assert_free q \<subseteq> assert_free p"

lemma follows_assert_free_subset:
"follows lev p q \<Longrightarrow> assert_free  q \<subseteq> assert_free  p"
  unfolding follows_def by auto

lemma
follows_refl[simp]:   "follows lev p p" and
follows_trans[intro]: "follows lev p q \<Longrightarrow> follows lev q r \<Longrightarrow> follows lev p r"
  unfolding follows_def by auto

lemma Sec_mono:
  "lev \<le> lev' \<Longrightarrow> follows lev'' (Rel (Sec x (level_expr lev))) (Rel (Sec x (level_expr lev')))"
  unfolding follows_def implies_def subset_def level_expr_def
  by (auto simp add: sec.simps)

lemma follows_frame:
"\<lbrakk>follows lev p q; star (assert_sem lev p) R (s,h) (s',h')\<rbrakk>
  \<Longrightarrow> star (assert_sem lev q) R (s,h) (s',h')"
  unfolding follows_def by auto

lemma follows_Star_intro[intro]:
"follows lev p q \<Longrightarrow> follows lev (Star p r) (Star q r)"
  unfolding follows_def by auto

lemma follows_Star_com[intro]:
"follows lev (Star p q) (Star q p)"
  unfolding follows_def subset_def using star_com by auto

definition lockinvs :: "ctx \<Rightarrow> assert" where
"lockinvs \<Gamma> = Stars (map lockinv (remdups \<Gamma>))"

fun atomic :: "cmd \<Rightarrow> bool" where
"atomic (Assign _ _) = True" |
"atomic (Load _ _) = True" |
"atomic (Store _ _) = True" |
"atomic (Lock _) = True" |
"atomic (Unlock _) = True" |
"atomic _ = False"

lemma atomic_is_atomic [simp]:
  "atomic c \<Longrightarrow> cmd_step sched (Run c \<Gamma> s h) (Run c' \<Gamma>' s' h') \<Longrightarrow> False"
  by(induction c, auto)

definition SecB :: "expr \<Rightarrow> expr \<Rightarrow> rexpr" where
"SecB e l \<equiv> Sec (Eq e ETrue) l"

inductive prove :: "bool \<Rightarrow> 'level \<Rightarrow> assert \<Rightarrow> cmd \<Rightarrow> assert \<Rightarrow> bool" where
skip:
"prove sh lev Emp Skip Emp" |

assume_:
"prove sh lev Emp (Assume p) (Rel p)" |

assign:
"\<lbrakk>x \<notin> expr_free e; x \<notin> invvars\<rbrakk>
  \<Longrightarrow> prove sh lev Emp (Assign x e) (Rel (Pure (Eq (Var x) e)))" |

load:
"\<lbrakk>x \<notin> expr_frees [p,e]; x \<notin> invvars\<rbrakk>
  \<Longrightarrow> prove sh lev (Star (Rel (if ct then (Sec p (level_expr lev)) else (Pure PTrue))) (Pto p e)) (Load x p) (Star (Rel (Pure (Eq (Var x) e))) (Pto p e))" |

store:
"prove sh lev (Star (Rel (if ct then (Sec p (level_expr lev)) else (Pure PTrue))) (Pto p e)) (Store p e') (Pto p e')" |

(*
shared_atomic:
"\<lbrakk>atomic c; prove False lev (Star P shared) c (Star Q shared)\<rbrakk>
  \<Longrightarrow> prove True lev P c Q" |
*)
lock:
"\<lbrakk>l \<in> locks\<rbrakk>
  \<Longrightarrow> prove sh lev Emp (Lock l) (lockinv l)" |

unlock:
"\<lbrakk>l \<in> locks\<rbrakk>
  \<Longrightarrow> prove sh lev (lockinv l) (Unlock l) Emp" |

if_:
"\<lbrakk>prove sh lev (Star (Rel (Pure b)) p) c1 q;
  prove sh lev (Star (Rel (Pure (Not b))) p) c2 q;
  expr_free b \<subseteq> assert_free p\<rbrakk>
  \<Longrightarrow> prove sh lev (Star (Rel (SecB b (level_expr lev))) p) (If b c1 c2) q" |

while:
"\<lbrakk>prove sh lev (Star (Rel (Pure b)) (Star (Rel (SecB b (level_expr lev))) p)) c (Star (Rel (SecB b (level_expr lev))) p);
  expr_free b \<subseteq> assert_free p\<rbrakk>
  \<Longrightarrow> prove sh lev (Star (Rel (SecB b (level_expr lev))) p) (While b c) (Star (Rel (Pure (Not b))) p)" |

seq:
"\<lbrakk>prove sh lev p c1 r; prove sh lev r c2 q\<rbrakk>
  \<Longrightarrow> prove sh lev p (Seq c1 c2) q" |

par:
"\<lbrakk>prove sh lev p1 c1 q1; prove sh lev p2 c2  q2;
  cmd_mod c1 \<inter> cmd_vars c2 = {};
  cmd_mod c2 \<inter> cmd_vars c1 = {};
  cmd_mod c1 \<inter> assert_free p2 = {};
  cmd_mod c1 \<inter> assert_free q2 = {};
  cmd_mod c2 \<inter> assert_free p1 = {};
  cmd_mod c2 \<inter> assert_free q1 = {}\<rbrakk>
  \<Longrightarrow> prove sh lev (Star p1 p2) (Par c1 c2) (Star q1 q2)" |

frame:
"cmd_mod c \<inter> assert_free r = {}
  \<Longrightarrow> prove sh lev p c q \<Longrightarrow> prove sh lev (Star p r) c (Star q r)" |

conseq:
"\<lbrakk>follows lev p p'; follows lev q' q\<rbrakk>
  \<Longrightarrow> prove sh lev p' c q' \<Longrightarrow> prove sh lev p c q" |

split:
"\<lbrakk>prove sh lev (Star (Rel (Pure b)) (Star (Rel (SecB b (level_expr lev))) p)) c q;
  prove sh lev (Star (Rel (Pure (Not b))) (Star (Rel (SecB b (level_expr lev))) p)) c q;
  expr_free b \<subseteq> assert_free p\<rbrakk>
  \<Longrightarrow> prove sh lev (Star (Rel (SecB b (level_expr lev))) p) c q" |

out:
  "prove sh lev (Rel (Sec eval elev)) (Output elev eval) Emp"


end

end