# Isabelle formalisation for Section 5

Of the extended security concurrent separation logic (Section 5)
from the paper
"Assume but Verify: Deductive Verification of Leaked Information in Concurrent Applications"

## Building

Tested with Isabelle2022 https://isabelle.in.tum.de/index.html

Install Isabelle, making sure that the `isabelle` binary is in your `PATH` and then run:

```
isabelle build -c -d . -v SecCSL
```

It should produce output like:
```
[...]

Session Pure/Pure
Session Misc/Tools
Session HOL/HOL (main)
Session Unsorted/SecCSL
Cleaned SecCSL
Running SecCSL ...
SecCSL: theory SecCSL.Syntax
SecCSL: theory SecCSL.Semantics
SecCSL: theory SecCSL.Separation
SecCSL: theory SecCSL.Logic
SecCSL: theory SecCSL.Locks
SecCSL: theory SecCSL.Shared
SecCSL: theory SecCSL.Soundness
SecCSL: theory SecCSL.SecCSL
SecCSL: theory SecCSL.Knowledge
Timing SecCSL (4 threads, 71.709s elapsed time, 202.865s cpu time, 14.145s GC time, factor 2.83)
Finished SecCSL (0:01:13 elapsed time, 0:03:23 cpu time, factor 2.78)
```

## Theory Organisation

Browse with

```
isabelle jedit Knowledge.thy
```

* `Knowledge.thy`: top-level theory with top-level soundness statement
* `SecCSL.thy`: almost top-level theory
* `Syntax.thy`: syntax of expressions, assertions, commands
* `Semantics.thy`: semantics of expressions, commands
* `Separation.thy`: separation logic connectives
* `Logic.thy`: definition of the logic's rules, entailment etc.
* `Locks.thy`: helper lemmas on the lock invariants
* `Soundness.thy`: details of the soundness proof


## Location of major definitions and theorems

The rules of the logic are defined by the judgement `prove` in `Logic.thy`

Attacker uncertainty (Definition 5.3)
is defined by `uncertainty2` in `Knowledge.thy`. Note that
observably equivalent schedules (Definition 5.2) is captured by the
conjunction of `sched_eq` and `ct_eq` in `uncertainty2`.

Release by assumption (Definition 5.5) is defined equivalently
by `allowed_release2` in `Knowledge.thy`. Note that the existence of
the schedule prefixes in which the failed assumption occurs
is written verbosely in `allowed_release2`, but is instead captured
equivalently in the paper by the Assumption failure definition
(Definition 5.4)

The top-level soundness theorem (Theorem 5.6) is proved and stated
equivalently by `controlled_release2` in `Knowlwedge.thy`. Note that
`controlled_release2` defines the soundness statement via the intermediate
definition `controlled_leakage2`.
