theory Soundness
  imports Shared
begin

context concurrency
begin

abbreviation maybe_shared :: "bool \<Rightarrow> assert" where
  "maybe_shared sh \<equiv> (if sh then shared else Emp)"

definition inv :: "bool \<Rightarrow> 'level \<Rightarrow> ctx \<Rightarrow> assert \<Rightarrow> assert \<Rightarrow> 'a pred" where
(* - shape of the local heap
   - invariants of shared, non-held locks
   - the shared part of the heap
   - fix a specific frame *)
"inv sh lev \<Gamma> p f
  = stars ([assert_sem lev p,
           assert_sem lev (lockinvs \<Gamma>),
           assert_sem lev (maybe_shared sh),
           assert_sem lev f])"

lemma inv_reorder:
"set \<Gamma> = set \<Gamma>' \<Longrightarrow> inv sh lev \<Gamma> p f = inv sh lev \<Gamma>' p f"
  unfolding inv_def apply (clarsimp)
  by (metis lockinvs_sem_reorder)

lemma inv_pure[simp]:
"inv sh lev \<Gamma> (Star (Rel (Pure b)) p) f (s, H) (s', H')
  = (pure_sem b s \<and> pure_sem b s' \<and> inv sh lev \<Gamma> p f (s, H) (s', H'))"
  unfolding inv_def by simp

lemma inv_low[simp]:
"inv sh lev \<Gamma> (Star (Rel (Sec b (level_expr lev))) p) f (s, H) (s', H')
  = (expr_sem b s = expr_sem b s' \<and> inv sh lev \<Gamma> p f (s, H) (s', H'))"
  unfolding inv_def 
  by (simp add: sec.simps)

lemma expr_sem_SecB[simp]:
  "rexpr_sem lev (SecB b (level_expr lev)) s s' =
    (pure_sem b s = pure_sem b s')"
  apply(simp add: SecB_def sec.simps pure_sem_def)
  apply(simp add: Eq_def ETrue_def true_neq_false)
  using true_neq_false by blast

lemma inv_low_SecB[simp]:
"inv sh lev \<Gamma> (Star (Rel (SecB b (level_expr lev))) p) f (s, H) (s', H')
  = (pure_sem b s = pure_sem b s' \<and> inv sh lev \<Gamma> p f (s, H) (s', H'))"
  unfolding inv_def 
  by (simp add: sec.simps split del: if_splits)

lemma inv_rel:
"inv sh lev \<Gamma> (Star (Rel r) p) f (s, H) (s', H')
  = (rel (rexpr_sem lev r) (s, Map.empty) (s',Map.empty) \<and> inv sh lev \<Gamma> p f (s, H) (s', H'))"
  unfolding inv_def 
  by (simp)

(* reverse only holds if resource invariants and frame are precise *)
lemma inv_and:
"inv sh lev \<Gamma> (AAnd p q) f (s, H) (s', H')
  \<Longrightarrow> (inv sh lev \<Gamma> p f (s, H) (s', H') \<and> inv sh lev \<Gamma> q f (s, H) (s', H'))"
   unfolding inv_def by (auto elim!: star_elim intro!: star_intro)

inductive step :: "bool \<Rightarrow> 'level \<Rightarrow> cmd \<Rightarrow> ctx \<Rightarrow> assert \<Rightarrow> ctx \<Rightarrow> assert \<Rightarrow> 'a rel" where
step_intro:
(* set \<Gamma>1 \<subseteq> locks \<Longrightarrow> *)
"\<lbrakk>assert_free p2 \<subseteq> assert_free p1 \<union> cmd_vars c1 \<union> invvars; 
  set \<Gamma>2 \<subseteq> locks\<rbrakk>
  \<Longrightarrow> step sh lev c1 \<Gamma>1 p1 \<Gamma>2 p2 s1 s2"

lemma step_reorder:
"set \<Gamma>1 = set \<Gamma>1' \<Longrightarrow> set \<Gamma>2 = set \<Gamma>2' \<Longrightarrow> step sh lev c1 \<Gamma>1 p1 \<Gamma>2 p2 s1 s2 = step sh lev c1 \<Gamma>1' p1 \<Gamma>2' p2 s1 s2"
  unfolding step.simps apply (clarsimp)
  (*using lockinvs_lows by auto *) done

inductive_cases step_elim: "step sh lev c1 \<Gamma>1 p1 \<Gamma>2 p2 s1 s2"

inductive pre :: "bool \<Rightarrow> 'level \<Rightarrow> ('level,'a) sched \<Rightarrow> ('level,'a) sched \<Rightarrow> locks \<Rightarrow> assert \<Rightarrow> assert \<Rightarrow> cmd \<Rightarrow> 'a conf \<Rightarrow> 'a conf \<Rightarrow> 'a pred" where
pre_intro:
"\<lbrakk>cmd_mod c1 \<inter> assert_free f = {};
  cmd_mod c1 \<inter> invvars = {};
  set \<Gamma>1 \<subseteq> locks;
  cmd_step \<sigma> (Run c1 (set \<Gamma>1) s1  H1)  k;
  cmd_step \<sigma>' (Run c1 (set \<Gamma>1) s1' H1') k';
  sched_eq \<sigma> \<sigma>';
  inv sh lev \<Gamma>1 p1 f (s1,H1) (s1',H1')\<rbrakk>
  \<Longrightarrow> pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) p1 f c1 k k' (s1,H1) (s1',H1')"

(* don't use this one *)
inductive_cases pre_elim_: "pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) p1 f c1 k k' (s1,H1) (s1',H1')"

lemma pre_elim:
  assumes
    "pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) p1 f c1 k k' (s1,H1) (s1',H1')"
  obtains
    "cmd_mod c1 \<inter> assert_free f = {}"
    "cmd_mod c1 \<inter> invvars = {}"
    "set \<Gamma>1 \<subseteq> locks"
    "cmd_step \<sigma> (Run c1 (set \<Gamma>1) s1  H1)  k"
    "cmd_step \<sigma>' (Run c1 (set \<Gamma>1) s1' H1') k'"
    "sched_eq \<sigma> \<sigma>'"
    "inv sh lev \<Gamma>1 p1 f (s1,H1) (s1',H1')"
    using assms by (auto elim!: pre_elim_ intro!: that, subst inv_reorder, auto)

lemma pre_pure[simp]:
"pre sh lev \<sigma> \<sigma>' (set \<Gamma>) (Star (Rel (Pure b)) p) f c k k' (s, H) (s', H')
  = (pure_sem b s \<and> pure_sem b s' \<and> pre sh lev \<sigma> \<sigma>' (set \<Gamma>) p f c k k' (s, H) (s', H'))"
  by (metis inv_pure pre_elim pre_elim_ pre_intro)

lemma pre_sec[simp]:
"pre sh lev \<sigma> \<sigma>' (set \<Gamma>) (Star (Rel (Sec b (level_expr lev))) p) f c k k' (s, H) (s', H')
  = (expr_sem b s = expr_sem b s' \<and> pre sh lev \<sigma> \<sigma>' (set \<Gamma>) p f c k k' (s, H) (s', H'))"
  using pre_elim pre_intro inv_low 
  by (metis pre_elim_)

lemma pre_elim_stop:
  assumes
    "pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) p1 f c1 (Stop \<Gamma>2 s2 H2) (Stop \<Gamma>2 s2' H2') (s1,H1) (s1',H1')"
  obtains (pre) \<sigma>  where
    "cmd_mod c1 \<inter> assert_free f = {}"
    "set \<Gamma>1 \<subseteq> locks"
    (* add more stuff here if needed *)
    "coincide s1 s2 (- cmd_mod c1)"
    "coincide s1' s2' (- cmd_mod c1)"
    "inv sh lev \<Gamma>1 p1 f (s1,H1) (s1',H1')"
  using assms by (auto elim!: pre_elim dest!: cmd_step_mod)

lemma pre_elim_run:
  assumes
    "pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) p1 f c1 (Run c2 \<Gamma>2 s2 H2) (Run c2 \<Gamma>2 s2' H2') (s1,H1) (s1',H1')"
  obtains (pre) blah where
    "cmd_mod c1 \<inter> assert_free f = {}"
    "cmd_mod c1 \<inter> invvars = {}"
    "set \<Gamma>1 \<subseteq> locks"
    (* add more stuff here if needed *)
    "cmd_mod c2 \<subseteq> cmd_mod c1"
    "cmd_vars c2 \<subseteq> cmd_vars c1"
    "coincide s1 s2 (- cmd_mod c1)"
    "coincide s1' s2' (- cmd_mod c1)"
    "inv sh lev \<Gamma>1 p1 f (s1,H1) (s1',H1')"
  using assms by (auto elim!: pre_elim, metis cmd_step_mod cmd_step_preserves_mod cmd_step_preserves_vars conf_mod.simps conf_vars.simps)

fun sched_asm :: "('level,'a) tick list \<Rightarrow> rexpr option" where
  "sched_asm ((Assumption p)#\<sigma>) = (Some p)" |
  "sched_asm (Tick#\<sigma>) = None" |
  "sched_asm (Fst#\<sigma>) = sched_asm \<sigma>" |
  "sched_asm (Snd#\<sigma>) = sched_asm \<sigma>" |
  "sched_asm ((MemLoad x)#\<sigma>) = None" |
  "sched_asm ((MemStore x)#\<sigma>) = None" |
  "sched_asm ((Out x y)#\<sigma>) = None" |
  "sched_asm ((Branch v)#\<sigma>) = None" |
  "sched_asm [] = None"

inductive
  asm :: "'level \<Rightarrow> ('level,'a) sched \<Rightarrow> 'a rel" where
  "sched_asm \<sigma> = Some p \<Longrightarrow> 
   \<not> (rel (rexpr_sem lev p) (s1, Map.empty) (s1', Map.empty)) \<Longrightarrow>
   asm lev \<sigma> s1 s1'"

lemma asm_tick[dest]:
  "asm lev \<sigma> s1 s1' \<Longrightarrow> hd \<sigma> = Tick \<or> (\<exists>y. hd \<sigma> = MemLoad y) \<or> (\<exists>y. hd \<sigma> = MemStore y) \<or> (\<exists> a b. hd \<sigma> = Out a b) \<Longrightarrow> False"
  by(induct \<sigma>, auto elim: asm.cases)


lemma asm_tick'[simp]:
  "asm lev [Tick] s1 s1'  \<Longrightarrow> False"
  "asm lev [MemLoad x] s1 s1'  \<Longrightarrow> False"
  "asm lev [MemStore x] s1 s1'  \<Longrightarrow> False"
  "asm lev [Out lev e] s1 s1' \<Longrightarrow> False"
  by(fastforce dest: asm_tick)+

definition
  outs :: "'level \<Rightarrow> ('level,'a) sched \<Rightarrow> ('level,'a) sched"
  where
  "outs lev \<sigma> = (filter (\<lambda>e. (\<exists>x l. e = Out l x \<and> l\<le>lev) \<or> (\<exists>v. e = Branch v)) \<sigma>)"

definition
  ct_eq :: "'level \<Rightarrow> ('level,'a) sched \<Rightarrow> ('level,'a) sched \<Rightarrow> bool"
  where
  "ct_eq lev \<sigma> \<sigma>' \<equiv> (outs lev \<sigma> = outs lev \<sigma>') \<and> 
                    (if ct then (map (\<lambda>ev. case ev of Out l x \<Rightarrow> Out l undefined | _ \<Rightarrow> ev) \<sigma> =
                             map (\<lambda>ev. case ev of Out l x \<Rightarrow> Out l undefined | _ \<Rightarrow> ev) \<sigma>')
                  else length \<sigma> = length \<sigma>')"

inductive
rec    :: "bool \<Rightarrow> 'level \<Rightarrow> nat \<Rightarrow> locks \<Rightarrow> assert \<Rightarrow> assert \<Rightarrow> cmd \<Rightarrow> 'a conf \<Rightarrow> 'a conf \<Rightarrow> assert \<Rightarrow> 'a pred"
  and
secure :: "bool \<Rightarrow> 'level \<Rightarrow> nat \<Rightarrow> assert \<Rightarrow> cmd \<Rightarrow> assert \<Rightarrow> bool"
where

secure_stop[intro!]:
"\<lbrakk>inv sh lev \<Gamma>' q f (s2,H2) (s2',H2');
  step sh lev c1 \<Gamma>1 p1 \<Gamma>' q s1 s2\<rbrakk>
  \<Longrightarrow> rec sh lev n (set \<Gamma>1) p1 f c1 (Stop (set \<Gamma>') s2 H2) (Stop (set \<Gamma>') s2' H2') q (s1,H1) (s1',H1')" |

secure_rec[intro!]:
"\<lbrakk>inv sh lev \<Gamma>2 p2 f (s2,H2) (s2',H2');
  step sh lev c1 \<Gamma>1 p1 \<Gamma>2 p2 s1 s2;
  secure sh lev n p2 c2 q\<rbrakk>
  \<Longrightarrow> rec sh lev n (set \<Gamma>1) p1 f c1 (Run c2 (set \<Gamma>2) s2 H2) (Run c2 (set \<Gamma>2) s2' H2') q (s1,H1) (s1',H1')" |

[simp]: "secure sh lev 0 p1 c1 q" |

secure_intro:
"\<lbrakk>\<And> \<Gamma>1 s1 s1' H1 H1' k k' f \<sigma> \<sigma>'.
        pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) p1 f c1 k k' (s1,H1) (s1',H1')
    \<Longrightarrow> ((rec sh lev n (set \<Gamma>1) p1 f c1 k k' q (s1,H1) (s1',H1') \<and> (ct_eq lev \<sigma> \<sigma>')) \<or>
         asm lev \<sigma> s1 s1')\<rbrakk>
  \<Longrightarrow> secure sh lev (Suc n) p1 c1 q" 

(* adds that m is non-zero, fixes some \<Gamma>1 *)
lemma secure_intro':
  assumes "\<And> n \<Gamma>1 s1 s1' H1 H1' k k' f \<sigma> \<sigma>'.
        m = Suc n
    \<Longrightarrow> pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) p1 f c1 k k' (s1,H1) (s1',H1')
    \<Longrightarrow> (rec sh lev n (set \<Gamma>1) p1 f c1 k k' q (s1,H1) (s1',H1') \<and> (ct_eq lev \<sigma> \<sigma>')) \<or>
        asm lev \<sigma> s1 s1'"
  shows "secure sh lev m p1 c1 q"
  apply (cases m, simp_all)
  using assms by (blast intro!: secure_intro)

(* for non-assumption steps; works better with automation *)
lemma secure_intro'':
  assumes "\<And> n \<Gamma>1 s1 s1' H1 H1' k k' f \<sigma> \<sigma>'.
        m = Suc n
    \<Longrightarrow> pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) p1 f c1 k k' (s1,H1) (s1',H1')
    \<Longrightarrow> rec sh lev n (set \<Gamma>1) p1 f c1 k k' q (s1,H1) (s1',H1') \<and> (ct_eq lev \<sigma> \<sigma>')"
  shows "secure sh lev m p1 c1 q"
  using secure_intro' assms by blast

(* don't use this *)
inductive_cases rec_elim_:   "rec sh lev n L1 p1 f c k k' q (s,H) (s',H')"

lemma rec_elim:
  assumes "rec sh lev n (set \<Gamma>1) p1 f c1 k k' q (s1,H1) (s1',H1')"
  obtains
    (stop) \<Gamma>' s2 H2 s2' H2' where
      "k = Stop (set \<Gamma>') s2 H2"
      "k' = Stop (set \<Gamma>') s2' H2'"
      "inv sh lev \<Gamma>' q f (s2,H2) (s2',H2')"
      "step sh lev c1 \<Gamma>1 p1 \<Gamma>' q s1 s2" |
    (rec) p2 \<Gamma>2 c2 s2 H2 s2' H2' where
      "k = Run c2 (set \<Gamma>2) s2 H2"
      "k' = Run c2 (set \<Gamma>2) s2' H2'"
      "inv sh lev \<Gamma>2 p2 f (s2,H2) (s2',H2')"
      "step sh lev c1 \<Gamma>1 p1 \<Gamma>2 p2 s1 s2"
      "secure sh lev n p2 c2 q"
proof -
  note stop = that(1)
  note rec = that(2)

  then show thesis proof (cases rule: rec_elim_[OF assms])
    case (1 \<Gamma>' s2 H2 s2' H2' \<Gamma>1)
    with inv_reorder[OF 1(1)]  step_reorder[OF 1(1)]
    show ?thesis by (auto intro!: stop)
  next
    case (2 \<Gamma>2 p2 s2 H2 s2' H2' \<Gamma>1 c2)
    with inv_reorder[OF 2(1)] step_reorder[OF 2(1)]
    show ?thesis by (auto intro!: rec)
  qed
qed

lemma step_pure_dest:
  assumes "step sh lev c1 \<Gamma>1 (Star (Rel (Pure b)) p1) \<Gamma>' q s1 s2"
  assumes "expr_free b \<subseteq> assert_free p1"
  shows   "step sh lev c1 \<Gamma>1 p1 \<Gamma>' q s1 s2"
    using assms by (elim step_elim, intro step_intro) auto

lemma rec_pure_dest:
  assumes "rec sh lev n (set \<Gamma>1) (Star (Rel (Pure b)) p1) f c1 k k' q (s1,H1) (s1',H1')"
  assumes "expr_free b \<subseteq> assert_free p1"
  shows   "rec sh lev n (set \<Gamma>1) p1 f c1 k k' q (s1,H1) (s1',H1') \<and> expr_free b \<subseteq> assert_free p1"
  using assms by (auto elim: rec_elim dest: step_pure_dest)

inductive_cases secure_elim: "secure sh lev (Suc n) p1 c1 q"

lemma secure_step:
  assumes "secure sh lev (Suc n) p1 c1 q"
  assumes "pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) p1 f c1 k k' (s1,H1) (s1',H1')"
  shows   "rec sh lev n (set \<Gamma>1) p1 f c1 k k' q (s1,H1) (s1',H1') \<and> (ct_eq lev \<sigma> \<sigma>') \<or>
           asm lev \<sigma> s1 s1'"
  using assms by (blast elim!: secure_elim)

lemma cmd_coincidences:
  assumes "coincide s s' (- cmd_mod c)"
  assumes "cmd_mod c \<inter> assert_free p = {}"
  obtains "coincide s s' (assert_free p)"
  using assms by blast

lemma secure_prefix:
  assumes "secure sh lev n p c q"
  assumes "m \<le> n"
  shows   "secure sh lev m p c q"
  using assms proof (induction m arbitrary: n p c q)
  case 0
  then show ?case
    by simp
next
  case (Suc m0)

  from Suc(3) obtain n0 where
    n0: "n = Suc n0" by (cases n) simp
  with Suc(2) have
    sec: "secure sh lev (Suc n0) p c q" by simp
  show ?case proof (rule secure_intro, goal_cases)
    case (1 \<Gamma>1 s1 s1' H1 H1' k k' f \<sigma> \<sigma>')
    show ?case proof(cases "asm lev \<sigma> s1 s1'")
      case True
      then show ?thesis by blast
    next
      case False
      with secure_step[OF sec "1"]
      have "rec sh lev n0 (set \<Gamma>1) p f c k k' q (s1, H1) (s1', H1') \<and> (ct_eq lev \<sigma> \<sigma>')" 
        by auto+
      thus ?thesis 
        using Suc(1) Suc(3) n0 by(auto elim!: rec_elim)
    qed
  qed
qed

lemma ct_eq_refl[intro!,simp]: "ct_eq lev \<sigma> \<sigma>"
  by(auto simp: ct_eq_def)

lemma secure_skip:
  shows "secure sh lev n Emp Skip Emp"
  by (auto elim!: pre_elim intro!: secure_intro'' step_intro)

lemma secure_assume:
  shows "secure sh lev n Emp (Assume p) (Rel p)"
  apply(case_tac n, simp_all)
  apply(rule secure_intro')
  by(auto elim!: pre_elim intro!: step_intro asm.intros simp: inv_def)

lemma secure_output_helper:
  "sec lev (expr_sem e) (expr_sem elev) s1 s1' \<Longrightarrow> ct_eq lev [Out (to_level (expr_sem elev s1)) (expr_sem e s1)]
            [Out (to_level (expr_sem elev s1')) (expr_sem e s1')]"
  by(clarsimp simp: ct_eq_def sec.simps outs_def)

lemma secure_output:
  shows "secure sh lev n (Rel (Sec e elev)) (Output elev e) Emp"
  apply(case_tac n, simp_all)
  apply(rule secure_intro')
  by(auto elim!: pre_elim intro!: step_intro asm.intros simp: inv_def secure_output_helper)

lemma secure_assign:
  assumes
    "x \<notin> expr_free e"
    "x \<notin> invvars"
  shows
    "secure sh lev n Emp (Assign x e) (Rel (Pure (Eq (Var x) e)))"
  using assms by (auto simp: inv_def invvars_def elim!: pre_elim star_elim intro!: secure_intro'' step_intro star_intro)


definition pre_extra where
  "pre_extra lev p \<equiv> (if ct then (Sec p (level_expr lev)) else (Pure PTrue))"

lemma concurrenty_pre_extra_simp[simp]:
  "ct \<Longrightarrow> (concurrency.pre_extra level_str True lev p) = 
            (Sec p (level_expr lev))"
  by (metis concurrency.pre_extra_def concurrency_axioms level_expr_sem order_refl rexpr_sem.simps(2) sec_elim to_val_to_level)  

lemma secure_load_helper:
  "rexpr_sem lev (pre_extra lev p) s1 s1' \<Longrightarrow>
    ct_eq lev [MemLoad (expr_sem p s1)] [MemLoad (expr_sem p s1')]"
  by(auto simp: ct_eq_def sec.simps outs_def)

lemma secure_load:
  assumes
    "x \<notin> expr_frees [p,e]"
    "x \<notin> invvars"
  shows
    "secure sh lev n (Star (Rel (pre_extra lev p)) (Pto p e)) (Load x p) (Star (Rel (Pure (Eq (Var x) e))) (Pto p e))"
  (* FIXME: horrible proof full of beer offences *)
  using assms
  apply (auto elim!: pre_elim intro!: secure_intro'' step_intro)
       apply (auto simp: inv_def)
  by (auto intro!: star_intro elim!: star_elim simp: secure_load_helper)


lemma secure_store:
  shows
    "secure sh lev n (Star (Rel (pre_extra lev p)) ((Pto p e))) (Store p e') ((Pto p e'))"
  by (auto simp: inv_def elim!: pre_elim intro!: secure_intro'' step_intro pto_frame_write, auto simp: sec.simps ct_eq_def outs_def sec.simps)

lemma shared_atomic_pre [simp]:
  "pre True lev \<sigma> \<sigma>' (set \<Gamma>1) P f c k k' (s1, H1) (s1', H1') \<Longrightarrow>
   pre False lev \<sigma> \<sigma>' (set \<Gamma>1) (Star P shared) f c k k' (s1, H1) (s1', H1')"
  apply(erule pre_elim)
  apply(rule pre_intro, simp_all)
  by(simp add: inv_def star_ac star_com star_ass)

lemma shared_atomic_inv [simp]:
  "local.inv False lev \<Gamma>' (Star Q shared) f (s2, H2) (s2', H2') \<Longrightarrow>
   local.inv True lev \<Gamma>' Q f (s2, H2) (s2', H2')"
  by(simp add: inv_def star_ac star_com star_ass)

lemma shared_atomic_step [simp]:
  "step False lev c \<Gamma>1 (Star P shared) \<Gamma>' (Star Q shared) s1 s2 \<Longrightarrow> 
   step True lev c \<Gamma>1 P \<Gamma>' Q s1 s2"
  apply(erule step_elim)
  apply(rule step_intro)
  using shared_free by auto

lemma atomic_is_not_assumption:
  assumes "pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) P f c k k' (s1, H1) (s1', H1')"
  assumes "atomic c"
  assumes "asm lev \<sigma> s1 s1'"
  shows "False"
  using assms apply(cases c)
  by(auto elim!: pre_elim)

(*
lemma secure_shared_atomic:
  assumes
    "atomic c"
    "secure False lev n (Star P shared) c (Star Q shared)"
  shows
    "secure True lev n P c Q"
proof(rule secure_intro'')
  fix n' \<Gamma>1 s1 s1' H1 H1' k k' f \<sigma> \<sigma>'
  assume n_def: "n = Suc n'" and
         pre: "pre True lev \<sigma> \<sigma>' (set \<Gamma>1) P f c k k' (s1, H1) (s1', H1')"

  have "n' < n" using n_def by blast

  with pre shared_atomic_pre secure_elim n_def "assms"(2) have 
  "rec False lev n' (set \<Gamma>1) (Star P shared) f c k k' (Star Q shared) (s1, H1) (s1', H1') \<and>  (ct \<longrightarrow> \<sigma> = \<sigma>')"
    using secure_step atomic_is_not_assumption assms(1) sledgehammer

  thus "rec True lev n' (set \<Gamma>1) P f c k k' Q (s1, H1) (s1', H1') \<and>  (ct \<longrightarrow> \<sigma> = \<sigma>')"
    apply(rule conjE)
    apply(erule rec_elim)
      apply(fastforce intro: secure_stop)
    using pre_elim[OF pre]
    by (metis assms(1) atomic_is_atomic)
qed
*)

lemma secure_lock:
  assumes "l \<in> locks"
  shows   "secure sh lev n Emp (Lock l) (lockinv l)"
proof (auto elim!: pre_elim intro!: secure_intro'')
  fix n \<Gamma>1 s1 s1' H1 H1' f
  assume
    l: "l \<in> set \<Gamma>1"
  then obtain \<Gamma>2 where
    \<Gamma>2: "set \<Gamma>2 = set \<Gamma>1 - {l}"
    using remove_lock by force
  assume
    "inv sh lev \<Gamma>1 Emp f (s1, H1) (s1', H1')"
    "set \<Gamma>1 \<subseteq> locks"
  with l \<Gamma>2 assms have
    "rec sh lev n (set \<Gamma>1) Emp f (Lock l) (Stop (set \<Gamma>2) s1 H1) (Stop (set \<Gamma>2) s1' H1') (lockinv l) (s1, H1) (s1', H1')"
    apply (intro secure_stop)
    apply (simp add: inv_def lockinvs_sem_lock)
    apply (rule step_intro)
     apply simp_all
    subgoal by (rule lockinv_free)
-     by blast
  with \<Gamma>2 show
    "rec sh lev n (set \<Gamma>1) Emp f (Lock l) (Stop (set \<Gamma>1 - {l}) s1 H1) (Stop (set \<Gamma>1 - {l}) s1' H1') (lockinv l) (s1, H1) (s1', H1')"
    by simp
qed

lemma secure_unlock:
  assumes "l \<in> locks"
  shows   "secure sh lev n (lockinv l) (Unlock l) Emp"
proof (auto elim!: pre_elim intro!: secure_intro'')
  fix n \<Gamma>1 s1 s1' H1 H1' f
  assume
    "inv sh lev \<Gamma>1 (lockinv l) f (s1, H1) (s1', H1')"
    "l \<notin> set \<Gamma>1"
    "set \<Gamma>1 \<subseteq> locks"
  with assms have
    "rec sh lev n (set \<Gamma>1) (lockinv l) f (Unlock l) (Stop (set (l#\<Gamma>1)) s1 H1) (Stop (set (l#\<Gamma>1)) s1' H1') Emp (s1, H1) (s1', H1')"
    apply (intro secure_stop)
    by (auto simp: inv_def intro!: step_intro) (* do the intro before simplifying set (l#\<Gamma>1) *)
  then show "rec sh lev n (set \<Gamma>1) (lockinv l) f (Unlock l) (Stop (insert l (set \<Gamma>1)) s1 H1) (Stop (insert l (set \<Gamma>1)) s1' H1') Emp (s1, H1) (s1', H1')"
    by simp
qed

lemma inv_conseq:
  assumes "inv sh lev \<Gamma> p f (s,H) (s',H')"
  assumes "follows lev p p'"
  shows   "inv sh lev \<Gamma> p' f (s,H) (s',H')"
  using assms by (simp add: inv_def follows_frame)

lemma pre_conseq:
  assumes "pre sh lev \<sigma> \<sigma>' (set \<Gamma>) p f c k k' (s, H1) (s', H1')"
  assumes "follows lev p p'"
  shows   "pre sh lev \<sigma> \<sigma>' (set \<Gamma>) p' f c k k' (s, H1) (s', H1')"
  using assms by (auto elim!: pre_elim intro!: pre_intro simp: inv_conseq)

lemma step_conseq:
  assumes c: "step sh lev c \<Gamma> p' \<Gamma>' q' s s'"
  assumes p: "follows lev p p'"
  assumes q: "follows lev q' q"
  shows   "step sh lev c \<Gamma> p \<Gamma>' q s s'"
  using c 
  follows_assert_free_subset[OF p] follows_assert_free_subset[OF q]
  by (auto elim!: step_elim intro!: step_intro) 

lemma secure_conseq_pre:
  assumes "secure sh lev n p' c q"
  assumes "follows lev p p'"
  shows   "secure sh lev n p c q"
  apply(rule secure_intro')
  apply(drule pre_conseq)
   apply(rule assms(2))
  apply(insert assms(1))
  apply(simp, drule secure_step, assumption)
  apply(erule disj_forward)
  using assms(2)
   apply(auto elim!: rec_elim intro: step_conseq inv_conseq)[1]
  by simp


lemma secure_conseq_post:
  assumes "secure sh lev n p c q'"
  assumes "follows lev q' q"
  shows   "secure sh lev n p c q"
  using assms
  apply (induction n arbitrary: c p q rule: nat_less_induct)
  apply(rule secure_intro')
  apply simp
  apply(drule (1) secure_step)
  apply(erule disj_forward)
   apply(auto elim!: rec_elim intro: step_conseq inv_conseq)[1] 
  apply clarsimp
  done

lemma secure_conseq:
  assumes "secure sh lev n p' c q'"
  assumes "follows lev p p'"
  assumes "follows lev q' q"
  shows   "secure sh lev n p c q"
  using assms secure_conseq_pre secure_conseq_post by blast

lemma secure_split:
  assumes c1: "secure sh lev n (Star (Rel (Pure b)) (Star (Rel (SecB b (level_expr lev))) p)) c q"
  assumes c2: "secure sh lev n (Star (Rel (Pure (Not b))) (Star (Rel (SecB b (level_expr lev))) p)) c q"
  assumes b:  "expr_free b \<subseteq> assert_free p"
  shows       "secure sh lev n (Star (Rel (SecB b (level_expr lev))) p) c q"
proof (rule secure_intro', goal_cases)
  case (1 m \<Gamma> s1 s1' H1 H1' k k' f \<sigma> \<sigma>')
  note m[simp] = 1(1)
  note pre = 1(2)
  from pre have bs: "pure_sem b s1 = pure_sem b s1'"
    by (auto elim: pre_elim)
  show ?case proof (cases "pure_sem b s1")
    case True
    have p': "follows lev (Star (Rel (Pure b)) (Star (Rel (SecB b (level_expr lev))) p)) (Star (Rel (Pure b)) p)"
      by (auto simp: follows_def implies_def)
    have pre1: "pre sh lev \<sigma> \<sigma>' (set \<Gamma>) (Star (Rel (Pure b)) (Star (Rel (SecB b (level_expr lev))) p)) f c k k' (s1, H1) (s1', H1')"
      using pre bs True by simp
    with secure_step[OF c1[simplified] pre1]

    show ?thesis
      by(auto dest: rec_pure_dest simp: SecB_def)
  next
    case False
    have p': "follows lev (Star (Rel (Pure (Not b))) (Star (Rel (SecB b (level_expr lev))) p)) (Star (Rel (Pure (Not b))) p)"
      by (auto simp: follows_def implies_def)
    have pre1: "pre sh lev \<sigma> \<sigma>' (set \<Gamma>) (Star (Rel (Pure (Not b))) (Star (Rel (SecB b (level_expr lev))) p)) f c k k' (s1, H1) (s1', H1')"
      using pre bs False by simp
    with secure_step[OF c2[simplified] pre1] show ?thesis
      using bs False by (auto dest: rec_pure_dest simp: SecB_def)
  qed
qed

lemma secure_if:
  assumes c1: "secure sh lev n (Star (Rel (Pure b)) p) c1 q"
  assumes c2: "secure sh lev n (Star (Rel (Pure (Not b))) p) c2 q"
  assumes b:  "expr_free b \<subseteq> assert_free p"
  shows       "secure sh lev (Suc n) (Star (Rel (SecB b (level_expr lev))) p) (If b c1 c2) q"
proof (rule secure_intro)
  fix \<Gamma> s1 s1' H1 H1' k k' f \<sigma> \<sigma>'
  assume "pre sh lev \<sigma> \<sigma>' (set \<Gamma>) (Star (Rel (SecB b (level_expr lev))) p) f (cmd.If b c1 c2) k k' (s1, H1) (s1', H1')"
  then have "rec sh lev n (set \<Gamma>) (Star (Rel (SecB b (level_expr lev))) p) f (cmd.If b c1 c2) k k' q (s1, H1) (s1', H1') \<and> (\<sigma> = \<sigma>')"
  proof (cases rule: pre_elim)
    case 1
    from 1(7) have
      bs: "pure_sem b s1 = pure_sem b s1'"
      by simp
    (* This whole thing is a bit brittle to avoid picking auto
       to instantiate the successor assertion from the plain invariant pre(4).
       Instead we're explicitly spelling out the strong invariant
       including the conditional evaluation *and* the security classification,
       and then get rid of the second by consequence
       but *keeping* the evaluation of the conditional to match the hypothesis c1/c2.
     *)
    show ?thesis proof (cases "pure_sem b s1")
      case True
      from True 1(7) bs have
        inv: "inv sh lev \<Gamma> (Star (Rel (Pure b)) (Star (Rel (SecB b (level_expr lev))) p)) f (s1, H1) (s1', H1')"
        by simp
      have "follows lev (Star (Rel (Pure b)) (Star (Rel (SecB b (level_expr lev))) p)) (Star (Rel (Pure b)) p)"
        by (auto simp: follows_def implies_def)
      with True 1(1-5) b bs c1 show ?thesis
        by (auto intro: inv step_intro secure_conseq_pre)
    next
      case False
      from False 1(7) bs have
        inv: "inv sh lev \<Gamma> (Star (Rel (Pure (Not b))) (Star (Rel (SecB b (level_expr lev))) p)) f (s1, H1) (s1', H1')"
        by simp
      have "follows lev (Star (Rel (Pure (Not b))) (Star (Rel (SecB b (level_expr lev))) p)) (Star (Rel (Pure (Not b))) p)"
        by (auto simp: follows_def implies_def)
      with False 1(1-5) b bs c2 show ?thesis 
        by (auto intro: inv step_intro secure_conseq_pre)
    qed
  qed
  (* FIXME: ugh *)
  thus 
  "rec sh lev n (set \<Gamma>) (Star (Rel (SecB b (level_expr lev))) p) f (cmd.If b c1 c2) k k'
        q (s1, H1) (s1', H1') \<and> (ct_eq lev \<sigma> \<sigma>') \<or>
   asm lev \<sigma> s1 s1'" by blast
qed

lemma secure_if':
  assumes c1: "\<And> n. secure sh lev n (Star (Rel (Pure b)) p) c1 q"
  assumes c2: "\<And> n. secure sh lev n (Star (Rel (Pure (Not b))) p) c2 q"
  assumes b:  "expr_free b \<subseteq> assert_free p"
  shows       "secure sh lev n (Star (Rel (SecB b (level_expr lev))) p) (If b c1 c2) q"
  using assms secure_if by (cases n, auto)

lemma inv_frame_fst:
"inv sh lev \<Gamma>' p (Star q f) (s2, H2) (s2', H2')
  = inv sh lev \<Gamma>' (Star p q) f (s2, H2) (s2', H2')"
  unfolding inv_def by (simp add: star_com23_)

lemma inv_frame_snd:
"inv sh lev \<Gamma>' q (Star p f) (s2, H2) (s2', H2') = inv sh lev \<Gamma>' (Star p q) f (s2, H2) (s2', H2')"
  unfolding inv_def by (simp add: star_com23_ star_com12_)

lemma pre_frame:
  assumes pre: "pre sh lev \<sigma> \<sigma>' (set \<Gamma>) (Star p1 p2) f c k k' (s1, H1) (s1', H1')"
  assumes p2:  "cmd_mod c \<inter> assert_free p2 = {}"
  shows "pre sh lev \<sigma> \<sigma>' (set \<Gamma>) p1 (Star p2 f) c k k' (s1, H1) (s1', H1')"
  using assms by (auto elim!: pre_elim intro!: pre_intro simp: Int_Un_distrib inv_frame_fst)  

lemma pre_par:
  assumes pre: "pre sh lev \<sigma> \<sigma>2 (set \<Gamma>) (Star p1 p2) f (Par c1 c2) k k' (s1, H1) (s1', H1')"
  assumes p1:  "cmd_mod c2 \<inter> assert_free p1 = {}"
  assumes p2:  "cmd_mod c1 \<inter> assert_free p2 = {}"
  obtains
    (fst) k1 k1' \<sigma>' \<sigma>2' where
      "k  = (par Fst c1 c2 k1)" 
      "k' = (par Fst c1 c2 k1')"
      "\<sigma> = Fst#\<sigma>'"
      "\<sigma>2 = Fst#\<sigma>2'"
      "pre sh lev \<sigma>' \<sigma>2' (set \<Gamma>) p1 (Star p2 f) c1 k1 k1' (s1, H1) (s1', H1')" |
    (snd) k2 k2' \<sigma>'  \<sigma>2' where
      "k  = (par Snd c1 c2 k2)" 
      "k' = (par Snd c1 c2 k2')"
      "\<sigma> = Snd#\<sigma>'"
      "\<sigma>2 = Snd#\<sigma>2'"
      "pre sh lev \<sigma>' \<sigma>2' (set \<Gamma>) p2 (Star p1 f) c2 k2 k2' (s1, H1) (s1', H1')"
using pre proof (cases rule: pre_elim)
  case 1
  from 1(4) show ?thesis proof (cases rule: step_par_split)
    case (fst \<sigma>' k1)
    from fst 1(1-3) 1(5-7) p2 show ?thesis (* don't add 1(4) or the last auto loops in simp *)
      by (auto elim!: step_par_fst intro!: that(1) pre_intro simp: Int_Un_distrib inv_frame_fst)
  next
    case (snd \<sigma>' k2)
    from snd 1(1-3) 1(5-7) p1 show ?thesis (* don't add pre(4) or the last auto loops in simp *)
      by (auto elim!: step_par_snd intro!: that(2) pre_intro simp: Int_Un_distrib inv_frame_snd)
  qed
qed

lemma step_par_fst':
  assumes "step sh lev c1 \<Gamma> p \<Gamma>' p' s1 s2"
  assumes "assert_lows lev q s1 \<subseteq> assert_lows lev q s2"
  shows   "step sh lev (Par c1 c2) \<Gamma> (Star p q) \<Gamma>' (Star p' q) s1 s2"
  using assms by (auto intro!: step_intro elim!: step_elim split: if_splits)

lemma step_par_snd':
  assumes "step sh lev c2 \<Gamma> q \<Gamma>' q' s1 s2"
  assumes "assert_lows lev p s1 \<subseteq> assert_lows lev p s2"
  shows   "step sh lev (Par c1 c2) \<Gamma> (Star p q) \<Gamma>' (Star p q') s1 s2"
  using assms by (auto intro!: step_intro elim!: step_elim split: if_splits)

lemma step_frame:
  assumes "step sh lev c \<Gamma> p \<Gamma>' p' s1 s2"
  assumes "assert_lows lev q s1 \<subseteq> assert_lows lev q s2"
  shows   "step sh lev c \<Gamma> (Star p q) \<Gamma>' (Star p' q) s1 s2"
  using assms by (auto intro!: step_intro elim!: step_elim split: if_splits)

lemma secure_seq:
  assumes c1: "secure sh lev n p c1 q"
  (* the additional assumption m \<le> is useful in the while rule *)
  assumes c2: "\<And> m. m \<le> n \<Longrightarrow> secure sh lev m q c2 r"
  shows   "secure sh lev n p (Seq c1 c2) r"
using assms proof (induction n arbitrary: c1 p, simp)
  case (Suc n)
  show ?case proof (intro secure_intro)
    fix \<Gamma>1 s1 s1' H1 H1' k k' f \<sigma> \<sigma>'
    assume
      "pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) p f (Seq c1 c2) k k' (s1, H1) (s1', H1')"
    then obtain  k1 k1' where
       k: "k = seq k1 c2" "k' = seq k1' c2" and
       "cmd_mod (Seq c1 c2) \<inter> assert_free f = {}"
       "cmd_mod (Seq c1 c2) \<inter> invvars = {}" "sched_eq \<sigma> \<sigma>'"
       "set \<Gamma>1 \<subseteq> locks" "cmd_step \<sigma> (Run c1 (set \<Gamma>1) s1 H1) k1" "cmd_step \<sigma>' (Run c1 (set \<Gamma>1) s1' H1') k1'" "local.inv sh lev \<Gamma>1 p f (s1, H1) (s1', H1')"
      by (auto elim: pre_elim dest!: step_seq)
    then have
      "pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) p f c1 k1 k1' (s1, H1) (s1', H1')"
      by (intro pre_intro)
         (simp add: Int_Un_distrib2)+
    from secure_step[OF Suc.prems(1) this]
    show "rec sh lev n (set \<Gamma>1) p f (Seq c1 c2) k k' r (s1, H1) (s1', H1') \<and> (ct_eq lev \<sigma> \<sigma>') \<or>
          asm lev \<sigma> s1 s1'" 
    proof(rule disj_forward, simp_all)
      assume "rec sh lev n (set \<Gamma>1) p f c1 k1 k1' q (s1, H1) (s1', H1') \<and> (ct_eq lev \<sigma> \<sigma>')"
      hence "rec sh lev n (set \<Gamma>1) p f c1 k1 k1' q (s1, H1) (s1', H1')" "(ct_eq lev \<sigma> \<sigma>')" by blast+
      thus "rec sh lev n (set \<Gamma>1) p f (Seq c1 c2) k k' r (s1, H1) (s1', H1')"
      proof (cases rule: rec_elim)
        case (stop \<Gamma>' s2 H2 s2' H2')
        with k show ?thesis
          by (auto elim!: step_elim intro!: step_intro Suc.prems(2) split: if_splits)
      next
        case (rec p2 \<Gamma>2 c2 s2 H2 s2' H2')
        with k show ?thesis 
          by (auto elim!: step_elim intro!: step_intro Suc.IH Suc.prems(2) split: if_splits)
      qed
    qed
  qed
qed

lemma secure_while:
  assumes body: "\<And> n. secure sh lev n (Star (Rel (Pure b)) (Star (Rel (SecB b (level_expr lev))) p)) c (Star (Rel (SecB b (level_expr lev))) p)"
  shows   "secure sh lev k (Star (Rel (SecB b (level_expr lev))) p) (While b c) (Star (Rel (Pure (Syntax.Not b))) p)"
proof (induction k rule: nat_less_induct, rule secure_intro')
  fix n m \<Gamma>1 s1 s1' H1 H1' k k' f \<sigma> \<sigma>'
  assume
    m: "n = Suc m"
  assume
    "\<forall>m<n. secure sh lev m (Star (Rel (SecB b (level_expr lev))) p) (While b c) (Star (Rel (Pure (Syntax.Not b))) p)"
  with m have
    hyp: "\<And> k. k\<le>m \<Longrightarrow> secure sh lev k (Star (Rel (SecB b (level_expr lev))) p) (While b c) (Star (Rel (Pure (Syntax.Not b))) p)"
    by auto
  assume
    "pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) (Star (Rel (SecB b (level_expr lev))) p) f (While b c) k k' (s1, H1) (s1', H1')"
  then show
    "rec sh lev m (set \<Gamma>1) (Star (Rel (SecB b (level_expr lev))) p) f (While b c) k k' (Star (Rel (Pure (Syntax.Not b))) p) (s1, H1) (s1', H1')
     \<and> (ct_eq lev \<sigma> \<sigma>')
      \<or>
     asm lev \<sigma> s1 s1'"
  proof (cases rule: pre_elim)
    case 1
    from 1(7) have
      bs: "pure_sem b s1 = pure_sem b s1'"
      by simp

    from cmd_step_while 1(4) 1(5) bs have seq: "\<sigma> = \<sigma>'" by metis
    show ?thesis proof (cases "pure_sem b s1")
      case True
      have inv: "inv sh lev \<Gamma>1 (Star (Rel (Pure b)) (Star (Rel (SecB b (level_expr lev))) p)) f (s1, H1) (s1', H1')"
        using True 1(7) bs by simp
      from True 1(1-5) bs show ?thesis
        apply (auto dest!: While_unwind)
        apply (rule inv)
        using seq by (auto intro!: step_intro secure_seq body hyp seq)
    next
      case False
      with seq 1 bs show ?thesis
        by (auto dest!: While_exit intro: step_intro)
    qed
  qed
qed

lemma secure_frame:
  assumes
    "secure sh lev n p c q"
    "cmd_mod c \<inter> assert_free r = {}"
  shows "secure sh lev n (Star p r) c (Star q r)"
using assms proof (induction n arbitrary: c p q rule: nat_less_induct)
  case (1 m)
  then have hyp:
    "\<And> n p c q.
      \<lbrakk>n < m;
      secure sh lev n p c q;
      cmd_mod c \<inter> assert_free r = {}\<rbrakk>
      \<Longrightarrow> secure sh lev n (Star p r) c (Star q r)"
    by blast

  note sec = 1(2)
  note r = 1(3)

  show ?case proof (rule secure_intro')
    fix n \<Gamma>1 s1 s1' H1 H1' k k' f \<sigma> \<sigma>'
    assume
      m: "m = Suc n"
    then have
      "n < m"  by simp
    note hyp = hyp[OF this]

    from m sec have
      sec: "secure sh lev (Suc n) p c q" by simp
    assume
      pre: "pre sh lev \<sigma> \<sigma>' (set \<Gamma>1) (Star p r) f c k k' (s1, H1) (s1', H1')"

    note pre_frame[OF pre r]
    note rec = secure_step[OF sec this]

    from rec
    show "rec sh lev n (set \<Gamma>1) (Star p r) f c k k' (Star q r) (s1, H1) (s1', H1') \<and>
          (ct_eq lev \<sigma> \<sigma>') \<or>
          asm lev \<sigma> s1 s1'"
    proof(rule disj_forward,simp_all)
      assume ras: "rec sh lev n (set \<Gamma>1) p (Star r f) c k k' q (s1, H1) (s1', H1') \<and> (ct_eq lev \<sigma> \<sigma>')"
      hence ra: "rec sh lev n (set \<Gamma>1) p (Star r f) c k k' q (s1, H1) (s1', H1')"  and
            seq: "(ct_eq lev \<sigma> \<sigma>')" by blast+
      have "rec sh lev n (set \<Gamma>1) (Star p r) f c k k' (Star q r) (s1, H1) (s1', H1')"
      proof (cases rule: rec_elim[OF ra])
        case (1 s2 H2 s2' H2')
        with pre r show ?thesis 
          by (auto simp: inv_frame_fst intro!: step_frame elim!: pre_elim_stop elim: cmd_coincidences)
      next
        case (2 p2 \<Gamma>2 c2 s2 H2 s2' H2')
        with pre r show ?thesis
          by (auto simp: inv_frame_fst intro!: hyp step_frame elim!: pre_elim_run elim: cmd_coincidences)
      qed
      thus "rec sh lev n (set \<Gamma>1) (Star p r) f c k k' (Star q r) (s1, H1) (s1', H1')" 
        by blast
    qed
  qed
qed

lemma secure_frame_com:
"secure sh lev n (Star p1 p2) c (Star q1 q2)
  = secure sh lev n (Star p2 p1) c (Star q2 q1)"
  by (auto intro: secure_conseq)

lemma asm_par_fst:
  "asm lev \<sigma>1 s1 s1' \<Longrightarrow>
   asm lev (Fst#\<sigma>1) s1 s1'"
  by(auto elim!: asm.cases intro: asm.intros)

lemma asm_par_snd:
  "asm lev \<sigma>1 s1 s1' \<Longrightarrow>
   asm lev (Snd#\<sigma>1) s1 s1'"
  by(auto elim!: asm.cases intro!: asm.intros simp: inv_def star_com12_)

lemma ct_eq_Cons_refl[intro,simp]: "ct_eq lev \<sigma> \<sigma>' \<Longrightarrow> ct_eq lev (a#\<sigma>) (a#\<sigma>')"
  by(clarsimp simp: ct_eq_def outs_def)
  

lemma secure_par:
  assumes
    "secure sh lev k p1 c1 q1"
    "secure sh lev k p2 c2 q2"
    "cmd_mod c1 \<inter> cmd_vars c2 = {}"
    "cmd_mod c2 \<inter> cmd_vars c1 = {}"
    "cmd_mod c1 \<inter> assert_free p2 = {}"
    "cmd_mod c1 \<inter> assert_free q2 = {}"
    "cmd_mod c2 \<inter> assert_free p1 = {}"
    "cmd_mod c2 \<inter> assert_free q1 = {}"
  shows "secure sh lev k (Star p1 p2) (Par c1 c2) (Star q1 q2)"
using assms proof (induction k arbitrary: c1 p1 c2 p2 rule: nat_less_induct)
  case (1 n)
  note sec1 = 1(2)
  note sec2 = 1(3)
  note mod = 1(4-)

  from 1(1) have hyp:
    "\<And> m c1 p1 c2 p2.
      m < n \<Longrightarrow>
      \<lbrakk>secure sh lev m p1 c1 q1;
       secure sh lev m p2 c2 q2;
       cmd_mod c1 \<inter> cmd_vars c2 = {};
       cmd_mod c2 \<inter> cmd_vars c1 = {};
       cmd_mod c1 \<inter> assert_free p2 = {};
       cmd_mod c1 \<inter> assert_free q2 = {};
       cmd_mod c2 \<inter> assert_free p1 = {};
       cmd_mod c2 \<inter> assert_free q1 = {}\<rbrakk>
    \<Longrightarrow> secure sh lev m (Star p1 p2) (Par c1 c2)(Star q1 q2)"
    by auto

  show ?case proof (rule secure_intro')
    fix m \<Gamma>1 s1 s1' H1 H1' k k' f \<sigma> \<sigma>2

    assume
      m: "n = Suc m"
    then have "m < n" by simp
    note hyp = hyp[OF this]

    from m sec1 sec2 have
      sec1: "secure sh lev (Suc m) p1 c1 q1" and
      sec2: "secure sh lev (Suc m) p2 c2 q2"
      by simp_all

    assume
      pre: "pre sh lev \<sigma> \<sigma>2 (set \<Gamma>1) (Star p1 p2) f (Par c1 c2) k k' (s1, H1) (s1', H1')"

    from pre have
      invvars: "cmd_mod c1 \<inter> invvars = {}" "cmd_mod c2 \<inter> invvars = {}"
      by (elim pre_elim, simp add: Int_Un_distrib2)+

    from pre mod(5) mod(3) show
      "rec sh lev m (set \<Gamma>1) (Star p1 p2) f (Par c1 c2) k k' (Star q1 q2) (s1, H1) (s1', H1')
       \<and> (ct_eq lev \<sigma> \<sigma>2) \<or>
       asm lev \<sigma> s1 s1'"

    proof (cases rule: pre_par)
      case (fst k1 k1' \<sigma>1 \<sigma>2')
      note pre1 = fst(5)
      note rec1 = secure_step[OF sec1 pre1]

      then show ?thesis proof(rule disj_forward)
        assume "rec sh lev m (set \<Gamma>1) p1 (Star p2 f) c1 k1 k1' q1 (s1, H1) (s1', H1') \<and> (ct_eq lev \<sigma>1 \<sigma>2')"
        hence "rec sh lev m (set \<Gamma>1) p1 (Star p2 f) c1 k1 k1' q1 (s1, H1) (s1', H1')" and seq: "(ct_eq lev \<sigma>1 \<sigma>2')" by blast+
        thus " rec sh lev m (set \<Gamma>1) (Star p1 p2) f (Par c1 c2) k k' (Star q1 q2) (s1, H1) (s1', H1') \<and> (ct_eq lev \<sigma> \<sigma>2)"
        proof (cases rule: rec_elim)
          case (stop s2 H2 s2' H2')
          have "secure sh lev m p2 c2 q2"
            by (rule secure_prefix, rule sec2) simp
          from this mod(6) have
            "secure sh lev m (Star q1 p2) c2 (Star q1 q2)"
            unfolding secure_frame_com by (rule secure_frame)
          with seq stop fst mod(3) show ?thesis
            by (auto simp: inv_frame_fst Int_Un_distrib elim: cmd_coincidences intro!: step_par_fst' elim!: pre_elim_stop)
        next
          case (rec p1' \<Gamma>2 c1' s2 H2 s2' H2')
          have
            "secure sh lev m p2 c2 q2"
            by (rule secure_prefix, rule sec2) simp
          with rec fst mod show ?thesis
            apply (auto simp: inv_frame_fst Int_Un_distrib elim: cmd_coincidences intro!: step_par_fst' elim!: pre_elim_run)
          proof (rule hyp, simp_all)
            assume
              "cmd_mod c1' \<subseteq> cmd_mod c1"
              "cmd_vars c1' \<subseteq> cmd_vars c1"
            with mod show
              "cmd_mod c1' \<inter> cmd_vars c2  = {}" 
              "cmd_mod c2  \<inter> cmd_vars c1' = {}"
              "cmd_mod c1' \<inter> assert_free p2 = {}"
              "cmd_mod c1' \<inter> assert_free q2 = {}"
              by auto
          next
            assume
              "step sh lev c1 \<Gamma>1 p1 \<Gamma>2 p1' s1 s2"
            then have
              p1': "assert_free p1' \<subseteq> assert_free p1 \<union> cmd_vars c1 \<union> invvars"
              by (auto elim!: step_elim)
            assume
              "cmd_mod c1' \<subseteq> cmd_mod c1"
              "cmd_vars c1' \<subseteq> cmd_vars c1"
            with p1' mod invvars show "cmd_mod c2 \<inter> assert_free p1' = {}"
              by auto
          next
            assume  "\<sigma> = Fst # \<sigma>1" "\<sigma>2 = Fst # \<sigma>2'"
            thus "ct_eq lev (Fst # \<sigma>1) (Fst # \<sigma>2')" using seq  by simp
          qed
        qed
      next
        assume "asm lev \<sigma>1 s1 s1'"
        thus "asm lev \<sigma> s1 s1' "
          using asm_par_fst
          by (simp add: fst(3))
      qed
    next
      case (snd k2 k2' \<sigma>1 \<sigma>2')
      note pre2 = snd(5)
      note rec2 = secure_step[OF sec2 pre2]
      then show ?thesis proof (rule disj_forward) 
        assume "rec sh lev m (set \<Gamma>1) p2 (Star p1 f) c2 k2 k2' q2 (s1, H1) (s1', H1')  \<and> (ct_eq lev \<sigma>1 \<sigma>2')"
        hence "rec sh lev m (set \<Gamma>1) p2 (Star p1 f) c2 k2 k2' q2 (s1, H1) (s1', H1')" and seq: "(ct_eq lev \<sigma>1 \<sigma>2')" by blast+
        thus "rec sh lev m (set \<Gamma>1) (Star p1 p2) f (Par c1 c2) k k' (Star q1 q2) (s1, H1) (s1', H1') \<and>
              (ct_eq lev \<sigma> \<sigma>2)"
        proof(cases rule: rec_elim)
          case (stop s2 H2 s2' H2')
          have
            "secure sh lev m p1 c1 q1"
            by (rule secure_prefix, rule sec1) simp
          note secure_frame[OF this mod(4)]
          with seq stop snd mod(5) show ?thesis
            by (auto simp: inv_frame_snd Int_Un_distrib elim: cmd_coincidences intro!: step_par_snd' elim!: pre_elim_stop)
        next
          case (rec p2' \<Gamma>2 c2' s2 H2 s2' H2')
          have
            "secure sh lev m p1 c1 q1"
            by (rule secure_prefix, rule sec1) simp
          with rec snd mod show ?thesis
            apply (auto simp: inv_frame_snd Int_Un_distrib elim: cmd_coincidences intro!: step_par_snd' elim!: pre_elim_run)
          proof (rule hyp, simp_all)
            assume
              "cmd_mod c2' \<subseteq> cmd_mod c2"
              "cmd_vars c2' \<subseteq> cmd_vars c2"
            with mod show
              "cmd_mod c1  \<inter> cmd_vars c2' = {}" 
              "cmd_mod c2' \<inter> cmd_vars c1  = {}"
              "cmd_mod c2' \<inter> assert_free p1 = {}"
              "cmd_mod c2' \<inter> assert_free q1 = {}"
              by auto
          next
            assume
              "step sh lev c2 \<Gamma>1 p2 \<Gamma>2 p2' s1 s2"
            then have
              p2': "assert_free p2' \<subseteq> assert_free p2 \<union> cmd_vars c2 \<union> invvars"
              by (auto elim!: step_elim)
            assume
              "cmd_mod c2' \<subseteq> cmd_mod c2"
              "cmd_vars c2' \<subseteq> cmd_vars c2"
            with p2' mod invvars show "cmd_mod c1 \<inter> assert_free p2' = {}"
              by auto
          next
            assume "\<sigma> = Snd # \<sigma>1" "\<sigma>2 = Snd # \<sigma>2'"
            thus "ct_eq lev (Snd # \<sigma>1) (Snd # \<sigma>2')" using seq  by simp
          qed
        qed
      next
        assume "asm lev \<sigma>1 s1 s1'"
        thus "asm lev \<sigma> s1 s1'"
          using asm_par_snd
          by (simp add: snd(3))
      qed
    qed
  qed
qed

lemma step_refl[simp]:
"step sh lev c \<Gamma> p \<Gamma> p s s = (set \<Gamma> \<subseteq> locks)"
  unfolding step.simps by auto

lemma step_trans:
  assumes "step sh lev c1 \<Gamma>1 p1 \<Gamma>2 p2 s1 s2"
  assumes "step sh lev c2 \<Gamma>2 p2 \<Gamma>3 p3 s2 s3"
  assumes "cmd_vars c2 \<subseteq> cmd_vars c1"
  shows   "step sh lev c1 \<Gamma>1 p1 \<Gamma>3 p3 s1 s3"
  using assms by (elim step_elim, intro step_intro, auto split: if_splits)

fun no_asm :: "('level,'a) sched \<Rightarrow> bool" where
  "no_asm ((Assumption p)#\<sigma>) = False" |
  "no_asm (a#\<sigma>) = no_asm \<sigma>" |
  "no_asm [] = True"

lemma no_asm_Cons:
  "no_asm (a#as) = (if (\<exists>p. a = Assumption p) then False else no_asm as)"
  by(cases a, auto)

lemma no_asm_filter:
  "no_asm xs = (filter (\<lambda>x. \<exists> p. x = Assumption p) xs = [])"
  by(induct xs, auto simp: no_asm_Cons)

lemma no_asm_ConsD:
  "no_asm (a#\<sigma>) \<Longrightarrow> no_asm \<sigma>"
  by(case_tac a, auto)

lemma no_asm_appD:
  "no_asm \<sigma> \<Longrightarrow> \<sigma> = \<sigma>1@\<sigma>2 \<Longrightarrow> no_asm \<sigma>1 \<and> no_asm \<sigma>2"
  using no_asm_filter 
  by simp

lemma sched_asm_no_asm[simp]:
  "sched_asm \<sigma> = Some p \<Longrightarrow> no_asm \<sigma> \<Longrightarrow> False"
  by(induct rule: sched_asm.induct, auto)

lemma asm_no_asm[dest]:
  "asm lev \<sigma>1 s1 s1' \<Longrightarrow> no_asm \<sigma>1 \<Longrightarrow> False"
  apply(induct rule: asm.induct)
  by(blast dest: sched_asm_no_asm)

lemma sched_eq_appI:
  "sched_eq \<sigma>1 \<sigma>1' \<Longrightarrow> sched_eq \<sigma>2 \<sigma>2' \<Longrightarrow> sched_eq (\<sigma>1 @ \<sigma>2) (\<sigma>1' @ \<sigma>2')"
  by(induct \<sigma>1 \<sigma>1'  arbitrary: \<sigma>2 \<sigma>2' rule: sched_eq.induct, auto)

lemma ct_eq_appI[intro]:
  "ct_eq lev \<sigma> \<sigma>' \<Longrightarrow> ct_eq lev \<sigma>2 \<sigma>2' \<Longrightarrow> ct_eq lev (\<sigma>@\<sigma>2) (\<sigma>'@\<sigma>2')"
  by(auto simp: ct_eq_def outs_def)


lemma secure_steps_ex:
  assumes
    "cmd_steps \<sigma> (Run c1 (set \<Gamma>1) s1  H1)  k"
  assumes
    "secure sh lev (length \<sigma>) p1 c1 q"
    "no_asm \<sigma>"
    "cmd_mod c1 \<inter> assert_free f = {}"
    "cmd_mod c1 \<inter> invvars = {}"
    "set \<Gamma>1 \<subseteq> locks"
    "inv sh lev \<Gamma>1 p1 f (s1,H1) (s1',H1')"
  obtains
    (stop) \<sigma>' \<Gamma> s H s' H' k' where
      "cmd_steps \<sigma>' (Run c1 (set \<Gamma>1) s1' H1') k'"
      "k = Stop (set \<Gamma>) s H"
      "k' = Stop (set \<Gamma>) s' H'"
      "inv sh lev \<Gamma> q f (s,H) (s',H')"
      "step sh lev c1 \<Gamma>1 p1 \<Gamma> q s1 s"
      "sched_eq \<sigma> \<sigma>'" "ct_eq lev \<sigma> \<sigma>'" |
    (rec) p \<sigma>' \<Gamma> c s H s' H' k' where
      "cmd_steps \<sigma>' (Run c1 (set \<Gamma>1) s1' H1') k'"
      "k = Run c (set \<Gamma>) s H"
      "k' = Run c (set \<Gamma>) s' H'"
      "inv sh lev \<Gamma> p f (s,H) (s',H')"
      "step sh lev c1 \<Gamma>1 p1 \<Gamma> p s1 s"
      "sched_eq \<sigma> \<sigma>'" "ct_eq lev \<sigma> \<sigma>'"
  using assms
proof (induction  \<sigma> "(Run c1 (set \<Gamma>1) s1  H1)" k arbitrary: c1 \<Gamma>1 s1 H1 p1 q s1' H1' rule: cmd_steps.induct)
  case cmd_steps_refl
  then show ?case
    by auto
next
  case (cmd_steps_step \<sigma>1 k1 \<sigma>2 k2)
  from cmd_step_Run[OF this(1)] obtain \<sigma>1' k1' where 
    step': "cmd_step \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k1'" and seq: "sched_eq \<sigma>1 \<sigma>1'" by metis
  have p: "pre sh lev \<sigma>1 \<sigma>1' (set \<Gamma>1) p1 f c1 k1 k1' (s1,H1) (s1',H1')"
    using pre_intro 
    using \<open>cmd_step \<sigma>1' (Run c1 (set \<Gamma>1) s1' H1') k1'\<close> cmd_steps_step.hyps(1) cmd_steps_step.prems(5) cmd_steps_step.prems(6) cmd_steps_step.prems(7) cmd_steps_step.prems(8) seq by blast
  have ne': "\<sigma>1' \<noteq> []" using cmd_step_sched_nonempty step' by metis
  hence ne: "\<sigma>1 \<noteq> []" using seq by(cases \<sigma>1, auto)
  from this obtain n where n: "length (\<sigma>1 @ \<sigma>2) = Suc n" 
    by (meson append_is_Nil_conv length_0_conv not0_implies_Suc)
  with  secure_step[OF _ p] have
    "rec sh lev n (set \<Gamma>1) p1 f c1 k1 k1' q (s1, H1) (s1', H1') \<and> (ct_eq lev \<sigma>1 \<sigma>1') \<or> asm lev \<sigma>1 s1 s1'"
    using cmd_steps_step.prems(3) by auto
  with `no_asm (\<sigma>1 @ \<sigma>2)` have
    r: "rec sh lev n (set \<Gamma>1) p1 f c1 k1 k1' q (s1, H1) (s1', H1') \<and> (ct_eq lev \<sigma>1 \<sigma>1')"
    using no_asm_appD by blast

  (* FIXME: clean up this mess *)
  thus ?case
    apply -
    apply(erule conjE)
    apply(erule rec_elim)
      apply(rule cmd_steps_step(4))
         apply(rule cmd_steps.cmd_steps_step)
            apply(rule step')
         apply clarsimp
           apply(rule conjI, rule HOL.refl, rule HOL.refl)

    using cmd_steps_Stop
    using cmd_steps_step.hyps(2) apply blast
         apply(rule HOL.refl)
(*
    using cmd_steps_Stop
    using cmd_steps_step.hyps(2) apply blast
       apply(rule HOL.refl) *)
      apply assumption
       apply assumption 
      apply(rule sched_eq_appI, rule seq)
      apply(subgoal_tac "\<sigma>2 = []")
       apply simp

    using cmd_steps_step.hyps(2) 
      apply blast

    using cmd_steps_step.hyps(2) 
    apply (metis Stop_dest append_Nil2)

    apply(rule cmd_steps_step(3))
            apply fastforce
           apply(rule cmd_steps_step(4))
               apply(rule cmd_steps.cmd_steps_step)
                apply(rule step') 
               apply clarsimp
               apply assumption
              apply fastforce
             apply fastforce
            apply assumption 
            apply clarsimp
            apply(rule step_trans)
              apply assumption
             apply assumption
    using cmd_step_preserves_vars cmd_steps_step.hyps(1) conf_vars.simps(3) 
             apply blast
            apply(blast intro: sched_eq_appI seq)
           apply blast
          apply(rule cmd_steps_step(5))
              apply(rule cmd_steps.cmd_steps_step)
               apply(rule step')
              apply clarsimp
              apply assumption
             apply fastforce
            apply fastforce
           apply assumption
          apply clarsimp
          apply(rule step_trans)
            apply assumption
           apply assumption
    using cmd_step_preserves_vars cmd_steps_step.hyps(1) conf_vars.simps(3) 
            apply blast
           apply(blast intro: sched_eq_appI seq)
          apply blast
         apply(erule secure_prefix)
    using n ne 
         apply(cases \<sigma>1, auto)[1]
    using `no_asm (\<sigma>1@\<sigma>2)` no_asm_appD apply metis
    using cmd_step_preserves_mod cmd_steps_step.prems(5) conf_mod.simps(3) step' apply blast
    using cmd_step_preserves_mod cmd_steps_step.prems(6) conf_mod.simps(3) step' apply blast         
    using step_elim apply blast          
    by assumption
qed

end

end